/*
 * common.cpp
 *
 *   This is the common code used by UNet user. Mostly, the code of the file
 *   comes from libunet.c, with fixes in compiling warnings or errors.
 *
 * Matt Welsh
 * Xun Wilson Huang
 * Hoi So
 */

#include "unetspi.h"
#include <unetuser.h>
#include <libunet.h>

int str_to_addr(UNET_DEV_TYPE devtype, char *str, UNET_CHANNEL_ADDRESS *addr) 
{
	if (devtype == UNET_DEVID_TULIP) {
		unsigned int mac_addr[6];
		unsigned int port;
		if (sscanf(str,"%x:%x:%x:%x:%x:%x.%x",
				&(mac_addr[0]),
				&(mac_addr[1]),
				&(mac_addr[2]),
				&(mac_addr[3]),
				&(mac_addr[4]),
				&(mac_addr[5]),
				&port) != 7) return -1;

		addr->fe.macAddr[0] = (unsigned char)mac_addr[0];
		addr->fe.macAddr[1] = (unsigned char)mac_addr[1];
		addr->fe.macAddr[2] = (unsigned char)mac_addr[2];
		addr->fe.macAddr[3] = (unsigned char)mac_addr[3];
		addr->fe.macAddr[4] = (unsigned char)mac_addr[4];
		addr->fe.macAddr[5] = (unsigned char)mac_addr[5];
		addr->fe.port = (unsigned char)port;
		return 0;
	}

	if ((devtype == UNET_DEVID_SBA200) || (devtype == UNET_DEVID_PCA200)) {
		if (sscanf(str,"%d",&(addr->atm.vci)) != 1) return -1;
		return 0;
	}

	return -1;
}

int UnetAddr_addr ( UNET_DEV_TYPE devtype,  sockaddr_unet_fe * feaddr, 
		UNET_CHANNEL_ADDRESS  * addr)
{
	if ( devtype == UNET_DEVID_TULIP ) {
		addr->fe.port = (unsigned char )feaddr->port;
		for (int i=0;i<6;i++) {
			addr->fe.macAddr[i] = (unsigned char)(feaddr->addr.addr[i]);
		}
	}
	return 0;
}

int str_to_unetaddr ( char* str, unsigned short port, sockaddr_unet_fe* addr)
{
	unsigned int mac_addr[6];
	if (sscanf(str,"%x:%x:%x:%x:%x:%x", &(mac_addr[0]),
			&(mac_addr[1]), &(mac_addr[2]), &(mac_addr[3]),
			&(mac_addr[4]), &(mac_addr[5])) != 6) 
		return -1;
	for (int  i=0;i<6;i++)
		addr->addr.addr[i] = (char)mac_addr[i];
	addr->port = port;
	return 0;
}

int compose_addr(UNET_DEV_TYPE devtype, 
		char                       *str,
		unsigned int               port, 
		UNET_CHANNEL_ADDRESS       *addr) 
{
	if (devtype == UNET_DEVID_TULIP) {
		unsigned int mac_addr[6];
		int i;

		if (sscanf(str,"%x:%x:%x:%x:%x:%x", &(mac_addr[0]),
				&(mac_addr[1]), &(mac_addr[2]), &(mac_addr[3]),
				&(mac_addr[4]), &(mac_addr[5])) != 6) 
			return -1;

		for ( i=0;i<6;i++)
			addr->fe.macAddr[i] = (unsigned char)mac_addr[i];
		addr->fe.port = (unsigned char)port;
		return 0;
	}

	if ((devtype == UNET_DEVID_SBA200) || (devtype == UNET_DEVID_PCA200)) {
		if (sscanf(str,"%d",&(addr->atm.vci)) != 1) return -1;
		return 0;
	}

	return -1;
}

/* Create an endpoint on the given (open) U-Net device */
struct unet_endpt_user *unet_create_endpt(HANDLE hDriver,
		UNET_INFO_RESPONSE *info,
		UNET_CREATE_ENDPOINT_REQUEST *create,
		int num_rx_buffers) {

	struct unet_endpt_user *endpt;
	UNET_CREATE_ENDPOINT_RESPONSE response;
	// caddr_t comseg;
	int nbytes;
	unsigned int i;

	if (!DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_CREATE_ENDPOINT,
			create, sizeof(UNET_CREATE_ENDPOINT_REQUEST),
			&response, sizeof(UNET_CREATE_ENDPOINT_RESPONSE),
			(unsigned long *) &nbytes, 0)) {
		Perror("IOCTL_UNET_CREATE_ENDPOINT");
		return NULL;
	}

	/* Create endpoint struct */
	endpt = (struct unet_endpt_user *)malloc(sizeof(struct unet_endpt_user));
	if (endpt == NULL) {
		CloseHandle(hDriver);  /* XXX need to destroy endpoint here */
		return NULL;
	}

	endpt->endpointId = response.endpointId;
	endpt->hDriver = hDriver;
	endpt->info = *info;

	endpt->tx_first = response.txFifo;
	endpt->tx_next = endpt->tx_first;
	endpt->tx_last = endpt->tx_first + (create->numTxFifoEntries - 1);
	endpt->tx_wall = (volatile unsigned int *) response.txWall;
	endpt->tx_wall_cache = (UNET_TX_BUFFER_DESCRIPTOR *)((u_int)endpt->tx_first +
			*(endpt->tx_wall));

	endpt->free_first = response.freeFifo;
	endpt->free_next = endpt->free_first;
	endpt->free_last = endpt->free_first + (create->numFreeFifoEntries - 1);
	endpt->free_wall = (volatile unsigned int *) response.freeWall;

	endpt->rx_first = response.rxFifo;
	endpt->rx_next = endpt->rx_first;
	endpt->rx_last = endpt->rx_first + (create->numRxFifoEntries - 1);
	endpt->rx_wall = (volatile unsigned int *) response.rxWall;

	endpt->buffer_area = (unsigned char *) response.bufferArea;
	endpt->rxb_area = endpt->buffer_area;
	endpt->rxb_numbuff = num_rx_buffers;
	// endpt->rxb_buffsize = create->rxBufferSize;
	endpt->txb_area = endpt->buffer_area + (num_rx_buffers * create->rxBufferSize);

	/* Set up the Rx FIFO */
	for (i = 0; i < create->numRxFifoEntries; i++) {
		(endpt->rx_first + i)->chan = UNET_CHANNEL_FREE;
	}

	/* Set up the free FIFO. Note that we only push on as many entries as
	 * we have Rx buffers for: Also important is that free_next is incremented
	 * as we do so. This means that new entries are pushed on starting AFTER
	 * these initial entries (if there are more free FIFO entries than Rx
	 * buffers).
	 */

	/* Check that the Rx buffers will be properly aligned for the device */
	if (create->rxBufferSize & (endpt->info.rx_buf_align - 1)) {
		CloseHandle(hDriver);  /* XXX Should destroy endpoint here */
		free(endpt);
		return(NULL);
	}

	for (i = 0; i < MIN((create->numFreeFifoEntries), (unsigned int)num_rx_buffers); i++) {
		endpt->free_next->buf = (i * create->rxBufferSize);
		if (endpt->free_next == endpt->free_last) 
			endpt->free_next = endpt->free_first;
		else
			endpt->free_next++;
	}

	/* Initialize the Rx wall pointer */
	*(endpt->rx_wall) = (u_int)endpt->rx_last - (u_int)endpt->rx_first;

	endpt->rxwall_deferred = 0;

	return endpt;
}


/* Destroy the endpoint on the (given) open U-Net device */
int unet_destroy_endpt(HANDLE hDriver, UNET_ENDPOINT_ID endpointId) {

	UNET_DESTROY_ENDPOINT_REQUEST destroy;
	int ret, nbytes;

	destroy.endpointId = endpointId;

	if (!(ret = DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_DESTROY_ENDPOINT,
			&destroy, sizeof(UNET_DESTROY_ENDPOINT_REQUEST),
			NULL, 0, (unsigned long *)&nbytes, 0))) {
		Perror("IOCTL_UNET_DESTROY_ENDPT");
		return ret;
	}
	return ret;
}

/* Activate a channel on the given U-Net device */
UNET_CHANNEL_ID unet_activate_chan(HANDLE hDriver,
		UNET_ACTIVATE_CHANNEL_REQUEST *activate) {

	int nbytes;
	UNET_ACTIVATE_CHANNEL_RESPONSE response;
	if (!DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_ACTIVATE_CHANNEL,
			activate, sizeof(UNET_ACTIVATE_CHANNEL_REQUEST),
			&response, sizeof(response),
			(unsigned long *)&nbytes, 0)) {
		Perror("IOCTL_UNET_ACTIVATE_CHANNEL");
		return (UNET_CHANNEL_ID)-1;
	}
	return response.chan;
}

/* Deactivate a channel on the given U-Net device */
int unet_deactivate_chan(HANDLE hDriver,
		UNET_DEACTIVATE_CHANNEL_REQUEST *deactivate) 
{
	int nbytes, ret;
	if (!(ret = DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_DEACTIVATE_CHANNEL,
			deactivate,
			sizeof(UNET_DEACTIVATE_CHANNEL_REQUEST),
			NULL, 0, (unsigned long *)&nbytes, 0))) {
		Perror("IOCTL_UNET_DEACTIVATE_CHANNEL");
		return ret;
	}
	return ret;
}

char * Strerror(int errnum)
{
	char *MsgBuf;

	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER,
			NULL,  // message format
			errnum,
			0,   // language
			(char*) &MsgBuf,
			0,   // buffer size, ignored for ALLOC_BUF
			NULL // args for msg format
							 );
	return MsgBuf;
}

void Perror(char *s)
{
	char *error;
	int errno;

	errno = GetLastError();
	error = Strerror(errno);
	printf("%s: (%d) %s\n", s, errno, error);
	LocalFree(error);
}


HANDLE unet_opendev(UNET_DEV_TYPE devtype, UNET_DEV_INDEX devindex,
		UNET_INFO_RESPONSE *info) 
{
	HANDLE hDriver;
	int nadapters, nbytes, i;
	UNET_INFO_RESPONSE *adapterlist;

	hDriver = CreateFile("\\\\.\\UNET", GENERIC_READ | GENERIC_WRITE,
			0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
			NULL);
	if (hDriver == (HANDLE)-1) {
		Perror("unet_opendev: CreateFile");
	}

	if (!DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_ADAPTERS, NULL, 0,
			&nadapters, sizeof(ULONG), (unsigned long*)&nbytes, 0)) {
		Perror("IOCTL_UNET_ADAPTERS");
		CloseHandle(hDriver);
		return (HANDLE)NULL;
	}

	adapterlist = (UNET_INFO_RESPONSE *)malloc(nadapters * 
			sizeof(UNET_INFO_RESPONSE));
	if (!adapterlist) return (HANDLE)NULL;

	if (!DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_ADAPTER_LIST, NULL, 0,
			adapterlist, nadapters * sizeof(UNET_INFO_RESPONSE),
			(unsigned long*)&nbytes, 0)) {
		Perror("IOCTL_UNET_ADAPTER_LIST");
		CloseHandle(hDriver);
		return (HANDLE)NULL;
	}

	for (i = 0; i < nadapters; i++) {
		if ((adapterlist[i].dev_type == devtype) &&
				(adapterlist[i].dev_index == devindex)) {
			*info = adapterlist[i];
			free(adapterlist);
			return hDriver;
		}
	}

	free(adapterlist);
	CloseHandle(hDriver);
	return (HANDLE)NULL;
}
