/*
 * Copyright (c) 1996 by Cornell University.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL CORNELL UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF CORNELL UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * CORNELL UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND CORNELL UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 */

#include <ntddk.h>
#include "string.h"
#include "unetuser.h"
#include "unetkernel.h"

#ifndef UNET_USE_SLOW_TRAP
#include "utrap.h"
#endif

#undef UNET_DEBUG

#ifdef UNET_DEBUG
#define PRINTK DbgPrint
#else
#define PRINTK 0 &&
#endif

//
// Define driver entry routine.
//

NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    );

//
// Define the local routines used by this driver module.
//
NTSTATUS
UnetConfigure(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath,
    IN PUNET_OBJECT UnetObject
    );

static
NTSTATUS
UnetDispatch(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    );

VOID
UnetUnload(
    IN PDRIVER_OBJECT DriverObject
    );

static
BOOLEAN
UnetFastIoRead(
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    OUT PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN PDEVICE_OBJECT DeviceObject
    );

static
BOOLEAN
UnetFastIoWrite(
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    IN PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN PDEVICE_OBJECT DeviceObject
    );

static
BOOLEAN
UnetFastIoDeviceControl(
    IN PFILE_OBJECT FileObject,
    IN BOOLEAN Wait,
    IN PVOID InputBuffer OPTIONAL,
    IN ULONG InputBufferLength,
    OUT PVOID OutputBuffer OPTIONAL,
    IN ULONG OutputBufferLength,
    IN ULONG IoControlCode,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN PDEVICE_OBJECT DeviceObject
    );

NTSTATUS
UnetCreateEndpoint(
    IN PFILE_OBJECT FileObject,
    IN PUNET_CREATE_ENDPOINT_REQUEST UnetCreateEndpointRequest,
    OUT PUNET_CREATE_ENDPOINT_RESPONSE UnetCreateEndpointResponse,
    IN PDEVICE_OBJECT DeviceObject
    );

NTSTATUS
UnetDestroyEndpoint(
    IN PFILE_OBJECT FileObject,
    IN PUNET_DESTROY_ENDPOINT_REQUEST UnetDestroyEndpointRequest,
    IN PDEVICE_OBJECT DeviceObject
    );

NTSTATUS
UnetActivateChannel(
    IN PFILE_OBJECT FileObject,
    PUNET_ACTIVATE_CHANNEL_REQUEST  UnetActivateChannelRequest,
    PUNET_ACTIVATE_CHANNEL_RESPONSE UnetActivateChannelResponse,
    IN PDEVICE_OBJECT DeviceObject
    );

NTSTATUS
UnetDeactivateChannel(
    IN PFILE_OBJECT FileObject,
    PUNET_DEACTIVATE_CHANNEL_REQUEST UnetDeactivateChannelRequest,
    IN PDEVICE_OBJECT DeviceObject
    );

NTSTATUS
UnetInformation(
    IN PFILE_OBJECT FileObject,
    PUNET_ENDPOINT_ID UnetEndpointId,
    OUT PUNET_INFO_RESPONSE UnetInfoResponse,
    IN PDEVICE_OBJECT DeviceObject
    );

PUNET_ENDPOINT
UnetCreateEndpointObject(
    IN PUNET_OBJECT UnetObject
    );


NTSTATUS
UnetDestroyEndpointObject(
    PUNET_ENDPOINT  UnetEndpoint
    );

PUNET_ENDPOINT
UnetFindEndpointObject(
    PUNET_OBJECT UnetObject,
    UNET_ENDPOINT_ID UnetEndpointId
    );

PUNET_ADAPTER
UnetCreateAdapterObject(
    UNET_DEV_TYPE UnetDevType,
    UNET_DEV_INDEX UnetDevIndex
    );

NTSTATUS
UnetDestroyAdapterObject(
    UNET_DEV_TYPE UnetDevType,
    UNET_DEV_INDEX UnetDevIndex
    );

PUNET_ADAPTER
UnetFindAdapterObject(
    UNET_DEV_TYPE UnetDevType,
    UNET_DEV_INDEX UnetDevIndex
    );

NTSTATUS
UnetNumAdapters(
    IN PFILE_OBJECT FileObject,
    OUT PULONG UnetAdapters,
    IN PDEVICE_OBJECT DeviceObject
    );

NTSTATUS
UnetAdapterList(
    IN PFILE_OBJECT FileObject,
    IN ULONG Nadapters,
    OUT PUNET_INFO_RESPONSE UnetAdapters,
    IN PDEVICE_OBJECT DeviceObject
    );

NTSTATUS
UnetCloseEndpoints(
    IN PFILE_OBJECT FileObject,
    IN PDEVICE_OBJECT DeviceObject
    );

NTSTATUS
UnetCloseAllEndpoints(
    IN PDEVICE_OBJECT DeviceObject
    );

NTSTATUS
UnetTrapFunction(
    IN ULONG TrapArgument
    );

NTSTATUS
UnetStartTestAdapter();

//
// Assign text sections for each routine.
//

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, DriverEntry)
#pragma alloc_text(PAGE, UnetDispatch)
#pragma alloc_text(PAGE, UnetFastIoRead)
#pragma alloc_text(PAGE, UnetFastIoWrite)
#endif

static PUNET_ADAPTER UnetAdapters = NULL;
static PUNET_OBJECT  TrapObject = NULL;

NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
    UNICODE_STRING              deviceNameUnicodeString;
    UNICODE_STRING              deviceLinkUnicodeString;
    PDEVICE_OBJECT              deviceObject;
    NTSTATUS                    status;
    PFAST_IO_DISPATCH           fastIoDispatch;

    PRINTK("unet: Entered the Unet driver\n");

    RtlInitUnicodeString( &deviceNameUnicodeString, UNET_DEVICE_NAME);

    status = IoCreateDevice( DriverObject,
                             sizeof(UNET_OBJECT),
                             &deviceNameUnicodeString,
                             FILE_DEVICE_UNET,
                             0,
                             FALSE,
                             &deviceObject );

    if (!NT_SUCCESS( status )) {
        return status;
    }

    status = UnetConfigure(DriverObject, RegistryPath,
                    (PUNET_OBJECT)deviceObject->DeviceExtension);

    if (!NT_SUCCESS( status )) {
        IoDeleteDevice (deviceObject);
        return status;
    }

    DriverObject->MajorFunction[IRP_MJ_CREATE]            =
    DriverObject->MajorFunction[IRP_MJ_CLOSE]             =
    DriverObject->MajorFunction[IRP_MJ_READ]              =
    DriverObject->MajorFunction[IRP_MJ_WRITE]             =
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]    = UnetDispatch;
    DriverObject->DriverUnload = UnetUnload;

    RtlInitUnicodeString(&deviceLinkUnicodeString, UNET_DOSDEVICE_NAME);

    status = IoCreateSymbolicLink (&deviceLinkUnicodeString,
                                         &deviceNameUnicodeString);

    if (!NT_SUCCESS(status))
    {
    	PRINTK("unet: IoCreateSymbolicLink failed\n");
    	/* try to unlink it if it was hanging around from the last time */
        IoDeleteSymbolicLink (&deviceLinkUnicodeString);
        IoDeleteDevice (deviceObject);
    	return status;
    }

    fastIoDispatch = ExAllocatePool( NonPagedPool, sizeof( FAST_IO_DISPATCH ) );
    if (!fastIoDispatch) {
        IoDeleteDevice( deviceObject );
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    RtlZeroMemory( fastIoDispatch, sizeof( FAST_IO_DISPATCH ) );
    fastIoDispatch->SizeOfFastIoDispatch = sizeof( FAST_IO_DISPATCH );
    fastIoDispatch->FastIoRead           = UnetFastIoRead;
    fastIoDispatch->FastIoWrite          = UnetFastIoWrite;
    fastIoDispatch->FastIoDeviceControl  = UnetFastIoDeviceControl;
    DriverObject->FastIoDispatch = fastIoDispatch;

#if 0
    status = UnetStartTestAdapter();
    if (status != 0)
    {
    	PRINTK("unet: UnetStartTestAdapter failed\n");
    	IoDeleteSymbolicLink (&deviceLinkUnicodeString);
    	ExFreePool(fastIoDispatch);
        IoDeleteDevice (deviceObject);
    	return status;
    }
#endif
    // The first of the true hacks, I have to globalize this object, otherwise the
    // trap can't reach the endpoints. This effectively means that we cannot have
    // multiple unet devices. I should reprogram to try to take that level of
    // indirection out of the code
    TrapObject = (PUNET_OBJECT)deviceObject->DeviceExtension;

#ifndef UNET_USE_SLOW_TRAP
    UTrapConnect(UNET_TRAP_VECTOR, UnetTrapFunction);
#endif

    return STATUS_SUCCESS;
}

NTSTATUS
UnetConfigure(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath,
    IN PUNET_OBJECT UnetObject
    )
{
    WCHAR                       subKeyString[] = L"\\Parameters";
    UNICODE_STRING              paramPath;
    RTL_QUERY_REGISTRY_TABLE    paramTable[10];
    NTSTATUS                    status;

    ULONG   defaultMaxEndpoints     = UNET_DEFAULT_MAX_ENDPOINTS;
    ULONG   defaultMaxComSegSz      = UNET_DEFAULT_MAX_COMSEG_SIZE;
    ULONG   defaultMaxTotalComSegSz = UNET_DEFAULT_MAX_TOTAL_COMSEG_SIZE;
    ULONG   defaultDefaultComSegSz  = UNET_DEFAULT_COMSEG_SIZE;
    ULONG   defaultDefaultTxFifoSz  = UNET_DEFAULT_TX_FIFO_SIZE;
    ULONG   defaultDefaultRxFifoSz  = UNET_DEFAULT_RX_FIFO_SIZE;
    ULONG   defaultDefaultFreeFifoSz= UNET_DEFAULT_FREE_FIFO_SIZE;
    ULONG   defaultDefaultRxBufferSz= UNET_DEFAULT_RX_BUFFER_SIZE;

    PRINTK("unet: configuring\n");

    paramPath.MaximumLength = RegistryPath->Length + sizeof(subKeyString);
    paramPath.Buffer = ExAllocatePool(PagedPool, paramPath.MaximumLength);

    if (paramPath.Buffer != NULL)
    {
        RtlMoveMemory(paramPath.Buffer, RegistryPath->Buffer, RegistryPath->Length);
        RtlMoveMemory(&paramPath.Buffer[RegistryPath->Length / 2], subKeyString,
                                                                sizeof(subKeyString));
        paramPath.Length = paramPath.MaximumLength - 2;

        RtlZeroMemory(&paramTable[0], sizeof(paramTable));

        paramTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[0].Name = L"MaxNmEndpoints";
        paramTable[0].EntryContext = &UnetObject->maxEndpoints;
        paramTable[0].DefaultType = REG_DWORD;
        paramTable[0].DefaultData = &defaultMaxEndpoints;
        paramTable[0].DefaultLength = sizeof(ULONG);

        paramTable[1].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[1].Name = L"MaxComSegSz";
        paramTable[1].EntryContext = &UnetObject->maxComSegSz;
        paramTable[1].DefaultType = REG_DWORD;
        paramTable[1].DefaultData = &defaultMaxComSegSz;
        paramTable[1].DefaultLength = sizeof(ULONG);

        paramTable[2].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[2].Name = L"MaxTotalComSegSz";
        paramTable[2].EntryContext = &UnetObject->maxTotalComSegSz;
        paramTable[2].DefaultType = REG_DWORD;
        paramTable[2].DefaultData = &defaultMaxTotalComSegSz;
        paramTable[2].DefaultLength = sizeof(ULONG);

        paramTable[3].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[3].Name = L"DefaultComSegSz";
        paramTable[3].EntryContext = &UnetObject->defaultComSegSz;
        paramTable[3].DefaultType = REG_DWORD;
        paramTable[3].DefaultData = &defaultDefaultComSegSz;
        paramTable[3].DefaultLength = sizeof(ULONG);

        paramTable[4].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[4].Name = L"DefaultTxFifoSz";
        paramTable[4].EntryContext = &UnetObject->defaultTxFifoSz;
        paramTable[4].DefaultType = REG_DWORD;
        paramTable[4].DefaultData = &defaultDefaultTxFifoSz;
        paramTable[4].DefaultLength = sizeof(ULONG);

        paramTable[5].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[5].Name = L"DefaultRxFifoSz";
        paramTable[5].EntryContext = &UnetObject->defaultRxFifoSz;
        paramTable[5].DefaultType = REG_DWORD;
        paramTable[5].DefaultData = &defaultDefaultRxFifoSz;
        paramTable[5].DefaultLength = sizeof(ULONG);

        paramTable[6].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[6].Name = L"DefaultFreeFifoSz";
        paramTable[6].EntryContext = &UnetObject->defaultFreeFifoSz;
        paramTable[6].DefaultType = REG_DWORD;
        paramTable[6].DefaultData = &defaultDefaultFreeFifoSz;
        paramTable[6].DefaultLength = sizeof(ULONG);

        paramTable[7].Flags = RTL_QUERY_REGISTRY_DIRECT;
        paramTable[7].Name = L"DefaultRxBufferSz";
        paramTable[7].EntryContext = &UnetObject->defaultRxBufferSz;
        paramTable[7].DefaultType = REG_DWORD;
        paramTable[7].DefaultData = &defaultDefaultRxBufferSz;
        paramTable[7].DefaultLength = sizeof(ULONG);

        status = RtlQueryRegistryValues(
                RTL_REGISTRY_ABSOLUTE | RTL_REGISTRY_OPTIONAL,
                paramPath.Buffer, &paramTable[0], NULL, NULL);
    }

    if ((paramPath.Buffer == NULL) || (!NT_SUCCESS( status ))) {
        PRINTK("unet: QueryRegistry failed, setting defaults\n");
        UnetObject->maxEndpoints      = defaultMaxEndpoints;
        UnetObject->maxComSegSz       = defaultMaxComSegSz;
        UnetObject->maxTotalComSegSz  = defaultMaxTotalComSegSz;
        UnetObject->defaultComSegSz   = defaultDefaultComSegSz;
        UnetObject->defaultTxFifoSz   = defaultDefaultTxFifoSz;
        UnetObject->defaultRxFifoSz   = defaultDefaultRxFifoSz;
        UnetObject->defaultFreeFifoSz = defaultDefaultFreeFifoSz;
        UnetObject->defaultRxBufferSz = defaultDefaultRxBufferSz;
    }

    if (paramPath.Buffer != NULL)
        ExFreePool(paramPath.Buffer);

    PRINTK("unet: config ME 0x%x MC 0x%x MT 0x%x\n",
                            UnetObject->maxEndpoints,
                            UnetObject->maxComSegSz,
                            UnetObject->maxTotalComSegSz);

    PRINTK("unet: config DC 0x%x DT 0x%x DR 0x%x\n",
                            UnetObject->defaultComSegSz,
                            UnetObject->defaultTxFifoSz,
                            UnetObject->defaultRxFifoSz);

    PRINTK("unet: config DF 0x%x DB 0x%x\n",
                            UnetObject->defaultFreeFifoSz,
                            UnetObject->defaultRxBufferSz);

    UnetObject->endpoints = ExAllocatePool(NonPagedPool,
                    UnetObject->maxEndpoints * sizeof(UNET_ENDPOINT));

    if (UnetObject->endpoints == NULL) {
        PRINTK("unet: config failed to allocate endpoint states\n");
        return STATUS_INSUFFICIENT_RESOURCES;
    }
    RtlZeroMemory(UnetObject->endpoints, UnetObject->maxEndpoints * sizeof(UNET_ENDPOINT));
    return STATUS_SUCCESS;
}

static
NTSTATUS
UnetDispatch(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    NTSTATUS status;
    PIO_STACK_LOCATION irpSp;
    PFILE_OBJECT fileObject;

    UNREFERENCED_PARAMETER( DeviceObject );

    irpSp = IoGetCurrentIrpStackLocation( Irp );

    PRINTK("Unet: Dispatch Maj %x Min %x Flags %x Ctl %x\n",
	   irpSp->MajorFunction, irpSp->MinorFunction, irpSp->Flags, irpSp->Control);

    switch (irpSp->MajorFunction) {

    case IRP_MJ_CREATE:
        PRINTK("Unet: Create\n");
        fileObject = irpSp->FileObject;
        // I want to put some per fileobject resources management
        // here if that is usefull
        //
        // fileObject->FsContext = (PVOID) UnetAllocState();
        // if (fileObject->FsContext == NULL)
        //    Irp->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
        // else
            Irp->IoStatus.Status = STATUS_SUCCESS;
        Irp->IoStatus.Information = 0;
        break;

    case IRP_MJ_CLOSE:
    	PRINTK("Unet: Close\n");
        UnetCloseEndpoints(irpSp->FileObject, DeviceObject);
        Irp->IoStatus.Status = STATUS_SUCCESS;
        Irp->IoStatus.Information = 0;
        break;

    case IRP_MJ_READ:
        PRINTK("Unet: Read %d\n",irpSp->Parameters.Read.Length);
        Irp->IoStatus.Status = STATUS_END_OF_FILE;
        Irp->IoStatus.Information = 0;
        break;

    case IRP_MJ_WRITE:
        PRINTK("Unet: Write %d\n",irpSp->Parameters.Write.Length);
        Irp->IoStatus.Status = STATUS_SUCCESS;
        Irp->IoStatus.Information = irpSp->Parameters.Write.Length;
        break;

    case IRP_MJ_DEVICE_CONTROL:
    	PRINTK("Unet: Control %x\n", irpSp->Parameters.DeviceIoControl.IoControlCode);
        (VOID)UnetFastIoDeviceControl(
                            irpSp->FileObject,
                            FALSE,
                            Irp->AssociatedIrp.SystemBuffer,
                            irpSp->Parameters.DeviceIoControl.InputBufferLength,
                            Irp->UserBuffer,
                            irpSp->Parameters.DeviceIoControl.OutputBufferLength,
                            irpSp->Parameters.DeviceIoControl.IoControlCode,
                            &Irp->IoStatus,
                            DeviceObject
                            );
  	break;

    }

    status = Irp->IoStatus.Status;
    IoCompleteRequest( Irp, 0 );
    return status;
}

static
BOOLEAN
UnetFastIoRead(
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    OUT PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN PDEVICE_OBJECT DeviceObject
    )
{
    PRINTK("Unet: FastIoRead\n");
    IoStatus->Status = STATUS_END_OF_FILE;
    IoStatus->Information = 0;
    return TRUE;
}


static
BOOLEAN
UnetFastIoWrite(
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset,
    IN ULONG Length,
    IN BOOLEAN Wait,
    IN ULONG LockKey,
    IN PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN PDEVICE_OBJECT DeviceObject
    )
{
    PRINTK("Unet: FastIoWrite\n");
    IoStatus->Status = STATUS_SUCCESS;
    IoStatus->Information = Length;
    return TRUE;
}

BOOLEAN
UnetFastIoDeviceControl(
    IN PFILE_OBJECT FileObject,
    IN BOOLEAN Wait,
    IN PVOID InputBuffer OPTIONAL,
    IN ULONG InputBufferLength,
    OUT PVOID OutputBuffer OPTIONAL,
    IN ULONG OutputBufferLength,
    IN ULONG IoControlCode,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN PDEVICE_OBJECT DeviceObject
    )
{
    NTSTATUS    status;
    ULONG       nadapters;

    PRINTK("Unet: FastIoDeviceControl %x\n",IoControlCode);

    switch (IoControlCode) {

    case IOCTL_UNET_RESET:
        return 0;
        break;

    case IOCTL_UNET_CREATE_ENDPOINT:
        if (InputBufferLength != sizeof(UNET_CREATE_ENDPOINT_REQUEST)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        if (OutputBufferLength != sizeof(UNET_CREATE_ENDPOINT_RESPONSE)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        status = UnetCreateEndpoint(FileObject,
        		(PUNET_CREATE_ENDPOINT_REQUEST)InputBuffer,
        		(PUNET_CREATE_ENDPOINT_RESPONSE)OutputBuffer, DeviceObject);
        break;

    case IOCTL_UNET_DESTROY_ENDPOINT:
        if (InputBufferLength != sizeof(UNET_DESTROY_ENDPOINT_REQUEST)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        status = UnetDestroyEndpoint(FileObject,
        		(PUNET_DESTROY_ENDPOINT_REQUEST)InputBuffer, DeviceObject);
        break;

    case IOCTL_UNET_ACTIVATE_CHANNEL:
        if (InputBufferLength != sizeof(UNET_ACTIVATE_CHANNEL_REQUEST)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        if (OutputBufferLength != sizeof(UNET_ACTIVATE_CHANNEL_RESPONSE)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        status = UnetActivateChannel(FileObject,
        		(PUNET_ACTIVATE_CHANNEL_REQUEST)InputBuffer,
        		(PUNET_ACTIVATE_CHANNEL_RESPONSE)OutputBuffer, DeviceObject);
        break;

    case IOCTL_UNET_DEACTIVATE_CHANNEL:
        if (InputBufferLength != sizeof(UNET_DEACTIVATE_CHANNEL_REQUEST)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        status = UnetDeactivateChannel(FileObject,
        		(PUNET_DEACTIVATE_CHANNEL_REQUEST)InputBuffer, DeviceObject);
        break;

    case IOCTL_UNET_INFORMATION:
        if (InputBufferLength != sizeof(UNET_ENDPOINT_ID)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        if (OutputBufferLength != sizeof(UNET_INFO_RESPONSE)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        status = UnetInformation(FileObject, (PUNET_ENDPOINT_ID)InputBuffer, (PUNET_INFO_RESPONSE)OutputBuffer, DeviceObject);
        break;

    case IOCTL_UNET_ADAPTERS:
        if (OutputBufferLength != sizeof(ULONG)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        status = UnetNumAdapters(FileObject, (PULONG)OutputBuffer, DeviceObject);
        break;

    case IOCTL_UNET_ADAPTER_LIST:
        status = UnetNumAdapters(FileObject, &nadapters, DeviceObject);
        if (!NT_SUCCESS(status))
            break;
        if (OutputBufferLength != nadapters * sizeof(UNET_INFO_RESPONSE)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        status = UnetAdapterList(FileObject, nadapters, (PUNET_INFO_RESPONSE)OutputBuffer, DeviceObject);
        break;

		case IOCTL_UNET_ADAPTER_IOCTL:
			if (InputBufferLength != sizeof(UNET_ADAPTER_IOCTL_REQUEST)) {
				status = STATUS_BUFFER_TOO_SMALL;
				break;
			}
			if (OutputBufferLength != sizeof(UNET_ADAPTER_IOCTL_RESPONSE)) {
				status = STATUS_BUFFER_TOO_SMALL;
				break;
			}
				
			status = UnetAdapterIoctl(FileObject,
																(PUNET_ADAPTER_IOCTL_REQUEST)InputBuffer,
																(PUNET_ADAPTER_IOCTL_RESPONSE)OutputBuffer,
																DeviceObject);
			break;

	// This is provided only for backwards-compatibility. If 
	// the use of the U-Net fast trap isn't desirable, users can call
	// this ioctl instead.
	case IOCTL_UNET_SERVICE:
        if (InputBufferLength != sizeof(int)) {
            status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
		status = UnetTrapFunction(*(int *)InputBuffer);
		break;

    default:
        IoStatus->Status = STATUS_INVALID_PARAMETER;
        return FALSE;
    }

    IoStatus->Status = status;
    return TRUE;
}

VOID
UnetUnload(
    IN PDRIVER_OBJECT DriverObject
    )
{
    UNICODE_STRING         deviceLinkUnicodeString;

    PRINTK("Unet: Unload\n");

    UnetCloseAllEndpoints(DriverObject->DeviceObject);

#ifndef UNET_USE_SLOW_TRAP
    UTrapDisconnect(UNET_TRAP_VECTOR);
#endif

    RtlInitUnicodeString (&deviceLinkUnicodeString, UNET_DOSDEVICE_NAME);
    IoDeleteSymbolicLink (&deviceLinkUnicodeString);

    IoDeleteDevice( DriverObject->DeviceObject );
}

NTSTATUS
UnetCreateEndpoint(
    IN PFILE_OBJECT FileObject,
    IN PUNET_CREATE_ENDPOINT_REQUEST UnetCreateEndpointRequest,
    OUT PUNET_CREATE_ENDPOINT_RESPONSE UnetCreateEndpointResponse,
    IN PDEVICE_OBJECT DeviceObject
    )
{
    PUNET_OBJECT unetObject;
    PUNET_ADAPTER unetAdapter;
    PUNET_ENDPOINT unetEndpoint;
    NTSTATUS status;

    // get the object associated with the device instance
    unetObject = (PUNET_OBJECT)DeviceObject->DeviceExtension;

    // find the adapter the user wants to use
    unetAdapter = UnetFindAdapterObject(
       UnetCreateEndpointRequest->devType,
       UnetCreateEndpointRequest->devIndex);
    if (unetAdapter == NULL)
        return STATUS_INVALID_PARAMETER;

    // create an endpoint object and fill in the missing fields
    unetEndpoint = UnetCreateEndpointObject(unetObject);
    unetEndpoint->adapter = unetAdapter;
    unetEndpoint->fileObject = FileObject;

    UnetCreateEndpointResponse->endpointId = unetEndpoint->endpointId;
    
    // let the adapter create the endpoint
    ASSERT(unetAdapter->createEndpoint != NULL);
    status = unetAdapter->createEndpoint(unetEndpoint->endpointId, UnetCreateEndpointRequest, UnetCreateEndpointResponse);

    if (!NT_SUCCESS(status))
        UnetDestroyEndpointObject(unetEndpoint);
    return status;
}

NTSTATUS
UnetDestroyEndpoint(
    IN PFILE_OBJECT FileObject,
    IN PUNET_DESTROY_ENDPOINT_REQUEST UnetDestroyEndpointRequest,
    IN PDEVICE_OBJECT DeviceObject
    )
{
    PUNET_OBJECT unetObject;
    PUNET_ADAPTER unetAdapter;
    PUNET_ENDPOINT unetEndpoint;

    // get the object associate with the device instance
    unetObject = (PUNET_OBJECT)DeviceObject->DeviceExtension;
    
    if(FileObject == NULL)
    	return STATUS_INVALID_PARAMETER;

    // search for the endpoint where the channel needs to be activated for
    unetEndpoint = UnetFindEndpointObject(unetObject,
    		UnetDestroyEndpointRequest->endpointId);
    if (unetEndpoint == NULL)
        return STATUS_INVALID_PARAMETER;

    // see if this endpoint is associated with this user
    if (unetEndpoint->fileObject != FileObject)
        return STATUS_INVALID_PARAMETER;

    // notify the adapter first.
    unetAdapter  = unetEndpoint->adapter;
    ASSERT(unetAdapter != NULL);
    ASSERT(unetAdapter->destroyEndpoint != NULL);
    unetAdapter->destroyEndpoint(unetEndpoint->endpointId);

    // release the object
    (VOID)UnetDestroyEndpointObject(unetEndpoint);

    return STATUS_SUCCESS;
}

NTSTATUS
UnetActivateChannel(
    IN PFILE_OBJECT FileObject,
    PUNET_ACTIVATE_CHANNEL_REQUEST UnetActivateChannelRequest,
    PUNET_ACTIVATE_CHANNEL_RESPONSE UnetActivateChannelResponse,
    IN PDEVICE_OBJECT DeviceObject
    )
{
    PUNET_OBJECT unetObject;
    PUNET_ADAPTER unetAdapter;
    PUNET_ENDPOINT unetEndpoint;
    NTSTATUS status;

    // get the object associate with the device instance
    unetObject = (PUNET_OBJECT)DeviceObject->DeviceExtension;

    // search for the endpoint where the channel needs to be activated for
    unetEndpoint = UnetFindEndpointObject(unetObject,
    		UnetActivateChannelRequest->endpointId);
    if (unetEndpoint == NULL)
        return STATUS_INVALID_PARAMETER;

    // see if this endpoint is associated with this user
    if (unetEndpoint->fileObject != FileObject)
        return STATUS_INVALID_PARAMETER;

    // notify the adapter first.
    unetAdapter  = unetEndpoint->adapter;
    ASSERT(unetAdapter->activateChannel != NULL);
    status = unetAdapter->activateChannel(UnetActivateChannelRequest,
    		UnetActivateChannelResponse);

    return status;
}

NTSTATUS
UnetDeactivateChannel(
    IN PFILE_OBJECT FileObject,
    PUNET_DEACTIVATE_CHANNEL_REQUEST UnetDeactivateChannelRequest,
    IN PDEVICE_OBJECT DeviceObject
    )
{
    PUNET_OBJECT unetObject;
    PUNET_ADAPTER unetAdapter;
    PUNET_ENDPOINT unetEndpoint;
    NTSTATUS status;

    // get the object associate with the device instance
    unetObject = (PUNET_OBJECT)DeviceObject->DeviceExtension;

    // search for the endpoint where the channel needs to be deactivated for
    unetEndpoint = UnetFindEndpointObject(unetObject,
    		UnetDeactivateChannelRequest->endpointId);
    if (unetEndpoint == NULL)
        return STATUS_INVALID_PARAMETER;

    // see if this endpoint is associated with this user
    if (unetEndpoint->fileObject != FileObject)
        return STATUS_INVALID_PARAMETER;

    // notify the adapter first.
    unetAdapter  = unetEndpoint->adapter;
    ASSERT(unetAdapter->deactivateChannel != NULL);
    status = unetAdapter->deactivateChannel(UnetDeactivateChannelRequest);

    return status;
}

NTSTATUS
UnetInformation(
    IN PFILE_OBJECT FileObject,
    IN PUNET_ENDPOINT_ID UnetEndpointId,
    OUT PUNET_INFO_RESPONSE UnetInfoResponse,
    IN PDEVICE_OBJECT DeviceObject
    )
{
    PUNET_OBJECT unetObject;
    PUNET_ADAPTER unetAdapter;
    PUNET_ENDPOINT unetEndpoint;
    NTSTATUS status;

    // get the object associate with the device instance
    unetObject = (PUNET_OBJECT)DeviceObject->DeviceExtension;

    // search for the endpoint where the channel needs to be deactivated for
    unetEndpoint = UnetFindEndpointObject(unetObject, *UnetEndpointId);
    if (unetEndpoint == NULL)
        return STATUS_INVALID_PARAMETER;

    // see if this endpoint is associated with this user
    if (unetEndpoint->fileObject != FileObject)
        return STATUS_INVALID_PARAMETER;

    // pass on to the adapter
    unetAdapter  = unetEndpoint->adapter;
    ASSERT(unetAdapter->adapterInfo != NULL);
    if(UnetInfoResponse != NULL) {

			UnetInfoResponse->dev_type  = unetAdapter->devType;
			UnetInfoResponse->dev_index = unetAdapter->devIndex;

			status = unetAdapter->adapterInfo(UnetInfoResponse);
		} else 
			status=STATUS_INVALID_PARAMETER;
		return status;
}

PUNET_ENDPOINT
UnetCreateEndpointObject(
		IN PUNET_OBJECT UnetObject
												)
{
	ULONG i;
	PUNET_ENDPOINT unetEndpoint;

	for (i=0, unetEndpoint = UnetObject->endpoints;
			i < UnetObject->maxEndpoints; i++, unetEndpoint++)
		if (unetEndpoint->state == UNET_ENDPOINT_IDLE) {
			unetEndpoint->state = UNET_ENDPOINT_INUSE;
			unetEndpoint->endpointId = i;
			return &UnetObject->endpoints[i];
		}

	return NULL;
}

NTSTATUS
UnetDestroyEndpointObject(
		PUNET_ENDPOINT  UnetEndpoint
												 )
{
	PUNET_ADAPTER   unetAdapter;

	unetAdapter = UnetEndpoint->adapter;

	// clear the endpoint, marking it free for use
	RtlZeroMemory(UnetEndpoint, sizeof(UNET_ENDPOINT));

	return STATUS_SUCCESS;
}

PUNET_ENDPOINT
UnetFindEndpointObject(
		PUNET_OBJECT UnetObject,
		UNET_ENDPOINT_ID UnetEndpointId
											)
{
	if (UnetEndpointId > UnetObject->maxEndpoints)
		return NULL;

	return &UnetObject->endpoints[UnetEndpointId];
}

PUNET_ADAPTER
UnetCreateAdapterObject(
		UNET_DEV_TYPE UnetDevType,			
		UNET_DEV_INDEX UnetDevIndex
											 )
{
	PUNET_ADAPTER unetAdapter;

	unetAdapter = ExAllocatePool(NonPagedPool, sizeof(UNET_ADAPTER));
	if (unetAdapter == NULL)
		return NULL;

	RtlZeroMemory(unetAdapter, sizeof(UNET_ADAPTER));
	unetAdapter->devIndex    = UnetDevIndex;
	unetAdapter->devType     = UnetDevType;
	unetAdapter->nextAdapter = UnetAdapters;
	UnetAdapters = unetAdapter;
	return unetAdapter;
} 

NTSTATUS
UnetDestroyAdapterObject(
		UNET_DEV_TYPE UnetDevType,
		UNET_DEV_INDEX UnetDevIndex
												)
{
	PUNET_ADAPTER UnetAdapter = UnetFindAdapterObject(UnetDevType, UnetDevIndex);
	if (UnetAdapter == UnetAdapters) {
		UnetAdapters = UnetAdapters->nextAdapter;
	} else {
		PUNET_ADAPTER  adapter, prev;
		for (prev = UnetAdapters, adapter = UnetAdapters->nextAdapter;
				adapter != UnetAdapter;
				prev = adapter, adapter = adapter->nextAdapter) {
			if (adapter->nextAdapter == NULL)
				return STATUS_INVALID_PARAMETER;
		}
		prev->nextAdapter = adapter->nextAdapter;
	}
	ExFreePool(UnetAdapter);
	return STATUS_SUCCESS;
}

PUNET_ADAPTER
UnetFindAdapterObject(
		UNET_DEV_TYPE UnetDevType,
		UNET_DEV_INDEX UnetDevIndex
										 )
{
	PUNET_ADAPTER unetAdapter;

	unetAdapter = UnetAdapters;

	while (unetAdapter != NULL) {
		if ((unetAdapter->devType == UnetDevType) &&
				(unetAdapter->devIndex == UnetDevIndex))
			return unetAdapter;
		unetAdapter = unetAdapter->nextAdapter;
	}

	return NULL;
}

NTSTATUS
UnetNumAdapters(
		IN PFILE_OBJECT FileObject,
		OUT PULONG UnetNmAdapters,
		IN PDEVICE_OBJECT DeviceObject
							 )
{
	ULONG count;
	PUNET_ADAPTER adapter;

	PRINTK("Unet: adapters\n");
	count   = 0;
	adapter = UnetAdapters;

	while (adapter != NULL) {
		adapter = adapter->nextAdapter;
		++count;
	}

	RtlStoreUlong(UnetNmAdapters, count);

	PRINTK("Unet: adapters %d\n", count);

	return STATUS_SUCCESS;
}

NTSTATUS
UnetAdapterList(
		IN PFILE_OBJECT FileObject,
		IN ULONG Nadapters,
		OUT PUNET_INFO_RESPONSE UnetInfoResponseList,
		IN PDEVICE_OBJECT DeviceObject
							 )
{
	ULONG i;
	PUNET_ADAPTER adapter;
	NTSTATUS status;

	PRINTK("Unet: adapter list\n");
	// try{
		for (i=0, adapter = UnetAdapters; adapter != NULL && i < Nadapters;
				i++, adapter = adapter->nextAdapter) {
			PRINTK("Unet: info 0x%x:0x%x\n",
					adapter->devType, adapter->devIndex);
			ASSERT(adapter->adapterInfo != NULL);
			UnetInfoResponseList[i].dev_type = adapter->devType;
			UnetInfoResponseList[i].dev_index = adapter->devIndex;
			status = adapter->adapterInfo(&UnetInfoResponseList[i]);
		};
	// }     // won't compiled if it's on
	// except(EXCEPTION_EXECUTE_HANDLER){
	// PRINTK("An exception occurred in UnetAdapterList()\n");
	// };
	return status;
}

NTSTATUS
UnetAdapterIoctl(
		IN PFILE_OBJECT FileObject,
		PUNET_ADAPTER_IOCTL_REQUEST UnetAdapterIoctlRequest,
		PUNET_ADAPTER_IOCTL_RESPONSE UnetAdapterIoctlResponse,
		IN PDEVICE_OBJECT DeviceObject
								)
{

	PUNET_OBJECT unetObject;
	PUNET_ADAPTER unetAdapter;
	PUNET_ENDPOINT unetEndpoint;
	NTSTATUS status;

	// get the object associate with the device instance
	unetObject = (PUNET_OBJECT)DeviceObject->DeviceExtension;

	// search for the endpoint where the channel needs to be deactivated for
	unetEndpoint = UnetFindEndpointObject(unetObject,
			UnetAdapterIoctlRequest->endpointId);
	if (unetEndpoint == NULL)
		return STATUS_INVALID_PARAMETER;

	// see if this endpoint is associated with this user
	if (unetEndpoint->fileObject != FileObject)
		return STATUS_INVALID_PARAMETER;

	// pass on to the adapter
	unetAdapter  = unetEndpoint->adapter;
	if (unetAdapter->adapterIoctl == NULL) return STATUS_INVALID_PARAMETER;
	status = unetAdapter->adapterIoctl(UnetAdapterIoctlRequest,
			UnetAdapterIoctlResponse);

	return status;
}

NTSTATUS
UnetCloseEndpoints(
		IN PFILE_OBJECT FileObject,
		IN PDEVICE_OBJECT DeviceObject
									)
{
	ULONG i;
	PUNET_OBJECT unetObject;
	PUNET_ENDPOINT unetEndpoint;
	UNET_DESTROY_ENDPOINT_REQUEST unetDestroyEndpointRequest;

	// get the object associated with the device instance
	unetObject = (PUNET_OBJECT)DeviceObject->DeviceExtension;

	for (i=0, unetEndpoint = unetObject->endpoints;
			i < unetObject->maxEndpoints; i++, unetEndpoint++) {
		if (unetEndpoint->state == UNET_ENDPOINT_INUSE &&
				unetEndpoint->fileObject == FileObject) {
			unetDestroyEndpointRequest.endpointId = unetEndpoint->endpointId;
			UnetDestroyEndpoint(FileObject, &unetDestroyEndpointRequest, DeviceObject);
		}
	}
	return STATUS_SUCCESS;
}

NTSTATUS
UnetCloseAllEndpoints(
		IN PDEVICE_OBJECT DeviceObject
										 )
{
	ULONG i;
	PUNET_OBJECT unetObject;
	PUNET_ENDPOINT unetEndpoint;
	UNET_DESTROY_ENDPOINT_REQUEST unetDestroyEndpointRequest;
	FILE_OBJECT fileObject;

	// get the object associated with the device instance
	unetObject = (PUNET_OBJECT)DeviceObject->DeviceExtension;

	for (i=0, unetEndpoint = unetObject->endpoints;
			i < unetObject->maxEndpoints; i++, unetEndpoint++) {
		if (unetEndpoint->state == UNET_ENDPOINT_INUSE) {
			unetDestroyEndpointRequest.endpointId = unetEndpoint->endpointId;
			UnetDestroyEndpoint(NULL, &unetDestroyEndpointRequest, DeviceObject);
		}
	}
	return STATUS_SUCCESS;
}

NTSTATUS
UnetTrapFunction(
		IN ULONG TrapArgument
								)
{
	PUNET_ENDPOINT unetEndpoint;

	PRINTK("unet: trapfunction %d\n",TrapArgument);

	unetEndpoint = UnetFindEndpointObject(TrapObject, TrapArgument);
	if (unetEndpoint == NULL)
		return STATUS_INVALID_PARAMETER;

	ASSERT(unetEndpoint->adapter != NULL);
	ASSERT(unetEndpoint->adapter->transmit != NULL);
	return unetEndpoint->adapter->transmit(TrapArgument);
}
