/*
 * Copyright (c) 1996 by Cornell University.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL CORNELL UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF CORNELL UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * CORNELL UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND CORNELL UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * This version modified for U-Net/MM enhancements, M. Welsh, 16 Oct 96
 */

#ifndef _UNET_USER_H
#define _UNET_USER_H

#define UNET_USE_SLOW_TRAP

/*
 * Transmit, receive, free buffer descriptors
 */
 
// Transmit buffer descriptor (used in Tx Fifo)
typedef struct {
	short	chan;	// Channel number, or UNET_CHANNEL_* code below
	USHORT length; // Length of buffer in bytes
	ULONG buf; // User address of buffer
} UNET_TX_BUFFER_DESCRIPTOR, *PUNET_TX_BUFFER_DESCRIPTOR;

#define UNET_CHANNEL_FREE	(-1)	// Fifo entry is free
#define UNET_CHANNEL_BUSY	(-2)	// Fifo entry is busy
#define UNET_CHANNEL_ERROR	(-3)	// Fifo entry caused error

#define TX_ISFREE(chan) ((chan) == UNET_CHANNEL_FREE)
#define TX_ISBUSY(chan) ((chan) == UNET_CHANNEL_BUSY)
#define TX_ISERROR(chan) ((chan) == UNET_CHANNEL_ERROR)

// low order bit of buf indicates whether next buffer belongs to same packet
#define TX_CONTINUE(addr) ((addr) & 0x01)
#define TX_CONTINUEMASK(addr) ((addr) & ~0x01)

// Free buffer descriptor (used in Free Fifo)
typedef struct {
  ULONG buf;	// User address of buffer
} UNET_RECV_BUFFER_DESCRIPTOR, *PUNET_RECV_BUFFER_DESCRIPTOR;

#define UNET_RXB_FREE (-1)
#define RXB_ISFREE(buf) ((buf) == UNET_RXB_FREE)

// Receive buffer descriptor (used in Rx Fifo)

// size of each descriptor: 64-bytes which can hold 56 bytes of payload
#define RXD_PAYLOAD_ENTRIES 14
#define RXD_PAYLOAD_SIZE (sizeof(unsigned int) * RXD_PAYLOAD_ENTRIES)

typedef struct {
	short	chan;
	USHORT	length;
	ULONG	pad; 
	ULONG	pay_or_buf[RXD_PAYLOAD_ENTRIES];
} UNET_RECEIVE_DESCRIPTOR, *PUNET_RECEIVE_DESCRIPTOR;

#define RX_ISFREE(chan) ((chan) == UNET_CHANNEL_FREE)
#define RX_ISERROR(chan) ((chan) == UNET_CHANNEL_ERROR)

/*
 * Creating and destroying endpoints */

typedef ULONG   UNET_ENDPOINT_ID, *PUNET_ENDPOINT_ID;
typedef ULONG   UNET_DEV_TYPE;
typedef ULONG   UNET_DEV_INDEX;

typedef struct {
  UNET_DEV_TYPE     devType;
  UNET_DEV_INDEX    devIndex;
  ULONG             numTxFifoEntries;
  ULONG             numFreeFifoEntries;
  ULONG             numRxFifoEntries;
  ULONG             bufferAreaSize;
  ULONG             rxBufferSize;
} UNET_CREATE_ENDPOINT_REQUEST, *PUNET_CREATE_ENDPOINT_REQUEST;

typedef struct {
	UNET_ENDPOINT_ID			endpointId;
	PUNET_TX_BUFFER_DESCRIPTOR	txFifo;
	ULONG						*txWall;
	PUNET_RECV_BUFFER_DESCRIPTOR freeFifo;
	ULONG						*freeWall;
	PUNET_RECEIVE_DESCRIPTOR	rxFifo;
	ULONG						*rxWall;
	void						*bufferArea;
	ULONG           rxBufferAllocSize; // Bytes of Rx buffer to allocate
	ULONG           rxBufferUsedSize;  // Bytes of Rx buffer used by incoming packet

} UNET_CREATE_ENDPOINT_RESPONSE, *PUNET_CREATE_ENDPOINT_RESPONSE;

typedef struct {
	UNET_ENDPOINT_ID			endpointId;
} UNET_DESTROY_ENDPOINT_REQUEST, *PUNET_DESTROY_ENDPOINT_REQUEST;

/*
 * Activating and deactivating channels
 */

typedef short UNET_CHANNEL_ID;

typedef struct {
	UCHAR macAddr[6];
	UCHAR port;
} UNET_FE_ADDRESS, *PUNET_FE_ADDRESS;

typedef struct {
	ULONG vci;
} UNET_ATM_ADDRESS, *PUNET_ATM_ADDRESS;

typedef union {
	UNET_ATM_ADDRESS atm;
	UNET_FE_ADDRESS  fe;
} UNET_CHANNEL_ADDRESS, *PUNET_CHANNEL_ADDRESS;

typedef struct {
	UNET_ENDPOINT_ID		endpointId;
	UNET_CHANNEL_ID		chan;	// requested channel, or UNET_CHAN_ANY
	UNET_CHANNEL_ADDRESS	txAddr;	// Tx destination address
	UNET_CHANNEL_ADDRESS	rxAddr;	// Rx source address
} UNET_ACTIVATE_CHANNEL_REQUEST, *PUNET_ACTIVATE_CHANNEL_REQUEST;

#define UNET_CHAN_ANY (-1)

typedef struct {
	UNET_CHANNEL_ID			chan;
} UNET_ACTIVATE_CHANNEL_RESPONSE, *PUNET_ACTIVATE_CHANNEL_RESPONSE;

typedef struct  _UNET_DEACTIVATE_CHANNEL_REQUEST {
	UNET_ENDPOINT_ID	endpointId;
	UNET_CHANNEL_ID		chan;
} UNET_DEACTIVATE_CHANNEL_REQUEST, *PUNET_DEACTIVATE_CHANNEL_REQUEST;

typedef struct _UNET_ADAPTER_IOCTL_REQUEST {
	UNET_ENDPOINT_ID endpointId;
	ULONG code;
	ULONG bufferSize;
	PVOID buffer;
} UNET_ADAPTER_IOCTL_REQUEST, *PUNET_ADAPTER_IOCTL_REQUEST;

typedef struct _UNET_ADAPTER_IOCTL_RESPONSE {
	ULONG bufferSize;
	PVOID buffer;
} UNET_ADAPTER_IOCTL_RESPONSE, *PUNET_ADAPTER_IOCTL_RESPONSE;
	

/*
 * Information
 */

typedef struct {
  UNET_DEV_TYPE dev_type;		// device type ID tag (below)
  UNET_DEV_INDEX dev_index;              // index tag referring to device
  ULONG version;		// version number of the device
  
  ULONG rxd_payload_size;	// bytes of message payload in Unet_rxd
  ULONG max_pdu;			// maximum PDU supported by device
  ULONG max_endpts;		// max # of endpoints supported by device
  ULONG max_chans; 		// max # of chans supported per endpoint
  
  ULONG tx_buf_align;		// byte alignment for Tx buffers (1=8bit, 4=32bit)
  ULONG rx_buf_align; 	// byte alignment for Rx buffers (1=8bit, 4=32bit)
  
  ULONG min_tx_fifo;		// min and Max # entries for each FIFO
  ULONG max_tx_fifo;
  ULONG min_free_fifo;
  ULONG max_free_fifo;
  ULONG min_rx_fifo;
  ULONG max_rx_fifo;
	UNET_CHANNEL_ADDRESS local_addr;
} UNET_INFO_RESPONSE, *PUNET_INFO_RESPONSE;

#define UNET_DEVID_SBA200        0x01
#define UNET_DEVID_PCA200        0x02
#define UNET_DEVID_TULIP         0x03
#define UNET_DEVID_TEST          0x04
#define UNET_DEVID_ZEITNET       0x05
#define UNET_DEVID_NULL          0x06
#define UNET_DEVID_TULIP_NULL    0x07

#define UNET_DEV_VERSION  0x06

/* U-Net trap vector; on x86, place argument into %eax and issue software
 * interrupt at this index.
 */
#define UNET_TRAP_VECTOR 0x29

#ifdef UNET_USE_SLOW_TRAP

#define UNET_TRAP(hDriver, endpt) do {                            \
  int _nbytes, eId = (endpt);                                     \
  (void)DeviceIoControl((hDriver), (DWORD)IOCTL_UNET_SERVICE,     \
                        &(eId), sizeof(UNET_ENDPOINT_ID),         \
                        NULL, 0, &_nbytes, 0);                    \
} while(0);

#else 

/* Macro to call trap with given endpoint index as argument */
/* (hDriver argument unused, retained for compatibility with slow trap) */
#define UNET_TRAP(hDriver, endpt) { \
	int _eid = (endpt);      \
	__asm mov eax, _eid \
	__asm int UNET_TRAP_VECTOR \
}

#endif /* UNET_USE_SLOW_TRAP */


/*
 * NT I/O controls
 */

#define FILE_DEVICE_UNET  0x00008000
#define UNET_IOCTL_INDEX  0x800

#define IOCTL_UNET_RESET              CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX,   METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_CREATE_ENDPOINT    CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+1, METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_DESTROY_ENDPOINT   CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+2, METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_ACTIVATE_CHANNEL   CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+3, METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_DEACTIVATE_CHANNEL CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+4, METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_DEBUG              CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+5, METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_INFORMATION        CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+6, METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_ADAPTERS           CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+7, METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_ADAPTER_LIST       CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+8, METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_SERVICE            CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+9, METHOD_BUFFERED,FILE_ANY_ACCESS)
#define IOCTL_UNET_ADAPTER_IOCTL      CTL_CODE(FILE_DEVICE_UNET, UNET_IOCTL_INDEX+10, METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif

