/* utlb.c, M. Welsh (mdw@cs.cornell.edu)
 *
 * Stand-alone kernel module providing the U-Net/MM Translation Lookaside
 * Buffer mechanism, to be shared by all U-Net services.
 */

#include <ntddk.h>
#include "unetuser.h"
#include "unetkernel.h"
#include "utlb.h"

/* Define for verbose debug logging */
#undef UTLB_DEBUG

/* Define for raw operation benchmarking */
#undef UTLB_BENCHMARK

#ifdef UTLB_DEBUG
#define PRINTK DbgPrint
#else
#define PRINTK 0 &&
#endif

static UNET_STLB stlb[UNET_STLB_SIZE];
static PUNET_STLB_VCACHE victim_cache, victim_cache_last;
static int victim_cache_size;
static BOOLEAN victim_cache_cleanup_lock;

#ifdef UTLB_BENCHMARK
static __int64 hit_time, tx_miss_time, tx_unlock_time;
static __int64 rx_miss_time, rx_unlock_time;
static int num_hits, num_tx_misses, num_tx_unlocks;
static int num_rx_misses, num_rx_unlocks;
#endif

NTSTATUS
UTlbUnload (
    IN PDRIVER_OBJECT DriverObject
    );
    

NTSTATUS  
DriverEntry ( 
    IN PDRIVER_OBJECT DriverObject, 
    IN PUNICODE_STRING RegistryPath 
    )
{
	int i;
	
	DbgPrint("UTlb: Module loaded.\n");
	DriverObject->DriverUnload = UTlbUnload;

	for (i = 0; i < UNET_STLB_SIZE; i++) {
		stlb[i].page_id = 0;
		stlb[i].paddr = 0;
		stlb[i].saddr = 0;
		stlb[i].age = 0;
		stlb[i].refcnt = 0;
	}
	victim_cache = victim_cache_last = NULL;
	victim_cache_size = 0;
	victim_cache_cleanup_lock = FALSE;

#ifdef UTLB_BENCHMARK
	hit_time = tx_miss_time = tx_unlock_time = rx_miss_time = rx_unlock_time = 0;
	num_hits = num_tx_misses = num_tx_unlocks = num_rx_misses = num_rx_unlocks = 0;
#endif
		
	return STATUS_SUCCESS;
}

NTSTATUS
UTlbUnload (
    IN PDRIVER_OBJECT DriverObject
    )
{
	DbgPrint("UTlb: Module unloaded.\n");
	// XXX Do TLB cleanup here?

#ifdef UTLB_BENCHMARK
	DbgPrint("uTLB: hit time %ld ns (%d), Tx miss %ld ns (%d), Rx miss %ld ns (%d), Tx unlock %ld ns (%d), Rx unlock %ld ns (%d)\n",
					 get_nseconds(hit_time) / num_hits,
					 num_hits,
					 get_nseconds(tx_miss_time) / num_tx_misses,
					 num_tx_misses,
					 get_nseconds(rx_miss_time) / num_rx_misses,
					 num_rx_misses,
					 get_nseconds(tx_unlock_time) / num_tx_unlocks,
					 num_tx_unlocks,
					 get_nseconds(rx_unlock_time) / num_rx_unlocks,
					 num_rx_unlocks);
#endif
		
	return STATUS_SUCCESS;
}


// Do cleanup of victim cache if too large
// Assume that this is entered and left at DIRQL, although returns
// to origIrqL (which must be >= PASSIVE_LEVEL) to unlock pages

// XXX Should I use mutexes instead of interrupt-levels?
static void VictimCacheCleanup(KIRQL origIrqL) {
	PUNET_STLB_VCACHE node, prev, next;
	KIRQL oldIrqL = origIrqL;

	// Is someone else already cleaning up the list?
	// Note: This is atomic as we assume DIRQL here
	if (victim_cache_cleanup_lock) return;
	victim_cache_cleanup_lock = TRUE;
	
	if (victim_cache_size+1 > UNET_STLB_VCACHE_SIZE) {
		for (prev = NULL, node = victim_cache;
				 node && (victim_cache_size+1 > UNET_STLB_VCACHE_SIZE);) {
			if (node->entry.refcnt) goto skip_entry;
			if (!node->entry.paddr) goto skip_entry;
			
			PRINTK("UTlb: victim killing 0x%x:0x%lx -> 0x%lx (prot %d)\n",
						 PAGE_OFFSET(node->entry.page_id),
						 PAGE_ADDR(node->entry.page_id),
						 PAGE_ADDR(node->entry.paddr), PAGE_OFFSET(node->entry.paddr));
			
			if (node->entry.paddr & UNET_STLB_PROT_RW) {
				// Page used as Rx buffers. Walk free FIFO and push on bogus entries
				// so user can re-use the buffers
				// XXX mdw STLB fixup: How can we do this ??
				// Maybe -- push old entries onto a list, U-Net device can
				// pick them up later. Maybe only call VictimCacheCleanup in
				// an async thread, also ...
				PRINTK("XXX UTlb: Should walk free FIFO here\n");
			}
			
			if (!prev) {
				victim_cache = node->next;
			} else {
				prev->next = node->next;
				prev = node;
			}
			if (!node->next) victim_cache_last = node;
			next = node->next;
			victim_cache_size--;

			KeLowerIrql(oldIrqL);
			if (PAGE_ADDR(node->entry.paddr)) MmUnlockPages(&(node->entry.u.mdl));
			KeRaiseIrql(POWER_LEVEL, &oldIrqL);
			ExFreePool(node);
			node = next;
			continue;

	 skip_entry:
			if (!node->next) victim_cache_last = node;
			node = node->next;
			
		}
	}
	victim_cache_cleanup_lock = FALSE;
}

// Push new entry on end of victim cache list (which must be pre-allocated).
// Assumed to be atomic
static void VictimCachePush(PUNET_STLB entry) {
	memcpy(&(victim_cache_last->entry), entry, sizeof(UNET_STLB));
	victim_cache_size++;
}

// Allocate empty node on end of victim cache. This is separate from the
// push (which must be atomic) as this must be done at lowered IrqL.
static void VictimCacheAlloc(void) {
	PUNET_STLB_VCACHE node;
	KIRQL oldIrqL;
	node = ExAllocatePool(NonPagedPool, sizeof(UNET_STLB_VCACHE));
	if (!node) {
		DbgPrint("UTlb: Can't allocate new node for victim cache!\n");
		return;
	}
	node->next = NULL;
	node->entry.paddr = node->entry.page_id = 0;
	
	// Following must be atomic
	KeRaiseIrql(POWER_LEVEL, &oldIrqL);
	if (!victim_cache_last) {
		// First call
		victim_cache = victim_cache_last = node;
	} else {
		victim_cache_last->next = node;
		victim_cache_last = node;
	}
	KeLowerIrql(oldIrqL);
}

// Walk victim cache and return entry if present, assumed to be atomic
static PUNET_STLB_VCACHE VictimCacheFind(ULONG page_id) {
	PUNET_STLB_VCACHE node;
	for (node = victim_cache; node; node = node->next) {
		if (node->entry.page_id == page_id) break;
	}
	return node;
}

/* Translate vaddr to paddr using TLB, possibly sleep.
 * Assumes TLB entry refcnt to be incremented if prot == UNET_STLB_PROT_RO
 * (which must be done atomically).
 */
ULONG StlbTranslate(UNET_ENDPOINT_ID eid,
										ULONG vaddr, int prot, BOOLEAN handle_miss,
										PULONG system_addr) {
	PUNET_STLB entry;
	UNET_STLB tmp;
	int index;
  ULONG pAddr = 0xffffffff, sAddr;
	PUNET_STLB_VCACHE vcache, vcache_tmp;
	BOOLEAN pushed_vcache = FALSE;
	BOOLEAN relock = FALSE;
	KIRQL oldIrqL;

#ifdef UTLB_BENCHMARK
	__int64 start_time;
	start_time = get_pentium_cycle();
#endif

	KeRaiseIrql(POWER_LEVEL, &oldIrqL);
	
	index = UNET_STLB_HASH(eid, vaddr);
	entry = stlb + index;

	PRINTK("UTlb: Lookup 0x%x:0x%lx (prot %d)\n", eid, vaddr, prot);

	if ((entry->page_id == (vaddr|eid)) && (entry->paddr)) {
		PRINTK("UTlb: got valid tlb entry\n");
		// Got valid entry, right perms?
		if ((entry->paddr & UNET_STLB_PROT_RW) || (prot == UNET_STLB_PROT_RO)) {
			PRINTK("UTlb: Hit, 0x%x:0x%lx -> 0x%lx (prot %d)\n",
						 eid,vaddr,PAGE_ADDR(entry->paddr),PAGE_OFFSET(entry->paddr));
			if (prot == UNET_STLB_PROT_RO) entry->refcnt++;
			// Necessary as we can't look at entry after KeLowerIrql.
			sAddr = entry->saddr;
			pAddr = PAGE_ADDR(entry->paddr);
			KeLowerIrql(oldIrqL);
#ifdef UTLB_BENCHMARK
			hit_time += (get_pentium_cycle() - start_time); num_hits++;
#endif
			if (system_addr) *system_addr = sAddr;
			return pAddr;
		}
	}

	vcache = VictimCacheFind((vaddr|eid));
	if (vcache) {
		// Right entry, wrong perms?
		if ((vcache->entry.paddr & UNET_STLB_PROT_RW) || (prot == UNET_STLB_PROT_RO)) {
			PRINTK("UTlb: Vcache hit, 0x%x:0x%lx -> 0x%lx (prot %d)\n",
						 eid,vaddr, PAGE_ADDR(vcache->entry.paddr),
						 PAGE_OFFSET(vcache->entry.paddr));
			if (prot == UNET_STLB_PROT_RO) vcache->entry.refcnt++;
			sAddr = vcache->entry.saddr;
			pAddr = PAGE_ADDR(vcache->entry.paddr);
			KeLowerIrql(oldIrqL);
			if (system_addr) *system_addr = sAddr;
			return pAddr;
		}
	}

	if (!handle_miss) {
		KeLowerIrql(oldIrqL);
		if (system_addr) *system_addr = 0UL;
		return 0UL;
	}

	// OK, so this was a miss (or a relock, but relocks are counted as
	// misses). We need to _atomically_ push the old TLB entry to the
	// victim cache, then lower IrqL and lock the pages.

	// Possible that there's no victim cache yet (i.e. first call,
	// in which case this is an empty TLB entry, so it's ok).
	
	if (victim_cache_last && entry->page_id) {
		VictimCachePush(entry);
		pushed_vcache = TRUE;
	}

	// Now steal the TLB entry (atomically)
	entry->page_id = (vaddr|eid);
	entry->age = 0;
	entry->paddr = entry->saddr = 0;
	entry->refcnt = 0;

	// Can't use entry->mdl as MmProbeAndLock can sleep, during which entry
	// could be revoked
	memset(&(tmp.u), 0, sizeof(tmp.u));
	MmInitializeMdl(&(tmp.u.mdl), vaddr, PAGE_SIZE);
	
	// Shrink victim cache if need be (can sleep)
	// XXX mdw TLB fixup: Maybe a separate thread can do this ...
	if (victim_cache_last && pushed_vcache) {
		VictimCacheCleanup(oldIrqL);
	}
	
	KeLowerIrql(oldIrqL);

	// XXX mdw STLB fixup
	// Whoops. What if after we've done KeLowerIrqL, someone else comes along
	// and pushes a Tlb entry out before we've had a chance to allocate the
	// new entry???
	if (pushed_vcache)
		VictimCacheAlloc();

	try {
		MmProbeAndLockPages(&(tmp.u.mdl), KernelMode, (prot == UNET_STLB_PROT_RO)?IoReadAccess:IoModifyAccess);
	} except (EXCEPTION_EXECUTE_HANDLER) {
		PRINTK("UTlb: Exception in StlbTranslate 0x%x:0x%lx (prot %d)\n",
					 eid, vaddr, prot);
		pAddr = 0;
	}

	if (pAddr != 0) {
		pAddr = (*(ULONG *)((&(tmp.u.mdl))+1)) << PAGE_SHIFT;
		sAddr = MmGetSystemAddressForMdl(&(tmp.u.mdl));
	}
	
	KeRaiseIrql(POWER_LEVEL, &oldIrqL);

	// OK, now find our TLB entry (it might have been knocked out)
	
	if (entry->page_id != (vaddr|eid)) {
		vcache = VictimCacheFind((vaddr|eid));
		if (!vcache) {
			// Huh? Entry should at least have been pushed into the victim
			// cache.
			DbgPrint("UTlb: Lost TLB/victim cache entry after sleep?\n");
			KeLowerIrql(oldIrqL);
			if (system_addr) *system_addr = 0UL;
			return 0UL;
		}
		entry = &(vcache->entry);
	}

	// All right, we can initialize the entry. But ... it's possible that
	// a mapping for the page is already there, which means that our
	// initialized TLB entry went to the victim cache. So we need to arrange
	// for it to be discarded later.
	if (entry->paddr) {
		// Yep, someone beat us to it. If the permissions are right,
		// use the mapping already there. Otherwise replace with the new one.
		// XXX Assumes that the acquired paddr is the same.
		if ((entry->paddr & UNET_STLB_PROT_RW) || (prot == UNET_STLB_PROT_RO)) {
			PRINTK("UTlb: Stlb lost map race, 0x%x:0x%lx -> 0x%lx (prot %d)\n",
						 eid,vaddr,PAGE_ADDR(entry->paddr),PAGE_OFFSET(entry->paddr));
			if (prot == UNET_STLB_PROT_RO) entry->refcnt++;
			sAddr = entry->saddr;
			pAddr = PAGE_ADDR(entry->paddr);
			KeLowerIrql(oldIrqL);
			if (system_addr) *system_addr = sAddr;
			return pAddr;
		} else {
			// Permissions aren't right: We're an RW mapping replacing an RO.
			// XXX Replace the entry, but we want to knock the RO onto the
			// victim cache so it can be eventually freed.
			// XXX mdw 1-Nov-96: I'm going to leave this in as a deliberate
			// bug. It seems that the i960 firmware for the PCA-200 implementation
			// has the same problem.
			PRINTK("UTlb: Stlb lost map race but RW, 0x%x:0x%lx -> 0x%lx (prot %d)\n",
						 eid, vaddr, pAddr, prot);
			entry->paddr = pAddr | prot;
			entry->saddr = sAddr;
			memcpy(&(entry->u), &(tmp.u), sizeof(tmp.u));
			KeLowerIrql(oldIrqL);
			if (system_addr) *system_addr = sAddr;
			return pAddr;
		}
	} else {
		// Nope, we're filling this entry for the first time.
		PRINTK("UTlb: Stlb locked 0x%x:0x%lx -> 0x%lx (prot %d)\n",
					 eid, vaddr, pAddr, prot);
		if (prot == UNET_STLB_PROT_RO) entry->refcnt = 1;
		entry->paddr = pAddr | prot;
		entry->saddr = sAddr;
		memcpy(&(entry->u), &(tmp.u), sizeof(tmp.u));
		KeLowerIrql(oldIrqL);

#ifdef UTLB_BENCHMARK
		if (prot) {
			rx_miss_time += (get_pentium_cycle() - start_time);
			num_rx_misses++;
		} else {
			tx_miss_time += (get_pentium_cycle() - start_time);
			num_tx_misses++;
		}
#endif
		
		if (system_addr) *system_addr = sAddr;
		return pAddr;
	}

	DbgPrint("UTLB: XXXX SHOULD NEVER SEE THIS CASE\n");
	
}

// Walk STLB and victim cache for given endpoint, unlocking all entries.
// Doesn't regard entry refcnt, as this is called when the endpoint shuts
// down (refcnt should be zero anyway).
void StlbClear(UNET_ENDPOINT_ID eid) {
	int i;
	PUNET_STLB entry;
	KIRQL oldIrqL;
	UNET_STLB tmp;
	PUNET_STLB_VCACHE node, prev, next;

#ifdef UTLB_BENCHMARK
	__int64 start_time;
#endif

	PRINTK("UTlb: StlbClear called for endpoint %d\n", eid);
	
	KeRaiseIrql(POWER_LEVEL, &oldIrqL);

	for (i = 0; i < UNET_STLB_SIZE; i++) {
#ifdef UTLB_BENCHMARK
		start_time = get_pentium_cycle();
#endif
		entry = stlb + i;
		if ((PAGE_OFFSET(entry->page_id) == eid) && (entry->paddr)) {
			memcpy(&tmp, entry, sizeof(UNET_STLB));
			entry->page_id = entry->paddr = entry->saddr = entry->age = entry->refcnt = 0;
			memset(&(entry->u), 0, sizeof(entry->u));

			// Cleared the TLB, now can lower IRQL and do the unlock
			KeLowerIrql(oldIrqL);
			PRINTK("UTlb: StlbClear killing 0x%x:0x%lx -> 0x%lx (prot %d)\n",
						 PAGE_OFFSET(tmp.page_id),
						 PAGE_ADDR(tmp.page_id),
						 PAGE_ADDR(tmp.paddr),
						 PAGE_OFFSET(tmp.paddr));
			
			MmUnlockPages(&(tmp.u.mdl));
			MmPrepareMdlForReuse(&(tmp.u.mdl)); // Maybe necessary to remove sys mapping?
			KeRaiseIrql(POWER_LEVEL, &oldIrqL);

#ifdef UTLB_BENCHMARK
			if (PAGE_OFFSET(tmp.paddr)) {
				rx_unlock_time += (get_pentium_cycle() - start_time);
				num_rx_unlocks++;
			} else {
				tx_unlock_time += (get_pentium_cycle() - start_time);
				num_tx_unlocks++;
			}
#endif
			
		}
	}

	// XXX Need to clear victim cache here. Problem: Some other thread
	// might be holding the victim_cache_cleanup_lock. Should implement
	// that as a semaphore or mutex.
	//
	// Note that leaving 'old' entries in the victim cache can be dangerous,
	// not because they might be cleaned up after the process has died, but
	// because other processes might hit the victim cache where the entries
	// don't belong to them.
	
	KeLowerIrql(oldIrqL);
	
}

// Atomically decrement TLB entry reference count
void StlbDecRefcnt(UNET_ENDPOINT_ID eid, ULONG vaddr, ULONG paddr) {
	KIRQL oldIrqL;
	PUNET_STLB entry;
	int index;

	KeRaiseIrql(POWER_LEVEL, &oldIrqL);
	index = UNET_STLB_HASH(eid, vaddr);
	entry = stlb + index;

	if ((entry->page_id == (eid|vaddr)) && (entry->paddr == paddr)) {
		entry->refcnt--;
	} else {
		// Whoops, TLB entry replaced, get it from the victim cache
		PUNET_STLB_VCACHE vcache = VictimCacheFind((eid|vaddr));
		if (!vcache) {
			DbgPrint("UTlb: Lost vcache entry for reaping 0x%x->0x%lx ?\n",
							 vaddr, paddr);
		} else {
			vcache->entry.refcnt--;
		}
	}
	KeLowerIrql(oldIrqL);
}
