/* WinNT tulip.c,
 * Thorsten von Eicken (tve@cs.cornell.edu) and
 * Matt Welsh (mdw@cs.cornell.edu)
 *
 * A driver for the DC21140 'Tulip' 10/100 Mbps PCI Ethernet
 * controller chip. This driver implements U-Net/FE for the Tulip chip.
 *
 * Copyright (c) 1995, 1996 by Cornell University.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL CORNELL UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF CORNELL UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * CORNELL UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND CORNELL UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * M. Welsh, 21 Feb 96, Linux version
 * Thorsten von Eicken, July 96, WinNT4.0 version
 * M. Welsh, 15 Oct 96, WinNT4.0 version with MM enhancements
 */

#include <ntddk.h>
#include "unetuser.h"
#include "unetkernel.h"
#include "utlb.h"

/* Define for verbose debug logging */
#undef TULIP_DEBUG

#ifdef TULIP_DEBUG
#define PRINTK DbgPrint
#else
#define PRINTK 0 &&
#endif

/* Misc defines **************************************************************/

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))

#define PCI_VENDOR_ID_DEC 0x1011
#define PCI_DEVICE_ID_DEC_TULIP_FAST 0x0009

static PHYSICAL_ADDRESS topPhysAddr = { 0xffffffff, 0 }; // end of phys mem

// Pause for 1ms
#define SLOW_DOWN_IO do { LARGE_INTEGER t1, t2; KeQuerySystemTime(&t1); do { KeQuerySystemTime(&t2); } while ((t2.QuadPart - t1.QuadPart) < 1000); } while(0)

//#define OUT_PORT(addr, data) WRITE_PORT_ULONG((unsigned long *)(addr), data)
//#define IN_PORT(addr) READ_PORT_ULONG((unsigned long *)(addr))

/* Tulip parameters **********************************************************/

// Maxs
#define TULIP_MAX_DEVS 2
#define TULIP_MAX_ENDPT UNET_DEFAULT_MAX_ENDPOINTS
#define TULIP_MAX_CHANS 32
#define TULIP_MAX_PDU 1498  /* 1500 - 2 for the length short */
#define TULIP_MAX_PORTS 256

// Fifo size bounds
#define TULIP_MIN_TX 2
#define TULIP_MAX_TX 100
#define TULIP_MIN_FREE 2
#define TULIP_MAX_FREE 500
#define TULIP_MIN_RX 2
#define TULIP_MAX_RX 500

// TLB prefetch constants
#define START_PREFETCH_THRESH 1
#define NUM_PREFETCH 4

// register offsets
//enum TULIP_OFFSETS {
//	CSR0=0, CSR1=0x08, CSR2=0x10, CSR3=0x18, CSR4=0x20, CSR5=0x28,
//	CSR6=0x30, CSR7=0x38, CSR8=0x40, CSR9=0x48, CSR10=0x50, CSR11=0x58,
//	CSR12=0x60, CSR13=0x68, CSR14=0x70, CSR15=0x78
//};

// tulip status and control (CSR) registers
typedef struct {
	ULONG	CSR0,  pad0;		// bus mode
	ULONG	CSR1,  pad1;		// transmit poll demand
	ULONG	CSR2,  pad2;		// receive poll demand
	ULONG	CSR3,  pad3;		// rx ring base address
	ULONG	CSR4,  pad4;		// tx ring base address
	ULONG	CSR5,  pad5;		// status
	ULONG	CSR6,  pad6;		// operation mode
	ULONG	CSR7,  pad7;		// interrupt enable
	ULONG	CSR8,  pad8;		// missed frame counter
	ULONG	CSR9,  pad9;		// serial ROM
	ULONG	CSR10, pad10;		// reserved
	ULONG	CSR11, pad11;		// gen purp timer
	ULONG	CSR12, pad12;		// gen purp port
	ULONG	CSR13, pad13;		// reserved
	ULONG	CSR14, pad14;		// reserved
	ULONG	CSR15, pad15;		// watchdog timer
} TULIP_CSR, *PTULIP_CSR;

// Tulip device buffer ring definitions
#define RX_BUF_SZ 1536
#define TX_RING_SIZE 64
#define RX_RING_SIZE 64

typedef struct {
	unsigned long status;
	unsigned long length;
	unsigned long buffer1, buffer2;
} TULIP_DESC, *PTULIP_DESC;

/* Tulip data structures *****************************************************/


// Tx ring shadow
typedef struct {
	ULONG page_id;
	ULONG paddr;
	BOOLEAN chained;
} TULIP_TXRING_SHADOW;

// Channel
typedef struct {
	int inUse;
	UNET_FE_ADDRESS	txAddr;		// Tx destination address
	UNET_FE_ADDRESS	rxAddr;		// Rx destination address
	unsigned char	header[14];	// Ethernet header
} TULIP_CHAN, *PTULIP_CHAN;

typedef struct {
	char inuse;
	UNET_ENDPOINT_ID endpt;
	UNET_CHANNEL_ID  chan;
} TULIP_PORT, *PTULIP_PORT;

#define EEPROM_SIZE 128 // Max EEPROM size in bytes

// Device
typedef struct {
	// Device addresses and NT resource handles
	int index;                            // device index
	ULONG cbio;		                // I/O base addr for CSRs
	PTULIP_CSR cbma;	                // memory base addr for CSRs
	unsigned char eeprom[EEPROM_SIZE]; // EEPROM data
	unsigned char macAddr[6];		// MAC address from EEPROM
	PUNICODE_STRING registryPath;		// used for releasing resources
	PKINTERRUPT interruptObject;
	PDEVICE_OBJECT deviceObject;

	// The transmit and receive descriptor rings
	char *rxBufs;				// allocated region for rx buffers
	TULIP_DESC rxRing[RX_RING_SIZE];	// receive device ring
	TULIP_DESC txRing[TX_RING_SIZE];	// transmit device ring

	// Space for the per-txRing-entry Ethernet header
	char txRingHeader[TX_RING_SIZE][14+2];
	ULONG txRingHeaderPhys[TX_RING_SIZE]; // physical addresses of txRingHeader

	// Shadow of Tx ring for additional fields (for reap)
	TULIP_TXRING_SHADOW txRingShadow[TX_RING_SIZE];

	int curRx, curTx;			// Current Tx and Rx descriptors
	int dirtyRx, dirtyTx;			// Tx and Rx descriptor to reap
	int txFull;				// Set if Tx ring is full
	//struct wait_queue *tx_full_waitq;	// Used for full Tx ring

	int setupFrame[48]; 			// Address filter pseudo-frame

	/* Mapping from port numbers (for this device) to endpoint/channel. */
	TULIP_PORT portmap[TULIP_MAX_PORTS];

} TULIP_DEV, *PTULIP_DEV;

// Endpoint
typedef struct {
	PTULIP_DEV		dev;
	PEPROCESS     process;

	UNET_TX_BUFFER_DESCRIPTOR		*txFirst;
	volatile UNET_TX_BUFFER_DESCRIPTOR	*txNext;
	UNET_TX_BUFFER_DESCRIPTOR		*txLast;
	ULONG					*txWall;
	int					numTxFifoEntries;

	UNET_RECV_BUFFER_DESCRIPTOR	       *freeFirst;
	volatile UNET_RECV_BUFFER_DESCRIPTOR *freeNext;
	UNET_RECV_BUFFER_DESCRIPTOR	       *freeLast;
	ULONG					*freeWall;
	int					numFreeFifoEntries;

	// For prefetching of free buffers
	volatile UNET_RECV_BUFFER_DESCRIPTOR *freePrefetchNext;
	int freePrefetch;

	UNET_RECEIVE_DESCRIPTOR		*rxFirst;
	volatile UNET_RECEIVE_DESCRIPTOR	*rxNext;
	UNET_RECEIVE_DESCRIPTOR		*rxLast;
	ULONG					*rxWall;
	int					numRxFifoEntries;

	ULONG					fifoAreaSize;
	ULONG					bufferAreaSize;
	void					*bufferArea;
	ULONG					userBufferArea;
	ULONG					userFifoArea;
	ULONG					physBufferArea;

	int					numChans;
	TULIP_CHAN				chans[TULIP_MAX_CHANS];

	// Number of Tx ring entries for this endpoint
	// freed since last tx_wall update
	int					numFreed;
} TULIP_ENDPT, *PTULIP_ENDPT;

/* Global variables ******************************************************/

static PDRIVER_OBJECT tulipDriverObject;
static TULIP_DEV *tulipDev[TULIP_MAX_DEVS];
static int numtulipDev = 0;
static TULIP_ENDPT *tulipEndpt[TULIP_MAX_ENDPT];

static ULONG internalLoopback, promiscuousMode,fullDuplex;

/* Forward function declarations *************************************/

static NTSTATUS UnetUnload(PDRIVER_OBJECT DriverObject);
static void ReleaseTulip(int devIndex);
static NTSTATUS InitDevice(PDRIVER_OBJECT driverObject, int devIndex,
		int pciIndex, PPCI_COMMON_CONFIG pciConfig, PUNICODE_STRING registryPath);
static NTSTATUS InitRings(PTULIP_DEV Tulip);
static void ReleaseRings(PTULIP_DEV tulip);
static void InitFilter(PTULIP_DEV tulip);
static BOOLEAN TulipISR(PKINTERRUPT Interrupt, PVOID ServiceContext);
static VOID TulipRx(PTULIP_DEV tulip);
static void TulipFillAdapter(PUNET_ADAPTER UDIAdapter);
static void LogEvent(PUCHAR format);
static void TulipUpdateTxwall(
		TULIP_DEV *tulip,
		TULIP_ENDPT *endpt,
		int force
														 );
static void PfStart(PTULIP_DEV tulip, UNET_ENDPOINT_ID eid);

#define PENTIUM_MHZ 133
/* Pentium version --- PENTIUM_MHZ defined to MHz rating of processor */
static unsigned __int64 get_pentium_cycle(void) {
	__int64 now;
	unsigned long f1, f2;
	__asm {
		__asm _emit 0x0f
			__asm _emit 0x31
			__asm mov f1, eax
			__asm mov f2, edx
	}
	now = f1 | ((__int64)f2 << 32);
	return now;
}

static unsigned long get_nseconds(unsigned __int64 cycle) {
	return ((cycle*1000) / PENTIUM_MHZ);
}

/* Accessing the on-board EEPROM *********************************************/
/* EEPROM material stolen from Don Becker's newtulip.c */

/* EEPROM_Ctrl bits. */
#define EE_CS           0x01    // EEPROM chip select
#define EE_SHIFT_CLK    0x02    // EEPROM shift clock
#define EE_DATA_WRITE   0x04    // EEPROM chip data in
#define EE_DATA_READ    0x08    // EEPROM chip data out
#define EE_ENB          (0x4800 | EE_CS)	// EEPROM read enable

/* The EEPROM commands include the alway-set leading bit. */
#define EE_WRITE_CMD    (5)
#define EE_READ_CMD     (6)
#define EE_ERASE_CMD    (7)

static int				// returns the value read
ReadEeprom(
		IN ULONG *csr9,  // address of tulip CSR9 register
		IN int location,	// location to read eeprom MAC address into
		IN int big	 // TRUE for big (2k/4k) EEPROM, FALSE for small (.5k/1k)
					)
{
	int i, retval;
	int size = big ? 8 : 6; // number of EEPROM address bits
	int read_cmd = location | (EE_READ_CMD<<size);

	*csr9 = EE_ENB & ~EE_CS;
	*csr9 = EE_ENB;

	/* Shift the read command and address bits out. */
	for (i=size+3-1; i>=0; i--) {
		int dataval = read_cmd & (1<<i) ? EE_DATA_WRITE : 0;
		*csr9 = EE_ENB|dataval; SLOW_DOWN_IO;
		*csr9 = EE_ENB|dataval|EE_SHIFT_CLK; SLOW_DOWN_IO;
		*csr9 = EE_ENB|dataval; SLOW_DOWN_IO;
	}

	retval = 0;
	for (i=16; i>0; i--) {
		*csr9 = EE_ENB|EE_SHIFT_CLK; SLOW_DOWN_IO;
		retval = (retval<<1) | (*csr9 & EE_DATA_READ ? 1 : 0);
		SLOW_DOWN_IO;
		*csr9 = EE_ENB; SLOW_DOWN_IO;
	}

	/* Terminate the EEPROM access. */
	*csr9 = EE_ENB & ~EE_CS;

	//PRINTK("Tulip: EEPROM with %d address bits returned 0x%x\n",
	//		size, retval);
	return retval;
}

static NTSTATUS QueryRegParam(
		IN PUNICODE_STRING registryPath
														 );

/* Loading and unloading of driver *******************************************/

NTSTATUS
DriverEntry (
		IN PDRIVER_OBJECT driverObject,
		IN PUNICODE_STRING registryPath
						)
{
	NTSTATUS stat;
	PTULIP_DEV tulip;
	int i;

	PRINTK("Tulip driver module being loaded, RegistryPath=%S\n",
			registryPath->Buffer );
	PRINTK("File          = %s\n", __FILE__);
	PRINTK("Compiled Date = %s\n", __DATE__);
	PRINTK("Compiled Time = %s\n", __TIME__);
	PRINTK("Last Modified = %s\n", __TIMESTAMP__);

	driverObject->DriverUnload = UnetUnload;
	tulipDriverObject = driverObject;
	LogEvent("tulip: DriverEntry called.\n");

	QueryRegParam(registryPath);

	for (i = 0; i < TULIP_MAX_DEVS; i++)
		tulipDev[i] = NULL;

	for (i = 0; i < 32; i++) {
		PCI_COMMON_CONFIG pciConfig;
		ULONG l = HalGetBusData(PCIConfiguration, 0, i, &pciConfig, sizeof(pciConfig));
		if (l != sizeof(pciConfig)) {
			//PRINTK("Tulip: slot %d unreadable (got %d)\n", i, l);
		} else {
			//PRINTK("Tulip: slot %d -- vendor=0x%x device=0x%x\n", i,
			//     pciConfig.VendorID, pciConfig.DeviceID);
			if (pciConfig.VendorID == PCI_VENDOR_ID_DEC &&
					pciConfig.DeviceID == PCI_DEVICE_ID_DEC_TULIP_FAST) {

				stat = InitDevice(driverObject, numtulipDev, i, &pciConfig,
						registryPath);
				if (stat != STATUS_SUCCESS) return stat;
				numtulipDev++;
			}
		}
	}
	if(numtulipDev == 0) return STATUS_DEVICE_DOES_NOT_EXIST;

	return STATUS_SUCCESS;
}

static NTSTATUS
UnetUnload(
		IN PDRIVER_OBJECT DriverObject
					)
{
	int i;
	NTSTATUS stat;

	PRINTK("Tulip: module being unloaded\n");

	/* Free up the device resources */
	for (i=0; i<TULIP_MAX_DEVS; i++) {
		ReleaseTulip(i);
	}

	LogEvent("Tulip: module unload completed\n");
	return STATUS_SUCCESS;
}

static NTSTATUS
InitDevice(
		IN PDRIVER_OBJECT driverObject,
		IN int devIndex,
		IN int pciIndex,
		IN PPCI_COMMON_CONFIG pciConfig,
		IN PUNICODE_STRING registryPath
					)
{
	PTULIP_DEV tulip;
	PDEVICE_OBJECT deviceObject;
	PCM_RESOURCE_LIST resources;
	int i, j, k;
	int csr6 = 0;
	int resourceNum, resourceSet;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR resourceDescr;
	int irqLevel, irqMode, irqAffinity;
	long irqVector;
	ULONG vector;
	KIRQL irql;
	KAFFINITY affinity;
	NTSTATUS stat;

	/* Allocate a device structure */
	stat = IoCreateDevice(driverObject, sizeof(TULIP_DEV), 0,
			FILE_DEVICE_UNKNOWN, 0, FALSE, &deviceObject);
	if(stat != STATUS_SUCCESS) return stat;
	tulip = (PTULIP_DEV)deviceObject->DeviceExtension;
	tulipDev[devIndex] = tulip;
	tulip->index = devIndex;
	tulip->registryPath = registryPath;
	tulip->deviceObject = deviceObject;

	/* Print some device info */
	PRINTK("Tulip info: vendor=0x%x device=0x%x revision=0x%x cacheLine=%d\n"
			"latency=%d baseClass=%d subClass=%d headerType=%d\n",
			pciConfig->VendorID, pciConfig->DeviceID, pciConfig->RevisionID,
			pciConfig->CacheLineSize, pciConfig->LatencyTimer, pciConfig->BaseClass,
			pciConfig->SubClass, pciConfig->HeaderType);

	/* Get the required bus resources */
	stat = HalAssignSlotResources(registryPath, NULL, driverObject,
			NULL, PCIBus, 0, pciIndex, &resources);
	if (stat != STATUS_SUCCESS) {
		LogEvent("Tulip: Unable to assign slot resources.\n");
		IoDeleteDevice(deviceObject);
		tulipDev[devIndex] = NULL;
		return stat;
	}

	/* Parse what we've gotten... (you'd think this would be simple!) */
	ASSERT(resources->Count == 1 && resources->List);
	resourceNum = resources->List[0].PartialResourceList.Count;
	resourceDescr = resources->List[0].PartialResourceList.PartialDescriptors;
	resourceSet = 0;
	for (i=0; i<resourceNum; i++,resourceDescr++) {
		PHYSICAL_ADDRESS physAddr;
		ULONG addrSpace = 0x0;
		PRINTK("Tulip resource %d (%s): ", i,
				resourceDescr->ShareDisposition == CmResourceShareDeviceExclusive ?
				"(dev exclusive)" : resourceDescr->ShareDisposition ==
				CmResourceShareDriverExclusive ? "(drvr exclusive)" :
				"(shared)");
		switch (resourceDescr->Type) {
			case CmResourceTypePort:
				PRINTK("port 0x%x..0x%x (%s)\n",
						resourceDescr->u.Port.Start.LowPart,
						resourceDescr->u.Port.Start.LowPart
						+ resourceDescr->u.Port.Length - 1,
						resourceDescr->Flags == CM_RESOURCE_PORT_MEMORY ? "mem" : "io");
				tulip->cbio = resourceDescr->u.Port.Start.LowPart;
				resourceSet |= 1<<CmResourceTypePort;
				break;
			case CmResourceTypeInterrupt:
				PRINTK("interrupt level %d vector=0x%x affinity=%d\n",
						resourceDescr->u.Interrupt.Level,
						resourceDescr->u.Interrupt.Vector,
						resourceDescr->u.Interrupt.Affinity);
				irqLevel = resourceDescr->u.Interrupt.Level;
				irqVector = resourceDescr->u.Interrupt.Vector;
				irqAffinity = resourceDescr->u.Interrupt.Affinity;
				irqMode =
					resourceDescr->Flags == CM_RESOURCE_INTERRUPT_LATCHED ?
					Latched : LevelSensitive;
				resourceSet |= 1<<CmResourceTypeInterrupt;
				break;
			case CmResourceTypeMemory:
				PRINTK("memory 0x%x..0x%x (%s)\n",
						resourceDescr->u.Memory.Start.LowPart,
						resourceDescr->u.Memory.Start.LowPart
						+ resourceDescr->u.Memory.Length - 1,
						resourceDescr->Flags == CM_RESOURCE_MEMORY_READ_WRITE ? "rw" :
						resourceDescr->Flags == CM_RESOURCE_MEMORY_READ_ONLY ? "r":"w");
				if(!HalTranslateBusAddress(PCIBus, 0, resourceDescr->u.Memory.Start,
						&addrSpace, &physAddr)) {
					LogEvent("Tulip: Bus address translation failed!\n");
					break;
				}
				tulip->cbma = MmMapIoSpace(physAddr,
						resourceDescr->u.Memory.Length, FALSE);
				if(tulip->cbma == 0) {
					LogEvent("Tulip: failed to memory map CSRs.\n");
					break;
				}
				PRINTK("Tulip: memory mapped CSRs at 0x%x\n", tulip->cbma);
				resourceSet |= 1<<CmResourceTypeMemory;
				break;
			case CmResourceTypeDma:
				PRINTK("DMA channel %d\n",
						resourceDescr->u.Dma.Channel);
				break;
			case CmResourceTypeDeviceSpecific:
				PRINTK("Device specific\n");
				break;
			default:
				LogEvent("Tulip: Unknown resource!");
				break;
		}
	}
	ExFreePool(resources);

	/* Verify we got all the resources we need */
	if (resourceSet != (int)((1<<CmResourceTypePort) |
			(1<<CmResourceTypeInterrupt) | (1<<CmResourceTypeMemory))) {
		LogEvent("Tulip: missing PCI resource(s).\n");
		//IoDeleteDevice(tulipDev[devIndex]->deviceObject);
		//tulipDev[devIndex] = NULL;
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	// Check out PCI configuration make by HAL
	{
		PCI_COMMON_CONFIG pciConfig;
		LONG l;
		l = HalGetBusData(PCIConfiguration, 0, pciIndex, &pciConfig, 24);
		if(l != 24) {
			PRINTK("Get bus data failed: only got %d\n", l);
		} else {
			PRINTK("PCI command register is: 0x%x\n", pciConfig.Command);
			PRINTK("PCI latency timer is: %d\n", pciConfig.LatencyTimer);
			PRINTK("PCI I/O base address is: 0x%x\n",
					pciConfig.u.type0.BaseAddresses[0]);
			PRINTK("PCI memory base address is: 0x%x\n",
					pciConfig.u.type0.BaseAddresses[1]);
		}

		// Check latency timer and make it at least 100 cycles
		if (pciConfig.LatencyTimer < 100) {
			PRINTK("Tulip: Should set latency timer from %d to 100 clocks.\n",
					pciConfig.LatencyTimer);
		}
	}

	/* Initialize some fields */
	{
		int i;
		for (i = 0; i < TULIP_MAX_PORTS; i++) {
			tulip->portmap[i].inuse = 0;
		}
	}

	/* Reset the chip; hold CSR0<0> up for 10 PCI cycles */
	tulip->cbma->CSR0 = 0x00000001;
	SLOW_DOWN_IO;

	/* Install interrupt service routine */
	vector = HalGetInterruptVector(PCIBus, 0, irqLevel, irqVector,
			&irql, &affinity);
	stat = IoConnectInterrupt(&tulip->interruptObject, TulipISR, tulip, NULL,
			vector, irql, irql, irqMode, FALSE, affinity, FALSE);
	if (stat != STATUS_SUCCESS) {
		LogEvent("Tulip: Can't allocate IRQ.\n");
		//IoDeleteDevice(tulipDev[devIndex]->deviceObject);
		tulipDev[devIndex] = NULL;
		return stat;
	}

	// Install interrupt deferred processing routine
	// IoInitializeDpcRequest(deviceObject, TulipDPC);

	// Register driver with the U-Net interface
	{
		PUNET_ADAPTER adapter;
		adapter = UnetCreateAdapterObject(UNET_DEVID_TULIP, devIndex);
		if (adapter == NULL) {
			LogEvent("Tulip: Can't get adapter object\n");
			tulipDev[devIndex] = NULL;
			return STATUS_NO_MEMORY;
		}
		TulipFillAdapter(adapter);
	}

	/*
	 * Now initialize the network interface itself.
	 */

	// This sets 16-lw cache align and unlimited burst len (fff88000 for 0
	// max burst len)
	tulip->cbma->CSR0 = 0x0000e000;

	// Read and parse EEPROM.
	{
		int i;

		DbgPrint("Tulip: Reading EEPROM, please be patient...\n");

		for (i = 0; i < 24; i++) {
			((unsigned short *)tulip->eeprom)[i] = 
				ReadEeprom(&tulip->cbma->CSR9, i, 0);
		}
		for (i = 0; i < 8; i++) {
			if (tulip->eeprom[i] != tulip->eeprom[16 + i]) break;
		}
		if (i >= 8) {
			if (tulip->eeprom[0] == 0xff) {
				DbgPrint("Tulip: Missing EEPROM, multiport board?\n");
			}
			DbgPrint("Tulip: Old-style EEPROM, MAC addr at top.\n");
			tulip->cbma->CSR12 = 0;
			memcpy(tulip->macAddr, tulip->eeprom, 6);
			csr6 = 0x00800000;	// PCS function enabled
		} else {
			int j;
			struct infoleaf140 {
				short media; unsigned char csr12val; char count;
			} *infoleaf = (PVOID)(tulip->eeprom + tulip->eeprom[27]);

			for (i = 24; i < EEPROM_SIZE/2; i++) {
				((unsigned short *)tulip->eeprom)[i] = 
					ReadEeprom(&tulip->cbma->CSR9, i, 0);
			}

		DbgPrint("Tulip: New-Style EEPROM.\n");
		memcpy(tulip->macAddr, tulip->eeprom + 20, 6);

		DbgPrint("Tulip: Leaf node at %d, media %4.4x, csr12 %2.2x, %d blocks.\n",
				tulip->eeprom[27], infoleaf->media, tulip->cbma->CSR12, infoleaf->count);

		j = tulip->eeprom[27]+4;
		for (i = infoleaf->count; i > 0; i--) {
			if (tulip->eeprom[j] & 0x80) { // Extended block
				int blk_type = tulip->eeprom[j+1];
				unsigned char *p = &tulip->eeprom[j+2];
				DbgPrint("Tulip: Extended block for PHY %d, type %d length %d.\n",
						tulip->eeprom[j+2], blk_type, tulip->eeprom[j] & 0x7f);
				switch(blk_type) {
					case 0: {
										// XXX Non-MII transceiver. Steal code from Linux tulip.c
										// if this is needed.
										DbgPrint("Tulip: Extended block type %d not supported.\n",
												blk_type);
										break;
									}
					case 1: {
										int i;
										DbgPrint("Tulip: MII interface PHY %d.\n", p[0]);
										DbgPrint("Tulip: Doing a reset sequence of length %d.\n",
												p[2 + p[1]]);
										for (i = 0; i < p[2 + p[1]]; i++) {
											tulip->cbma->CSR12 = p[3 + p[1] + i]; SLOW_DOWN_IO;
										}
										DbgPrint("Tulip: Doing a transceiver setup sequence len %d.\n",
												p[1]);
										for (i = 0; i < p[1]; i++) {
											tulip->cbma->CSR12 = p[2 + i]; SLOW_DOWN_IO;
										}
										// The following is magic; I don't know why the PCS bit is
										// not set here, but it won't work if it is.
										csr6 = 0x020c0000;

										j += (tulip->eeprom[j] & 0x7f) + 1;
										break;
									}
					default:
									DbgPrint("Tulip: Unknown block type %d.\n");
									break;
				}
			} else {
				DbgPrint("Tulip: Compact block, not supported.\n");
				j += 4;
			}
		}
		}
	}

#ifdef USE_MULTICAST
	(tulip->macAddr)[0] = 0x01;
	(tulip->macAddr)[1] = 0x00;
	(tulip->macAddr)[2] = 0x5e;
	(tulip->macAddr)[3] = 0x00;
	(tulip->macAddr)[4] = 0xff;
	(tulip->macAddr)[5] = 0x00;
#endif

	DbgPrint("Tulip: MAC address ");
	for (j=0; j<6; j++) {
		DbgPrint("%02x%c", (tulip->macAddr)[j], j==5 ? '\n' : ':');
	}

	// Magic? This is a left-over from a DC21040 driver. This register
	// doesn't seem to exist in the 21140. Doing this doesn't hurt...
	tulip->cbma->CSR13 = 0x00000000;
	tulip->cbma->CSR13 = 0x00000004;

	/* Initialize the Tx and Rx rings */
	stat = InitRings(tulip);
	if (stat != STATUS_SUCCESS) {
		PRINTK("Tulip: Error initializing Tx/Rx rings\n");
		tulipDev[devIndex] = NULL;
		//ReleaseTulip(devIndex);
		return stat;
	}

	// Initialize the MAC address filtering
	InitFilter(tulip);

	// Tell the Tulip about the descriptor ring addresses
	tulip->cbma->CSR3 = MmGetPhysicalAddress(tulip->rxRing).LowPart;
	tulip->cbma->CSR4 = MmGetPhysicalAddress(tulip->txRing).LowPart;

	PRINTK("Tulip: Rx ring at 0x%x, Tx ring at 0x%x\n",
			MmGetPhysicalAddress(tulip->rxRing).LowPart,
			MmGetPhysicalAddress(tulip->txRing).LowPart);

	// Set interrupt mask: enable receive ints, disable transmit ints
	tulip->cbma->CSR7 = 0x0001a0e0;

	//tulip->txFullWaitq = NULL;
	// Set operating mode and start the Tx process.
	csr6 |= 0x00062000;	// 100Mbps MII port, min tx threshold, tx enabled
	// csr6 |= 0x00200000;	// store&forward TX
	csr6 |= 0x01000000;	// scrambler for TX (remove for FX)

	if (promiscuousMode) {
		PRINTK("Tulip: Using Promiscuous Mode\n");
		csr6 |= 0x00000040;
	}
	if (fullDuplex) {
		PRINTK("Tulip: Using Full-Duplex operation \n");
		csr6 |= 0x200;
	}
	if (internalLoopback) {
		PRINTK("Tulip: Using Internal Loopback Mode\n");
		csr6 |= 0x200;		// for internal loopback, full-duplex must be present
		csr6 |= 0x400;
	}

	tulip->cbma->CSR6 = csr6;


	// Trigger Tx request for setup frame
	tulip->cbma->CSR1 = 0;
	// See if we can wait 'til the setup frame is consumed
	for(i=0; i<1000 && (tulip->txRing[0].status&0x80000000); i++)
		SLOW_DOWN_IO;
	PRINTK("Tulip: Setup frame status is 0x%x\n",
			tulip->txRing[0].status);

	// Start the Rx process as well
	tulip->cbma->CSR6 = tulip->cbma->CSR6 & ~0xff | 0x02;
	PRINTK("Tulip: Mode register is 0x%x\n", tulip->cbma->CSR6);
	PRINTK("Tulip: Status register is 0x%x\n", tulip->cbma->CSR5);

	return STATUS_SUCCESS;
}

void
ReleaseTulip(
		IN int devIndex
						)
{
	NTSTATUS stat;

	PRINTK("Tulip: releasing device %d\n", devIndex);

	if (tulipDev[devIndex] != NULL) {
		PTULIP_DEV tulip = tulipDev[devIndex];

		UnetDestroyAdapterObject(UNET_DEVID_TULIP, devIndex);
		ReleaseRings(tulip);
		if (tulip->interruptObject)
			IoDisconnectInterrupt(tulip->interruptObject);
		stat = IoAssignResources(tulip->registryPath, NULL,
				tulipDriverObject, NULL, NULL, NULL);
		if (stat != STATUS_SUCCESS)
			PRINTK("Tulip: IoAssignResource failed\n");
		if (tulip->deviceObject != NULL) {
			IoDeleteDevice(tulip->deviceObject);
		}
		tulipDev[devIndex] = NULL;
	}
}

static NTSTATUS
InitRings(
		IN PTULIP_DEV tulip
				 )
{
	int i;
	PHYSICAL_ADDRESS pAddr;

	tulip->txFull = 0;
	tulip->curRx = tulip->curTx = 0;
	tulip->dirtyRx = tulip->dirtyTx = 0;

	// Allocate the Rx buffer region
	pAddr.LowPart = 0xffffffff;
	pAddr.HighPart = 0;
	tulip->rxBufs = MmAllocateContiguousMemory(RX_RING_SIZE*RX_BUF_SZ, pAddr);
	if (tulip->rxBufs == NULL) return STATUS_INSUFFICIENT_RESOURCES;

	// Init Rx ring
	// Each descriptor is owned by tulip, ready to recv; only buffer 1 is used
	// and buffer 2's length is 0, buffer1 points to physica (DMA) addr, we
	// use buffer2 to point to corresponding virtual addr
	for (i=0; i<RX_RING_SIZE; i++) {
		tulip->rxRing[i].status = 0x80000000; // Owned by Tulip
		tulip->rxRing[i].length = RX_BUF_SZ;  // other bits 0 ...
		tulip->rxRing[i].buffer1 =
			MmGetPhysicalAddress(tulip->rxBufs + i*RX_BUF_SZ).LowPart;
		tulip->rxRing[i].buffer2 = (ULONG)tulip->rxBufs + i*RX_BUF_SZ;
	}
	// Mark end of ring
	tulip->rxRing[i-1].length = RX_BUF_SZ | 0x02000000; // end of ring

	for (i = 0; i < TX_RING_SIZE; i++) {
		// Clear owner of Tx ring entries
		tulip->txRing[i].status = 0x00000000;
		// Clear the per-ring-entry Ethernet header
		memset(tulip->txRingHeader[i], 0, 16);
		// Translate Ethernet header addresses (XXX: hope it's contiguous!)
		tulip->txRingHeaderPhys[i] =
			MmGetPhysicalAddress(tulip->txRingHeader[i]).LowPart;
		// Clear per-ring-entry endpoint number (for reaping entries)
		tulip->txRingShadow[i].page_id = 0;
		tulip->txRingShadow[i].chained = FALSE;
		tulip->txRingShadow[i].paddr = 0;
	}

	return STATUS_SUCCESS;
}

static void
ReleaseRings(
		IN PTULIP_DEV tulip
						)
{
	if (tulip->rxBufs) MmFreeContiguousMemory(tulip->rxBufs);
	tulip->rxBufs = NULL;
}

/* Create the setup frame and push it onto the Tx ring */
static void
InitFilter(
		PTULIP_DEV tulip
					)
{
	unsigned short *eaddrs = (unsigned short *)tulip->macAddr;
	int *setup_frm = tulip->setupFrame;
	int i;

	// accept broadcasts
	*setup_frm++ = 0xffff;
	*setup_frm++ = 0xffff;
	*setup_frm++ = 0xffff;
	// accept our own ethernet address
	for (i=1; i<16; i++) {
		*setup_frm++ = eaddrs[0];
		*setup_frm++ = eaddrs[1];
		*setup_frm++ = eaddrs[2];
	}

	// push frame onto ring
	tulip->txRing[0].length = 0x08000000 | 192;
	tulip->txRing[0].buffer1 = MmGetPhysicalAddress(tulip->setupFrame).LowPart;
	tulip->txRing[0].buffer2 = 0;
	tulip->txRing[0].status = 0x80000000;
	tulip->curTx = 1;
	tulip->dirtyTx = 1;
}

/* Tulip UDI entry points ***********************************************/

static NTSTATUS
TulipAdapterInfo(
		OUT PUNET_INFO_RESPONSE resp
								)
{
	PTULIP_DEV tulip;

	try{
		tulip = tulipDev[resp->dev_index];
		resp->version          = UNET_DEV_VERSION;
		resp->rxd_payload_size = RXD_PAYLOAD_SIZE;
		resp->max_pdu          = TULIP_MAX_PDU;
		resp->max_endpts       = TULIP_MAX_ENDPT;
		resp->max_chans        = TULIP_MAX_CHANS;

		resp->tx_buf_align = 2;
		resp->rx_buf_align = 2;

		resp->min_tx_fifo   = TULIP_MIN_TX;
		resp->max_tx_fifo   = TULIP_MAX_TX;
		resp->min_free_fifo = TULIP_MIN_FREE;
		resp->max_free_fifo = TULIP_MAX_FREE;
		resp->min_rx_fifo   = TULIP_MIN_RX;
		resp->max_rx_fifo   = TULIP_MAX_RX;

		memcpy(resp->local_addr.fe.macAddr, tulip->macAddr, 6);
	}
	except(EXCEPTION_EXECUTE_HANDLER){
		PRINTK("An exception occurred in TulipAdapterInfo()\n");
	};
	return STATUS_SUCCESS;
}

#define CHK(val, min, max) do { \
	if(req->val < (min) || req->val > (max)) {\
		PRINTK("TulipCreateEndpoint: illegal %s (%d)\n", #val, req->val);\
			return STATUS_INVALID_PARAMETER;\
	}\
	endpt->val = req->val;\
} while(0)

	static NTSTATUS
	TulipCreateEndpoint(
			IN UNET_ENDPOINT_ID endptId,
			IN PUNET_CREATE_ENDPOINT_REQUEST req,
			OUT PUNET_CREATE_ENDPOINT_RESPONSE resp
										 )
{
	PTULIP_ENDPT endpt;
	PTULIP_DEV tulip;
	PMDL mdl;
	int i;

	PRINTK("Tulip: CreateEndpoint %d 0x%x 0x%x\n",
			endptId, req, resp);

	if(endptId < 0 || endptId > TULIP_MAX_ENDPT || tulipEndpt[endptId] != 0)
		return STATUS_INVALID_PARAMETER;
	if(req->devIndex < 0 || req->devIndex > TULIP_MAX_DEVS ||
			tulipDev[req->devIndex] == 0)
		return STATUS_INVALID_PARAMETER;

	// allocate endpoint structure
	endpt = ExAllocatePool(NonPagedPoolCacheAligned, sizeof(TULIP_ENDPT));
	if(endpt == 0) {
		PRINTK("Tulip: can't alloc endpt struct.\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	endpt->dev = tulipDev[req->devIndex];
	endpt->process = IoGetCurrentProcess();
	tulipEndpt[endptId] = endpt;
	tulip = endpt->dev;

	// check sizes
	CHK(numTxFifoEntries, TULIP_MIN_TX, TULIP_MAX_TX);
	CHK(numFreeFifoEntries, TULIP_MIN_FREE, TULIP_MAX_FREE);
	CHK(numRxFifoEntries, TULIP_MIN_RX, TULIP_MAX_RX);
	CHK(bufferAreaSize, 0, 5*1024*1024); // max due to bug in MmInitialize
	if(req->rxBufferSize != TULIP_MAX_PDU) {
		PRINTK("Tulip: TulipCreateEndpoint: illegal %s (%d)\n", "rxBufferSize",
				req->rxBufferSize);
		return STATUS_INVALID_PARAMETER;
	}

	// allocate buffer area
	endpt->bufferArea =
		MmAllocateContiguousMemory(endpt->bufferAreaSize, topPhysAddr);
	if(endpt->bufferArea == NULL) {
		PRINTK("Tulip: Can't allocate buffer area.\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	endpt->physBufferArea = MmGetPhysicalAddress(endpt->bufferArea).LowPart;
	RtlZeroMemory(endpt->bufferArea, endpt->bufferAreaSize);
	mdl = MmCreateMdl(0, endpt->bufferArea, endpt->bufferAreaSize);
	MmBuildMdlForNonPagedPool(mdl);
	endpt->userBufferArea = (long)MmMapLockedPages(mdl, UserMode);
	resp->bufferArea = (void *)endpt->userBufferArea;
	ExFreePool(mdl);
	PRINTK("Tulip: locked buffer area @0x%08x\n", resp->bufferArea);

	// allocate queues
	endpt->fifoAreaSize = 3*sizeof(void *) +
		endpt->numTxFifoEntries*sizeof(UNET_TX_BUFFER_DESCRIPTOR) +
		endpt->numFreeFifoEntries*sizeof(UNET_RECV_BUFFER_DESCRIPTOR) +
		endpt->numRxFifoEntries*sizeof(UNET_RECEIVE_DESCRIPTOR);
	endpt->txWall = ExAllocatePool(NonPagedPool, endpt->fifoAreaSize);
	if(endpt->txWall == NULL) {
		PRINTK("Tulip: Can't allocate fifos.\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	RtlZeroMemory(endpt->txWall, endpt->fifoAreaSize);
	mdl = MmCreateMdl(0, endpt->txWall, endpt->fifoAreaSize);
	MmBuildMdlForNonPagedPool(mdl);
	endpt->userFifoArea = (long)MmMapLockedPages(mdl, UserMode) +
		((ULONG)endpt->txWall & 0xfff); // XXX: 0xfff is page offset...
	resp->txWall = (void *)endpt->userFifoArea;
	ExFreePool(mdl);
	PRINTK("Tulip: locked fifo area @0x%08x\n", resp->txWall);

	// clear channels
	endpt->numChans = 0;
	for(i=0; i<TULIP_MAX_CHANS; i++) {
		endpt->chans[i].inUse = 0;
	}
	endpt->numFreed = 0;

	// init other endpt pointers
	endpt->freeWall = (long*)(endpt->txWall+1);
	endpt->rxWall = (long*)(endpt->freeWall+1);
	endpt->txFirst = (UNET_TX_BUFFER_DESCRIPTOR*)(endpt->rxWall+1);
	endpt->txNext = endpt->txFirst;
	endpt->txLast = endpt->txFirst + endpt->numTxFifoEntries-1;
	endpt->freeFirst = (UNET_RECV_BUFFER_DESCRIPTOR*)(endpt->txLast+1);
	endpt->freeNext = endpt->freeFirst;
	endpt->freeLast = endpt->freeFirst + endpt->numFreeFifoEntries-1;
	endpt->freePrefetchNext = endpt->freeFirst;
	endpt->freePrefetch = 0;
	endpt->rxFirst = (UNET_RECEIVE_DESCRIPTOR*)(endpt->freeLast+1);
	endpt->rxNext = endpt->rxFirst;
	endpt->rxLast = endpt->rxFirst + endpt->numRxFifoEntries-1;

	// initialize the Fifos
	for(i=0; i<endpt->numTxFifoEntries; i++) {
		endpt->txFirst[i].chan = UNET_CHANNEL_FREE;
	}
	for(i=0; i<endpt->numRxFifoEntries; i++) {
		endpt->rxFirst[i].chan = UNET_CHANNEL_FREE;
	}
	for(i=0; i<endpt->numFreeFifoEntries; i++) {
		endpt->freeFirst[i].buf = (ULONG)UNET_RXB_FREE;
	}

	// initialize the walls
	*endpt->txWall = (ULONG)endpt->txLast - (ULONG)endpt->txFirst;
	*endpt->freeWall = (ULONG)endpt->freeLast - (ULONG)endpt->freeFirst;
	*endpt->rxWall = (ULONG)endpt->rxLast - (ULONG)endpt->rxFirst;

	// construct the response
	resp->rxBufferAllocSize = req->rxBufferSize;
	resp->rxBufferUsedSize = req->rxBufferSize;
	resp->freeWall = (ULONG*)(resp->txWall+1);
	resp->rxWall = (ULONG*)(resp->freeWall+1);
	resp->txFifo = (UNET_TX_BUFFER_DESCRIPTOR*)(resp->rxWall+1);
	resp->freeFifo = (UNET_RECV_BUFFER_DESCRIPTOR*)(resp->txFifo +
			endpt->numTxFifoEntries);
	resp->rxFifo = (UNET_RECEIVE_DESCRIPTOR*)(resp->freeFifo +
			endpt->numFreeFifoEntries);

	return STATUS_SUCCESS;
}

static NTSTATUS
TulipDestroyEndpoint(
		IN UNET_ENDPOINT_ID endptId
										)
{
	static NTSTATUS TulipDeactivateChannel(PUNET_DEACTIVATE_CHANNEL_REQUEST req);
	PTULIP_ENDPT endpt = tulipEndpt[endptId];
	int chan;

	PRINTK("Tulip: DestroyEndpoint %d\n", endptId);

	if(endptId < 0 || endptId > TULIP_MAX_ENDPT)
		return STATUS_INVALID_PARAMETER;
	if(endpt == 0) return STATUS_SUCCESS;

	// Wait for Tx ring entries to drain off
	TulipUpdateTxwall(endpt->dev, endpt, TRUE);

	// destroy all channels
	for(chan=0; chan<TULIP_MAX_CHANS; chan++) {
		if(endpt->chans[chan].inUse) {
			UNET_DEACTIVATE_CHANNEL_REQUEST dcreq = {endptId, chan};
			(void)TulipDeactivateChannel(&dcreq);
		}
	}

	// Clear out TLB
	StlbClear(endptId);

	// Free Fifos and buffer area
	if(endpt->txWall != 0) {
		PMDL mdl = MmCreateMdl(0, endpt->txWall, endpt->fifoAreaSize);
		MmBuildMdlForNonPagedPool(mdl);
		MmUnmapLockedPages((void *)endpt->userFifoArea, mdl);
		ExFreePool(mdl);
		ExFreePool((void *)endpt->txWall);
	}
	if(endpt->bufferArea != 0) {
		PMDL mdl = MmCreateMdl(0, endpt->bufferArea, endpt->bufferAreaSize);
		MmBuildMdlForNonPagedPool(mdl);
		MmUnmapLockedPages((void *)endpt->userBufferArea, mdl);
		ExFreePool(mdl);
		MmFreeContiguousMemory((void *)endpt->bufferArea);
	}

	ExFreePool(endpt);
	tulipEndpt[endptId] = 0;
	return STATUS_SUCCESS;
}

NTSTATUS
TulipActivateChannel(
		IN PUNET_ACTIVATE_CHANNEL_REQUEST req,
		IN PUNET_ACTIVATE_CHANNEL_RESPONSE resp
										)
{
	int endptId = req->endpointId;
	PTULIP_ENDPT endpt = tulipEndpt[endptId];
	PTULIP_DEV tulip;
	int chan;

	PRINTK("Tulip: ActivateChannel %d\n", endptId);

	if(endptId < 0 || endptId > TULIP_MAX_ENDPT || tulipEndpt[endptId] == 0)
		return STATUS_INVALID_PARAMETER;

	tulip = endpt->dev;

	// get a free channel
	chan = req->chan;
	if(chan < 0 || chan >= TULIP_MAX_CHANS || endpt->chans[chan].inUse) {
		for(chan=0; chan<TULIP_MAX_CHANS; chan++) {
			if(endpt->chans[chan].inUse == 0) break;
		}
		if(chan == TULIP_MAX_CHANS) {
			PRINTK("Tulip: ActivateChannel ran out of channels.\n");
			return STATUS_INSUFFICIENT_RESOURCES;
		}
	}

	// Record addresses
	memcpy(&(endpt->chans[chan].txAddr), &(req->txAddr.fe), sizeof(UNET_FE_ADDRESS));
	memcpy(&(endpt->chans[chan].rxAddr), &(req->rxAddr.fe), sizeof(UNET_FE_ADDRESS));

	// Add Rx port to portmap */
	if (tulip->portmap[req->rxAddr.fe.port].inuse) {
		/* Whoops, port already in use */
		return STATUS_INVALID_PARAMETER;
	}
	tulip->portmap[req->rxAddr.fe.port].inuse = 1;
	tulip->portmap[req->rxAddr.fe.port].endpt = endptId;
	tulip->portmap[req->rxAddr.fe.port].chan  = chan;

	// Construct ethernet header for Tx
	memcpy(endpt->chans[chan].header,   req->txAddr.fe.macAddr, 6);
	memcpy(endpt->chans[chan].header+6, tulip->macAddr, 6);
	/* Our protocol type starts at 0x6000, second byte is the port ID */
	endpt->chans[chan].header[12] = 0x60;
	endpt->chans[chan].header[13] = req->txAddr.fe.port;

	endpt->numChans++;
	endpt->chans[chan].inUse = 1;

	/* Start up a buffer-prefetch cycle for the endpoint */
	if (endpt->freePrefetch <= START_PREFETCH_THRESH) PfStart(tulip, endptId);

	resp->chan = chan;
	return STATUS_SUCCESS;
}

static NTSTATUS
TulipDeactivateChannel(
		IN PUNET_DEACTIVATE_CHANNEL_REQUEST req
											)
{
	int endptId = req->endpointId;
	PTULIP_ENDPT endpt = tulipEndpt[endptId];
	PTULIP_DEV tulip;

	PRINTK("Tulip: DeactivateChannel endpt=%d chan=%d\n",
			endptId, req->chan);

	if(endptId < 0 || endptId > TULIP_MAX_ENDPT || tulipEndpt[endptId] == 0)
		return STATUS_INVALID_PARAMETER;
	if(req->chan < 0 || req->chan >= TULIP_MAX_CHANS)
		return STATUS_INVALID_PARAMETER;

	tulip = endpt->dev;

	// mark channel as free
	endpt->chans[req->chan].inUse = 0;

	/* Mark Rx port as free */
	tulip->portmap[endpt->chans[req->chan].rxAddr.port].inuse = 0;

	return STATUS_SUCCESS;
}

/* Tulip Stlb routines *****************************************************/

// Do a prefetch cycle on this endpoint
static void PfStart(PTULIP_DEV tulip, UNET_ENDPOINT_ID eid) {
	int i;
	PTULIP_ENDPT endpt = tulipEndpt[eid];
	ULONG vaddr;

	// XXX Assumes that free buffer fits into a single page
	PRINTK("Tulip: PfStart called on endpoint 0x%x\n",eid);
	while (endpt->freePrefetch < NUM_PREFETCH) {
		if (RXB_ISFREE(endpt->freePrefetchNext->buf)) break;
		vaddr = PAGE_ADDR(endpt->freePrefetchNext->buf);
		if (!StlbTranslate(eid, vaddr, UNET_STLB_PROT_RW, TRUE, NULL)) {
			PRINTK("Tulip: PfStart can't write-translate 0x%x:0x%lx\n", eid, vaddr);
		}
		endpt->freePrefetch++;
		endpt->freePrefetchNext = (endpt->freePrefetchNext == endpt->freeLast)?(endpt->freeFirst):(endpt->freePrefetchNext+1);
	}
}

/* Tulip reception *******************************************************/

static void
TulipRx(IN PTULIP_DEV tulip) {

	// Below parameters used if DPC processing used:
	//	 IN PKDPC Dpc,
	//	 IN PDEVICE_OBJECT DeviceObject,
	//	 IN PIRP Irp,
	//	 IN PVOID Context

	int entry = tulip->curRx;
	unsigned char port;

	PRINTK("Tulip: TulipRx called for rx entry %d\n", entry);

	// Walk down Rx ring until we hit an entry not owned by the Tulip
	while(!(tulip->rxRing[entry].status & 0x80000000)) {
		unsigned long status = tulip->rxRing[entry].status;
		//int length = status>>16;
		int length = *(unsigned short *)((unsigned char *)tulip->rxRing[entry].buffer2 + 14);
		char *buf = (char *)tulip->rxRing[entry].buffer2; // get *virtual* addr
		UNET_ENDPOINT_ID eid;
		TULIP_ENDPT *endpt;
		int chan;
		ULONG paddr, paddr2;

		PRINTK("Tulip: rx entry %d, status 0x%x, length %d, "
				"buf1 0x%x, buf2 0x%x\n", entry, status, length,
				tulip->rxRing[entry].buffer1, tulip->rxRing[entry].buffer2);

		// Any problem?
		if((status&0x8300) != 0x0300) {
			if((status & 0x0300) != 0x0300) {
				PRINTK("Tulip: Frame spanned multiple buffers?\n");
			} else if(status & 0x8000) {
				if(status&0x0890) PRINTK("Tulip: Frame too long or runt\n");
				if(status&0x0004) PRINTK("Tulip: Dribbling bit\n");
				if(status&0x0002) PRINTK("Tulip: CRC error on Rx\n");
				if(status&0x0001) PRINTK("Tulip: FIFO overflow on Rx\n");
			}
		goto drop;
		}

		// Process packet

#ifdef TULIP_DEBUG
		{
			int i;
			DbgPrint("rx: ");
			for (i = 0; i < 20; i++) {
				DbgPrint("%2.2x ", *((unsigned char *)buf + i));
			}
			DbgPrint("\n");
		}
#endif

		if (buf[12] != 0x60) goto drop;
		port = (unsigned char)buf[13];
		if (!tulip->portmap[port].inuse) {
			PRINTK("Tulip: Port 0x%x not in use, dropping packet.\n",port);
			goto drop;
		}
		eid = tulip->portmap[port].endpt;
		endpt = tulipEndpt[eid];
		chan =  tulip->portmap[port].chan;

		// Get next Rx FIFO entry
		if((long)endpt->rxNext - (long)endpt->rxFirst == (long)*endpt->rxWall) {
			// Whoops, Rx FIFO is full. Just drop it...
			// Maybe should signal user somehow
			PRINTK("Tulip: Rx fifo full, dropping packet\n");
			goto drop;
		}

		endpt->rxNext->chan = chan;
		endpt->rxNext->length = length;

		if(length <= RXD_PAYLOAD_SIZE) {
			// Can fit into descriptor itself
			memcpy((void *)(endpt->rxNext->pay_or_buf+0), buf+16, length);

		} else {
			// Actually need a buffer
			int numBufs = 0;

			// XXX Note that this 'do' loop is redundant, as we only allow
			// free buffers to be the size of TULIP_MAX_PDU.
			do {
				ULONG vaddr, vpage, voffset;

try_another_rxb:
				// Get an Rx buffer
				if(RXB_ISFREE(endpt->freeNext->buf)) {
					// Whoops, no buffer available. Drop out...
					endpt->rxNext->chan = UNET_CHANNEL_ERROR;
					break;
				}

				// translate buffer address
				vaddr = (unsigned long)endpt->freeNext->buf;
				vpage = PAGE_ADDR(vaddr);
				voffset = PAGE_OFFSET(vaddr);
				PRINTK("TulipRx: Free buf vaddr is 0x%lx\n",vaddr);

				if (voffset + length > PAGE_SIZE) {
					// Buffer spans two pages
					(void)StlbTranslate(eid, vpage, UNET_STLB_PROT_RW, FALSE, &paddr);
					(void)StlbTranslate(eid, vpage+PAGE_SIZE, UNET_STLB_PROT_RW, FALSE, &paddr2);
					PRINTK("TulipRx: Free buf saddrs are 0x%lx, 0x%lx\n", paddr, paddr2);

					endpt->freeNext->buf = (ULONG)UNET_RXB_FREE;
					endpt->freeNext = endpt->freeNext==endpt->freeLast ?
						endpt->freeFirst : endpt->freeNext+1;
					*endpt->freeWall = (long)endpt->freeNext - (long)endpt->freeFirst;
					endpt->freePrefetch--;

					if (!paddr || !paddr2) {
						PRINTK("Tulip: Can't get translation for recv address\n");
						if (endpt->freePrefetch) {
							PRINTK("TulipRx: Trying to get another free buffer\n");
							goto try_another_rxb;
						} else {
							// No more free buffers available (nothing pending prefetch),
							// just drop
							PRINTK("TulipRx: Dropping as free buffers not resident\n");
							endpt->rxNext->chan = UNET_CHANNEL_ERROR;
							break;
						}
					}
					endpt->rxNext->pay_or_buf[numBufs] = vaddr;
					// Copy the data out
					memcpy(paddr|voffset, buf+16, PAGE_SIZE-voffset);
					memcpy(paddr2, buf+16+(PAGE_SIZE-voffset), MIN(length, TULIP_MAX_PDU)-(PAGE_SIZE-voffset));
				} else {

					// Buffer fits into one page
					(void)StlbTranslate(eid, vpage, UNET_STLB_PROT_RW, FALSE, &paddr);
					PRINTK("TulipRx: Free buf saddr is 0x%lx\n",paddr);

					endpt->freeNext->buf = (ULONG)UNET_RXB_FREE;
					endpt->freeNext = endpt->freeNext==endpt->freeLast ?
						endpt->freeFirst : endpt->freeNext+1;
					*endpt->freeWall = (long)endpt->freeNext -
						(long)endpt->freeFirst;
					endpt->freePrefetch--;

					if (!paddr) {
						if (endpt->freePrefetch) {
							PRINTK("TulipRx: Trying to get another free buffer\n");
							goto try_another_rxb;
						} else {
							// No more free buffers available (nothing pending prefetch),
							// just drop
							PRINTK("TulipRx: Dropping as free buffers not resident\n");
							endpt->rxNext->chan = UNET_CHANNEL_ERROR;
							break;
						}
					}

					endpt->rxNext->pay_or_buf[numBufs] = vaddr;
					memcpy(paddr|voffset, buf+16, MIN(length, TULIP_MAX_PDU));
				}
				length -= MIN(length, TULIP_MAX_PDU);
				numBufs++;
			} while(length != 0 && numBufs < RXD_PAYLOAD_ENTRIES);
		}

		// update rxNext
		endpt->rxNext = endpt->rxNext==endpt->rxLast ?
			endpt->rxFirst : endpt->rxNext+1;

		// Give descriptor back to Tulip and bump curRx
drop:	tulip->rxRing[entry].status = 0x80000000;
			entry++; if(entry == RX_RING_SIZE) entry = 0;
	}

	tulip->curRx = entry;
}

static BOOLEAN
TulipISR (
		IN PKINTERRUPT Interrupt,
		IN PVOID ServiceContext
				 )
{
	TULIP_DEV *tulip = (PTULIP_DEV)ServiceContext;
	ULONG csr5;

	PRINTK("Tulip: INTERRUPT!\n");

	csr5 = tulip->cbma->CSR5;

	// ACK the interrupt to the tulip
	tulip->cbma->CSR5 = csr5 & 0x0001ffff;
	PRINTK("Tulip: Interrupt CSR5: 0x%08x, new CSR5: 0x%08x.\n",
			csr5, tulip->cbma->CSR5);

	// If no normal or abnormal interrupt... */
	if ((csr5 & 0x00018000) == 0) {
		PRINTK("Tulip: unknown interrupt cause!\n");
		return FALSE;
	}

	// Check for "Rx interrupt, no Tx interrupt, Tx process not stopped, and
	// no abnormal interrupt" all at once:
	if((csr5 & (0x0040|0x0001|0x0004|0x8000)) == 0x0040) {
		// Rx done interrupt only... fast path
		TulipRx(tulip);
		//  IoRequestDpc(tulip->deviceObject, (PIRP)0, tulip);

	} else {
		// it's more complicated...

		// Rx done interrupt
		if(csr5 & 0x0040) {
			TulipRx(tulip);
			//  IoRequestDpc(tulip->deviceObject, (PIRP)0, tulip);
		}

		// Tx done interrupt
		if (csr5 & 0x0001) {
			PRINTK("Tulip: Tx done interrupt?\n");
		}

		// Log errors
		if(csr5 & 0x8000) {   /* Abnormal error summary */
			PRINTK("Tulip: Abnormal interrupt, csr5: 0x%x\n", csr5);
			if(csr5 & 0x0008) PRINTK("Tulip: Tx jabber timeout\n");
			if(csr5 & 0x0100)
				PRINTK("Tulip: Rx process stopped, missed %ld frames\n",
						(tulip->cbma->CSR8 & 0xffff));
			if(csr5 & 0x0800) PRINTK("Tulip: Hmm... GT expired?\n");
			if(csr5 & 0x0080) PRINTK("Tulip: Rx buffer unavailable!\n");
		}
	}

	PRINTK("Tulip: Leaving interrupt\n");
	return TRUE;
}

/* Tulip transmission ****************************************************/

// Update Tx wall pointer to reflect entries that have been transmitted.
// If `force' is set, busy-wait until all entries have been reaped.
static void
TulipUpdateTxwall(
		TULIP_DEV *tulip,
		TULIP_ENDPT *endpt,
		int force
								 )
{
	// Walk down Tx_ring using the device-global dirty_tx pointer. Update
	// the 'num_freed' counted for _all_ endpoints, then update tx_wall
	// for our endpoint only.
	int num_spins = 0;

	while(tulip->txFull || (tulip->dirtyTx != tulip->curTx)) {
		int entry = tulip->dirtyTx;
		unsigned long status = tulip->txRing[entry].status;
		int endptNum;

		// This entry not Tx'd yet
		/* XXX If we get an 'Tx underflow error' this ends up spinning
		 * forever. Why? Maybe the 'own' bit is never cleared on entries
		 * which have errors. */
		if(status & 0x80000000) {
			if(force) {
				if (num_spins++ > 32) {
					DbgPrint("Tulip: Timed out waiting for Tx entries to drain, Tx ring may be corrupt.\n");
					break;
				} else {
					continue;	// gotta wait...
				}
			}
			else break;			// done
		}

		PRINTK("Tulip: Reaping Tx ring entry %d, cur=%d\n",
				tulip->dirtyTx, tulip->curTx);

		// log entry to reap in appropriate endpoint
		endptNum = PAGE_OFFSET(tulip->txRingShadow[entry].page_id);
		if(tulipEndpt[endptNum] && !(tulip->txRingShadow[entry].chained))
			tulipEndpt[endptNum]->numFreed++;
		tulip->txRingShadow[entry].chained = FALSE;

		StlbDecRefcnt(PAGE_OFFSET(tulip->txRingShadow[entry].page_id),
				PAGE_ADDR(tulip->txRingShadow[entry].page_id),
				tulip->txRingShadow[entry].paddr);

		// Clear tx_full
		tulip->txFull = 0;

		// bump index to next descriptor
		tulip->dirtyTx++;
		if(tulip->dirtyTx == TX_RING_SIZE) tulip->dirtyTx = 0;
	}

	PRINTK("Tulip: Num freed is %d\n", endpt->numFreed);

	// Now, free Tx FIFO entries up to current `num_freed' */
	if(endpt->numFreed) {
		int numWalked = 0;
		// XXX bug: we're trusting the txwall value which could have been
		// changed by the user to point into never-never land
		UNET_TX_BUFFER_DESCRIPTOR *curWall = (UNET_TX_BUFFER_DESCRIPTOR *)
			((long)endpt->txFirst + *(endpt->txWall));

		PRINTK("Tulip: txwall starting off at 0x%x\n",
				*(endpt->txWall));

		// Walk down Tx fifo and mark entries as free.
		// Also bump Tx wall past entries marked as UNET_CHANNEL_ERROR
		while(endpt->numFreed) {
			// we're trying to free the next entry...
			UNET_TX_BUFFER_DESCRIPTOR *nextWall =
				curWall==endpt->txLast ? endpt->txFirst : curWall+1;
			// if it's not in error, then mark as free
			if(nextWall->chan != UNET_CHANNEL_ERROR) {
				PRINTK("Tulip: freeing 0x%x\n",
						(ULONG)nextWall-(ULONG)endpt->txFirst);
				nextWall->chan = UNET_CHANNEL_FREE;
				endpt->numFreed--;
			}
			// bump wall
			curWall = nextWall;
			numWalked++;
		}

		PRINTK("Tulip: numWalked is %d, size %d, new txWall: 0x%lx\n",
				numWalked, sizeof(UNET_TX_BUFFER_DESCRIPTOR),
				(long)curWall - (long)endpt->txFirst);

		// Poke the new wall value
		*(endpt->txWall) = (long)curWall - (long)endpt->txFirst;
	}
}

// Trigger a transmit request on the Tulip.
#define TRIGGER_TX(tulip) tulip->cbma->CSR1 = 0

// Transmit new entries on the given endpoint
// XXX Known bug: Chained PDU's don't support buffers crossing pages
static void
TulipTransmit(
		TULIP_DEV *tulip,
		TULIP_ENDPT *endpt,
		int endptNum
						 )
{
	int prefix = 0;
	int nextPrefix;
	int entry;
	int firstTx, i;
	unsigned short totalLen;
	volatile UNET_TX_BUFFER_DESCRIPTOR *txBegin;
	ULONG vaddr, vpage, voffset, paddr, paddr2;

	// If the device Tx ring is full, try to free some entries
	if(tulip->txFull) {
		PRINTK("Tulip: Tx ring is full\n");
		TulipUpdateTxwall(tulip, endpt, FALSE);
		if(tulip->txFull) return; // still full: abandon
	}

	// Get the current Tx ring entry
	entry = tulip->curTx;

	// Definition of txWall: the txWall entry will NOT be pushed
	// into, it's always 'free'
	PRINTK("Tulip: Transmit: Channel=%d curTx=%d\n", endpt->txNext->chan, entry);

	while(endpt->txNext->chan >= 0) {
		int chan = endpt->txNext->chan;
		PRINTK("Tulip: Transmit: Channel=%d curTx=%d dirtyTx=%d\n",
				chan, entry, tulip->dirtyTx);

		// XXX For each 'continue' really need to bump the tx_next ptr as well */
		if(chan >= TULIP_MAX_CHANS || !endpt->chans[chan].inUse) {
			PRINTK("Tulip: Bad channel\n");
			endpt->txNext->chan = UNET_CHANNEL_ERROR;
			prefix = 0;
			continue;
		}
		// XXX For multi-buffer PDUs need to check of total length is too large */
		if(endpt->txNext->length > TULIP_MAX_PDU) {
			PRINTK("Tulip: Total length is too large\n");
			endpt->txNext->chan = UNET_CHANNEL_ERROR;
			prefix = 0;
			continue;
		}

		// Now we have a valid channel, buf, length
		// Push the entry onto the Tx ring. This code duplicates tulip_push_tx()...
		nextPrefix = TX_CONTINUE(endpt->txNext->buf);

		if(!prefix && !nextPrefix) {
			// This is a single-PDU entry... Optimize for that
			PRINTK("Tulip: Single-PDU entry\n");

			// Copy the per-Tx-ring header into place
			memcpy(tulip->txRingHeader[entry], endpt->chans[chan].header, 14);

			// Add the length
			*(unsigned short *)(tulip->txRingHeader[entry]+14) =
				endpt->txNext->length;

			// Translate address
			vaddr = endpt->txNext->buf & 0xfffffffe;
			vpage = PAGE_ADDR(vaddr); voffset = PAGE_OFFSET(vaddr);

			if (voffset + endpt->txNext->length > PAGE_SIZE) {
				// Buffer spans two pages
				if (((entry+1)%TX_RING_SIZE) == tulip->dirtyTx) {
					// Can't satisfy the two-buffer push, just return to caller
					PRINTK("Tulip: Can't do two-buffer push.\n");
					tulip->txFull = TRUE;
					break;
				}

				paddr = StlbTranslate(endptNum, vpage, UNET_STLB_PROT_RO, TRUE, NULL);
				paddr2 = StlbTranslate(endptNum, vpage+PAGE_SIZE, UNET_STLB_PROT_RO, TRUE, NULL);
				if (!paddr || !paddr2) {
					PRINTK("Tulip: Can't get translation for xmit address\n");
					endpt->txNext->chan = UNET_CHANNEL_ERROR;
					prefix = 0;
					continue;
				}
				PRINTK("Tulip: Tx got 2 pages 0x%lx -> 0x%lx, 0x%lx\n", vaddr, paddr|voffset, paddr2);

#ifdef TULIP_DEBUG
				{
					int i;
					DbgPrint("tx: ");
					for (i = 0; i < 16; i++ ) {
						DbgPrint("%2.2x ", (unsigned char)*(tulip->txRingHeader[entry]+i));
					}
					for (i = 0; i < 10; i++) {
						DbgPrint("%2.2x ", *((unsigned char *)vaddr + i));
					}
					DbgPrint("\n");
				}
#endif

				// Construct first descriptor
				tulip->txRing[entry].length =
					16 |				// length 1
					((PAGE_SIZE - voffset) << 11) |	// length 2
					0x20000000 |			// First of PDU
					(entry==TX_RING_SIZE-1 ? 0x02000000 : 0x0); // end-of-ring
				tulip->txRing[entry].buffer1 = tulip->txRingHeaderPhys[entry];
				tulip->txRing[entry].buffer2 = paddr | voffset;
				firstTx = entry;

				tulip->txRingShadow[entry].page_id = (vpage | endptNum);
				tulip->txRingShadow[entry].chained = TRUE;
				tulip->txRingShadow[entry].paddr = paddr;

				entry = entry==TX_RING_SIZE-1 ? 0 : entry+1;

				// Construct second descriptor
				tulip->txRing[entry].length =
					endpt->txNext->length - (PAGE_SIZE-voffset) |			// Length 1 only
					0x40000000 |				// Last descriptor for PDU
					(entry==TX_RING_SIZE-1 ? 0x02000000 : 0x0); // end-of-ring
				tulip->txRing[entry].buffer1 = paddr2;

				tulip->txRingShadow[entry].page_id = ((vpage+PAGE_SIZE)|endptNum);
				tulip->txRingShadow[entry].paddr = paddr2;

				// Now loop over all descriptors for this PDU in reverse order
				// and set the ownership bit so the Tulip can have them
				i = entry;
				for(;;) {
					tulip->txRing[i].status = 0x80000000;
					if(i == firstTx) break;
					i = i==0 ? TX_RING_SIZE-1 : i-1;
				}
				goto finish_txnext;
			}

			// Buffer fits within one page
			paddr = StlbTranslate(endptNum, vpage, UNET_STLB_PROT_RO, TRUE, NULL);
			if (!paddr) {
				PRINTK("Tulip: Can't get translation for xmit address\n");
				endpt->txNext->chan = UNET_CHANNEL_ERROR;
				prefix = 0;
				continue;
			}

			PRINTK("Tulip: Tx got 0x%lx -> 0x%lx\n", vaddr, paddr|voffset);

			// Construct the descriptor
			tulip->txRing[entry].length =
				16 |				// length 1
				(endpt->txNext->length << 11) |	// length 2
				0x60000000 |			// First and Last of PDU
				(entry==TX_RING_SIZE-1 ? 0x02000000 : 0x0); // end-of-ring
			tulip->txRing[entry].buffer1 = tulip->txRingHeaderPhys[entry];
			tulip->txRing[entry].buffer2 = paddr | voffset;

			tulip->txRingShadow[entry].page_id = (vpage | endptNum);
			tulip->txRingShadow[entry].paddr = paddr;
			tulip->txRing[entry].status = 0x80000000; // Go!

		}

		// XXX mdw: The below is currently unsupported in U-Net/MM.
		// Nothing's wrong with the design; just the above code needs to be
		// duplicated below.
#undef SUPPORT_CHAINED_TX
#ifdef SUPPORT_CHAINED_TX
		else if (!prefix) {
			// This is the first segment of a multi-segment PDU
			PRINTK("Tulip: First segment of multi-segment PDU\n");

			// save where we started to be able to go back if error
			txBegin = endpt->txNext;
			firstTx = entry;
			totalLen = endpt->txNext->length;

			// Copy the per-Tx-ring header into place
			memcpy(tulip->txRingHeader[entry], endpt->chans[chan].header, 14);

			// Construct descriptor (don't know PDU length yet)
			tulip->txRing[entry].length =
				16 |				// length 1
				(endpt->txNext->length << 11) |	// length 2
				0x20000000 |			// First of PDU
				(entry==TX_RING_SIZE-1 ? 0x02000000 : 0x0); // end-of-ring
			tulip->txRing[entry].buffer1 = tulip->txRingHeaderPhys[entry];
			tulip->txRing[entry].buffer2 = paddr | voffset;
			// Don't give it to the Tulip yet

		} else if(nextPrefix) {
			// Middle segment of a multi-segment PDU
			PRINTK("Tulip: Middle segment of multi-segment PDU\n");

			totalLen += endpt->txNext->length;

			tulip->txRing[entry].length =
				endpt->txNext->length |			// Length 1 only
				(entry=TX_RING_SIZE-1 ? 0x02000000 : 0x0); // end-of-ring
			tulip->txRing[entry].buffer1 = paddr | voffset;
			// Don't give it to the Tulip yet

		} else {
			// Last segment of a multi-segment PDU
			PRINTK("Tulip: Last segment of PDU\n");

			// Compute the total PDU length and push it into the header
			totalLen += endpt->txNext->length;
			*(unsigned short *)(tulip->txRingHeader[firstTx]+14) = totalLen;

			tulip->txRing[entry].length =
				endpt->txNext->length |			// Length 1 only
				0x40000000 |				// Last descriptor for PDU
				(entry==TX_RING_SIZE-1 ? 0x02000000 : 0x0); // end-of-ring
			tulip->txRing[entry].buffer1 = paddr | voffset;

			// Now loop over all descriptors for this PDU in reverse order
			// and set the ownership bit so the Tulip can have them
			i = entry;
			for(;;) {
				tulip->txRing[i].status = 0x80000000;
				if(i == firstTx) break;
				i = i==0 ? TX_RING_SIZE-1 : i-1;
			}
		}
#endif /* SUPPORT_CHAINED_TX */

finish_txnext:
	prefix = nextPrefix;

	// Bump pointer to next fifo entry
	endpt->txNext = endpt->txNext==endpt->txLast ?
		endpt->txFirst : endpt->txNext+1;

	// Bump the device txRing entry
	entry = entry==TX_RING_SIZE-1 ? 0 : entry+1;

	if(entry == tulip->dirtyTx) {
		// check to make sure you don't transmit half of a PDU
		if(nextPrefix) {
#ifdef SUPPORT_CHAINED_TX
			// restore old state
			entry = firstTx;
			endpt->txNext = txBegin;
#endif SUPPORT_CHAINED_TX
			// ring won't be really full but we can't fit the PDU so mark
			// it as full anyway
		}

		// We've hit the end of the free ring entries, mark ring as full
		tulip->txFull = TRUE;

		// And return to the caller; don't transmit any new entries
		break;
	}
	}

	tulip->curTx = entry;	// Update the device ring curTx pointer
	TRIGGER_TX(tulip);		// Start the Tx process
	return;
}

int
TulipService(
		IN UNET_ENDPOINT_ID index
						)
{
	TULIP_DEV *tulip;
	TULIP_ENDPT *endpt;

	if(index < 0 || index > TULIP_MAX_ENDPT || tulipEndpt[index] == 0)
		return FALSE;

	endpt = tulipEndpt[index];
	tulip = endpt->dev;
	PRINTK("Tulip: Status register is 0x%x\n", tulip->cbma->CSR5);

	// Push any new Tx fifo entries onto the device Tx ring
	TulipTransmit(tulip, endpt, index);

	// Update the wall
	TulipUpdateTxwall(tulip, endpt, FALSE);

	// Start prefetch for Rx buffers
	if (endpt->freePrefetch < NUM_PREFETCH) PfStart(tulip, index);

	PRINTK("Tulip: Status register is 0x%x\n", tulip->cbma->CSR5);
	return TRUE;
}

/* Collect all i/f routines in a struct **************************************/

static void
TulipFillAdapter(
		IN PUNET_ADAPTER UDIAdapter
								)
{
	UDIAdapter->createEndpoint = TulipCreateEndpoint;
	UDIAdapter->destroyEndpoint = TulipDestroyEndpoint;
	UDIAdapter->activateChannel = TulipActivateChannel;
	UDIAdapter->deactivateChannel = TulipDeactivateChannel;
	UDIAdapter->adapterInfo = TulipAdapterInfo;
	UDIAdapter->transmit = TulipService;
	UDIAdapter->stopAdapter = 0; // TulipStopAdapter;
}

/* Event logging (for debug purposes, mainly) ********************************/
static void
LogEvent (
		IN PUCHAR format
				 )
{
	ANSI_STRING          fmtString;
	UNICODE_STRING       uniString;
	UNICODE_STRING       labelString;
	PIO_ERROR_LOG_PACKET errorLogEntry;
	PWCHAR               dest;
	LARGE_INTEGER        dummy={0,0};

	RtlInitAnsiString           (&fmtString, format);
	RtlAnsiStringToUnicodeString(&uniString, &fmtString, TRUE);
	RtlInitUnicodeString        (&labelString, L"Tulip");

	errorLogEntry = IoAllocateErrorLogEntry(tulipDriverObject,
			(UCHAR)(sizeof(IO_ERROR_LOG_PACKET) + labelString.Length +
							uniString.Length + 2*sizeof(WCHAR))); 

	if (!errorLogEntry) return;

	errorLogEntry->MajorFunctionCode = 0 ;
	errorLogEntry->RetryCount       = 1;
	errorLogEntry->DumpDataSize     = 0;
	errorLogEntry->NumberOfStrings  = 2;
	errorLogEntry->StringOffset     = 0; 
	errorLogEntry->EventCategory    = 1;  // 0 before
	errorLogEntry->ErrorCode        = 1; // IO_ERR_DRIVER_ERROR;
	errorLogEntry->UniqueErrorValue = 0;
	errorLogEntry->FinalStatus      = STATUS_IO_DEVICE_ERROR;
	errorLogEntry->SequenceNumber   = 0;
	errorLogEntry->IoControlCode    = 0;
	errorLogEntry->DeviceOffset     = dummy;

	dest = (PWCHAR)((char*)errorLogEntry + sizeof(IO_ERROR_LOG_PACKET));
	RtlMoveMemory(dest, labelString.Buffer, labelString.Length) ;
	dest[labelString.Length/sizeof(WCHAR)] = L'\0' ;
	dest = (PWCHAR)((char*)dest + labelString.Length+sizeof(WCHAR));
	RtlMoveMemory(dest, uniString.Buffer, uniString.Length) ;
	dest[uniString.Length/sizeof(WCHAR)] =L'\0';

	IoWriteErrorLogEntry(errorLogEntry);
	RtlFreeUnicodeString(&uniString);
}

NTSTATUS QueryRegParam(
		IN PUNICODE_STRING registryPath
											)
{
	NTSTATUS status;
	try{
		RTL_QUERY_REGISTRY_TABLE    paramTable[10];

		RtlZeroMemory(paramTable, sizeof(RTL_QUERY_REGISTRY_TABLE)*10);

		paramTable[0].QueryRoutine = NULL;
		paramTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
		paramTable[0].Name = L"FullDuplex";
		paramTable[0].EntryContext = &fullDuplex;
		paramTable[0].DefaultLength = sizeof(ULONG);
		paramTable[0].DefaultData = NULL;

		paramTable[1].QueryRoutine = NULL;
		paramTable[1].Flags = RTL_QUERY_REGISTRY_DIRECT;
		paramTable[1].Name = L"InternalLoopBack";
		paramTable[1].EntryContext = &internalLoopback;
		paramTable[1].DefaultLength = sizeof(ULONG);
		paramTable[1].DefaultData = NULL;

		paramTable[2].QueryRoutine = NULL;
		paramTable[2].Flags = RTL_QUERY_REGISTRY_DIRECT;
		paramTable[2].Name = L"PromiscuousMode";
		paramTable[2].EntryContext = &promiscuousMode;
		paramTable[2].DefaultLength = sizeof(ULONG);
		paramTable[2].DefaultData = NULL;

		paramTable[3].QueryRoutine = NULL;

		status =  RtlQueryRegistryValues(
				RTL_REGISTRY_ABSOLUTE | RTL_REGISTRY_OPTIONAL,
				registryPath->Buffer, paramTable, NULL, NULL);
	}
	except(EXCEPTION_EXECUTE_HANDLER){
		PRINTK("An exception occurred in QueryRegParam()\n");
	}

	return status;
}
