/* Winnt zeitnet.c,
 * Matt Welsh (mdw@cs.cornell.edu)
 *
 * U-Net/MM driver for the Zeitnet ATM NIC.
 *
 * Copyright (c) 1995, 1996 by Cornell University.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL CORNELL UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF CORNELL UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * CORNELL UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND CORNELL UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * M. Welsh, 18 Nov 96
 */

/* All zwait/zout(foo, CMR) and zpokel's should be protected by
 * KeRaiseIrql(POWER_LEVEL)/KeLowerIrql.
 */

#include <ntddk.h>
#include "unetuser.h"
#include "unetkernel.h"
#include "zeitnet.h"
#include "zeitnetuser.h"
#include "utlb.h"

/* Define for verbose debug logging */
#undef ZEITNET_DEBUG
#undef ZEITNET_DEBUG_PRINTPACKETS
/* Define for internal loopback */
#undef ZEITNET_LOOPBACK

#ifdef ZEITNET_DEBUG
#define PRINTK DbgPrint
#else
#define PRINTK 0 &&
#endif

/* Global variables ******************************************************/

static PDRIVER_OBJECT zeitnetDriverObject;
static ZEITNET_DEV *zeitnetDev[ZEITNET_MAX_DEVS];
static int numZeitnetDev = 0;
static ZEITNET_ENDPT *zeitnetEndpt[ZEITNET_MAX_ENDPT];
static ULONG dummy[2] = {0, 0};  // For empty pools

/* Forward function declarations *************************************/

static NTSTATUS ZeitnetUnload(PDRIVER_OBJECT DriverObject);
static void ReleaseZeitnet(int devIndex);
static NTSTATUS InitDevice(PDRIVER_OBJECT driverObject, int devIndex,
													 int pciIndex, PPCI_COMMON_CONFIG pciConfig, PUNICODE_STRING registryPath);
static void InitShaper(PZEITNET_DEV zeitnet, int shaper, UCHAR I, ULONG M, UCHAR P, UCHAR C);
static BOOLEAN ZeitnetISR(PKINTERRUPT Interrupt, PVOID ServiceContext);
static void ZeitnetFillAdapter(PUNET_ADAPTER UDIAdapter);
static void PfStart(PZEITNET_DEV zeitnet, UNET_ENDPOINT_ID eid);
static void ZeitnetRx(ZEITNET_DEV *zeitnet);
static void ZeitnetUpdateTxwall(ZEITNET_DEV *zeitnet,
										ZEITNET_ENDPT *endpt,
										BOOLEAN force);
		 
/* Utility functions ******************************************************/

#define PENTIUM_MHZ 133
/* Pentium version --- PENTIUM_MHZ defined to MHz rating of processor */
static unsigned __int64 get_pentium_cycle(void) {
	__int64 now;
	unsigned long f1, f2;
	__asm {
		__asm _emit 0x0f
		__asm _emit 0x31
		__asm mov f1, eax
		__asm mov f2, edx
	}	 
	now = f1 | ((__int64)f2 << 32);
	return now;
}

static unsigned long get_nseconds(unsigned __int64 cycle) {
	return ((cycle*1000) / PENTIUM_MHZ);
}

// Pause for 1ms
static void slow_down_io(void) {
	unsigned long t1, t2;
	t1 = get_nseconds(get_pentium_cycle());
	do {
		t2 = get_nseconds(get_pentium_cycle());
	}  while ((t2 - t1) < 1000000) ;
}

/* Poke a value to control memory */
static void zpokel(PZEITNET_DEV zeitnet, ULONG value, ULONG addr) {
	zwait(0);
	zout(value, CER);
	zout(0x0f000000 | addr, CMR);
}

/* Peek a value from control memory */
static ULONG zpeekl(PZEITNET_DEV zeitnet, ULONG addr) {
	zwait(0);
	zout(0x1f000000 | addr, CMR);
	zwait(0);
	return zin(CER);
}

/* Loading and unloading of driver *******************************************/

NTSTATUS
DriverEntry (
	     IN PDRIVER_OBJECT driverObject,
	     IN PUNICODE_STRING registryPath
	     )
{
  NTSTATUS stat;
  int i;

  PRINTK("Zeitnet driver module being loaded, RegistryPath=%S\n", registryPath);
  driverObject->DriverUnload = ZeitnetUnload;
  zeitnetDriverObject = driverObject;

  for (i = 0; i < ZEITNET_MAX_DEVS; i++)
    zeitnetDev[i] = NULL;

  for (i = 0; i < 32; i++) {
    PCI_COMMON_CONFIG pciConfig;
    ULONG l = HalGetBusData(PCIConfiguration, 0, i, &pciConfig, sizeof(pciConfig));
    if (l != sizeof(pciConfig)) {
      //PRINTK("Zeitnet: slot %d unreadable (got %d)\n", i, l);
    } else {
      //PRINTK("Zeitnet: slot %d -- vendor=0x%x device=0x%x\n", i,
      //     pciConfig.VendorID, pciConfig.DeviceID);
      if (pciConfig.VendorID == PCI_VENDOR_ID_ZEITNET) {
				if ((pciConfig.DeviceID == PCI_DEVICE_ID_ZEITNET_1221) ||
						(pciConfig.DeviceID == PCI_DEVICE_ID_ZEITNET_1225)) {
					stat = InitDevice(driverObject, numZeitnetDev, i, &pciConfig,
														registryPath);
					if (stat != STATUS_SUCCESS) return stat;
					numZeitnetDev++;
				}
			}
    }
  }
  if(numZeitnetDev == 0) return STATUS_DEVICE_DOES_NOT_EXIST;
	
  return STATUS_SUCCESS;
}

static NTSTATUS
ZeitnetUnload(
	   IN PDRIVER_OBJECT DriverObject
	   )
{
  int i;
  NTSTATUS stat;

  PRINTK("Zeitnet: module being unloaded\n");
	
  /* Free up the device resources */
  for (i=0; i<ZEITNET_MAX_DEVS; i++) {
    ReleaseZeitnet(i);
  }

  return STATUS_SUCCESS;
}


// Initialize PHY device. Assumes uPD98402.
static NTSTATUS InitPhy(PZEITNET_DEV zeitnet) {
	// XXX Only add this code if it's necessary later on
	return STATUS_SUCCESS;
}

static NTSTATUS InitMbx(PZEITNET_DEV zeitnet,
												int mbx_num, int mbx_size, int entry_size) {

	ULONG addr, paddr;
	
	addr = (ULONG)MmAllocateContiguousMemory(2*mbx_size*entry_size, topPhysAddr);
	if (!addr) {
		DbgPrint("Zeitnet: Can't allocate %d bytes for mailbox %d.\n",
						 2*mbx_size*entry_size, mbx_num);
    return STATUS_INSUFFICIENT_RESOURCES;
	}

	if ((addr^(addr+(mbx_size * entry_size))) & ~0xffff)
		addr = (addr & ~0xffff)+0x10000;

	zeitnet->mbx[mbx_num] = (PULONG)addr;
	zeitnet->mbx_end[mbx_num] = (PULONG)((addr + mbx_size*entry_size) & 0xffff);
	zeitnet->mbx_next[mbx_num] = zeitnet->mbx_dirty[mbx_num] = (PULONG)addr;

  paddr =	zeitnet->p_mbx[mbx_num] = MmGetPhysicalAddress(addr).LowPart;
	zout(paddr>>16, MSH(mbx_num));
	zout(paddr, MSL(mbx_num));
	zout((paddr + mbx_size*entry_size)&0xffff, MBA(mbx_num));
	zout(paddr & 0xffff, MTA(mbx_num));
	zout(paddr & 0xffff, MWA(mbx_num));

	return STATUS_SUCCESS;
}


static NTSTATUS
InitDevice(
	   IN PDRIVER_OBJECT driverObject,
	   IN int devIndex,
	   IN int pciIndex,
	   IN PPCI_COMMON_CONFIG pciConfig,
	   IN PUNICODE_STRING registryPath
	   )
{
  PZEITNET_DEV zeitnet;
  PDEVICE_OBJECT deviceObject;
  PCM_RESOURCE_LIST resources;
  int i, j, k;
  int resourceNum, resourceSet;
  PCM_PARTIAL_RESOURCE_DESCRIPTOR resourceDescr;
  int irqLevel, irqMode, irqAffinity;
  long irqVector;
  ULONG vector;
  KIRQL irql;
  KAFFINITY affinity;
  NTSTATUS stat;

  /* Allocate a device structure */
  stat = IoCreateDevice(driverObject, sizeof(ZEITNET_DEV), 0,
			FILE_DEVICE_UNKNOWN, 0, FALSE, &deviceObject);
  if(stat != STATUS_SUCCESS) return stat;
  zeitnet = (PZEITNET_DEV)deviceObject->DeviceExtension;
  zeitnetDev[devIndex] = zeitnet;
  zeitnet->index = devIndex;
  zeitnet->registryPath = registryPath;
  zeitnet->deviceObject = deviceObject;
	zeitnet->adapter_created = FALSE;

  /* Print some device info */
  DbgPrint("Zeitnet: vendor=0x%x device=0x%x revision=0x%x cacheLine=%d\n"
	 "latency=%d baseClass=%d subClass=%d headerType=%d\n",
	 pciConfig->VendorID, pciConfig->DeviceID, pciConfig->RevisionID,
	 pciConfig->CacheLineSize, pciConfig->LatencyTimer, pciConfig->BaseClass,
	 pciConfig->SubClass, pciConfig->HeaderType);
    	
  /* Get the required bus resources */
  stat = HalAssignSlotResources(registryPath, NULL, driverObject,
				NULL, PCIBus, 0, pciIndex, &resources);
  if (stat != STATUS_SUCCESS) {
    DbgPrint("Zeitnet: Unable to assign slot resources.\n");
    IoDeleteDevice(deviceObject);
    zeitnetDev[devIndex] = NULL;
    return stat;
  }

  /* Parse what we've gotten... (you'd think this would be simple!) */
  ASSERT(resources->Count == 1 && resources->List);
  resourceNum = resources->List[0].PartialResourceList.Count;
  resourceDescr = resources->List[0].PartialResourceList.PartialDescriptors;
  resourceSet = 0;
  for (i=0; i<resourceNum; i++,resourceDescr++) {
    PHYSICAL_ADDRESS physAddr;
    ULONG addrSpace = 0x0;
    PRINTK("Zeitnet resource %d (%s): ", i,
					 resourceDescr->ShareDisposition == CmResourceShareDeviceExclusive ?
					 "(dev exclusive)" : resourceDescr->ShareDisposition ==
					 CmResourceShareDriverExclusive ? "(drvr exclusive)" :
					 "(shared)");
    switch (resourceDescr->Type) {
    case CmResourceTypePort:
      PRINTK("port 0x%x..0x%x (%s)\n",
						 resourceDescr->u.Port.Start.LowPart,
						 resourceDescr->u.Port.Start.LowPart
						 + resourceDescr->u.Port.Length - 1,
						 resourceDescr->Flags == CM_RESOURCE_PORT_MEMORY ? "mem" : "io");
      zeitnet->cbio = resourceDescr->u.Port.Start.LowPart;
      resourceSet |= 1<<CmResourceTypePort;
      break;
    case CmResourceTypeInterrupt:
      PRINTK("interrupt level %d vector=0x%x affinity=%d\n",
						 resourceDescr->u.Interrupt.Level,
						 resourceDescr->u.Interrupt.Vector,
						 resourceDescr->u.Interrupt.Affinity);
      irqLevel = resourceDescr->u.Interrupt.Level;
      irqVector = resourceDescr->u.Interrupt.Vector;
      irqAffinity = resourceDescr->u.Interrupt.Affinity;
      irqMode =
				resourceDescr->Flags == CM_RESOURCE_INTERRUPT_LATCHED ?
				Latched : LevelSensitive;
      resourceSet |= 1<<CmResourceTypeInterrupt;
      break;
    case CmResourceTypeMemory:
      PRINTK("memory 0x%x..0x%x (%s)\n",
						 resourceDescr->u.Memory.Start.LowPart,
						 resourceDescr->u.Memory.Start.LowPart
						 + resourceDescr->u.Memory.Length - 1,
						 resourceDescr->Flags == CM_RESOURCE_MEMORY_READ_WRITE ? "rw" :
						 resourceDescr->Flags == CM_RESOURCE_MEMORY_READ_ONLY ? "r":"w");
			if (!HalTranslateBusAddress(PCIBus, 0, resourceDescr->u.Memory.Start,
																	&addrSpace, &physAddr)) {
				DbgPrint("Zeitnet: Bus address translation failed.\n");
				break;
			}
			zeitnet->cbma = MmMapIoSpace(physAddr, resourceDescr->u.Memory.Length, FALSE);
			if (zeitnet->cbma == 0) {
				DbgPrint("Zeitnet: Failed to map CSRs.\n");
				break;
			}
			PRINTK("Zeitnet: memory-mapped CSRs at 0x%x\n", zeitnet->cbma);
			resourceSet |= 1<<CmResourceTypeMemory;
      break;
    case CmResourceTypeDma:
      PRINTK("DMA channel %d\n",
						 resourceDescr->u.Dma.Channel);
      break;
    case CmResourceTypeDeviceSpecific:
      PRINTK("Device specific\n");
      break;
    default:
      DbgPrint("Zeitnet: Unknown resource!");
      break;
    }
  }
  ExFreePool(resources);
  
  /* Verify we got all the resources we need */
  if (resourceSet != (int)((1<<CmResourceTypePort) |
													 (1<<CmResourceTypeInterrupt) |
													 (1<<CmResourceTypeMemory))) {
    DbgPrint("Zeitnet: missing PCI resource(s).\n");
    IoDeleteDevice(zeitnetDev[devIndex]->deviceObject);
    zeitnetDev[devIndex] = NULL;
    return STATUS_INSUFFICIENT_RESOURCES;
  }
	
  // Check out PCI configuration make by HAL
  {
    PCI_COMMON_CONFIG pciConfig;
    LONG l;
    l = HalGetBusData(PCIConfiguration, 0, pciIndex, &pciConfig, 24);
    if(l != 24) {
      PRINTK("Get bus data failed: only got %d\n", l);
    } else {
      PRINTK("PCI command register is: 0x%x\n", pciConfig.Command);
      DbgPrint("PCI latency timer is: %d\n", pciConfig.LatencyTimer);
      PRINTK("PCI I/O base address is: 0x%x\n",
						 pciConfig.u.type0.BaseAddresses[0]);
      PRINTK("PCI memory base address is: 0x%x\n",
						 pciConfig.u.type0.BaseAddresses[1]);
    }

    // Check latency timer and make it at least 100 cycles
    if (pciConfig.LatencyTimer < 100) {
      DbgPrint("Zeitnet: Should set latency timer from %d to 100 clocks.\n",
	     pciConfig.LatencyTimer);
    }
	}

	/* Initialize some fields */
	if (pciConfig->DeviceID == PCI_DEVICE_ID_ZEITNET_1225)
		zeitnet->copper = TRUE;
	else
		zeitnet->copper = FALSE;
	
	/* Reset uPD98401 */
	zout(0,SWR);
	while (!(zin(GSR) & 0x01000000)) ;
	zout(0x80000000, GMR);

	/* Get memory size */
	{
		int i, last;
		last = ZN_MAX_CRAM_SIZE;
		for (i = last - 1024; i >= 0; i -= 1024) {
			zpokel(zeitnet, 0x55555555, i);
			if (zpeekl(zeitnet, i) != 0x55555555) {
				last = i;
			} else {
				zpokel(zeitnet,0xAAAAAAAA, i);
				if (zpeekl(zeitnet, i) != 0xAAAAAAAA) {
					last = i;
				} else {
					zpokel(zeitnet, i, i);
				}
			}
		}
		for (i = 0; i < last; i += 1024) {
			if (zpeekl(zeitnet, (ULONG)i) != (ULONG)i) break;
		}
		zeitnet->memsz = (i << 2);
		while (i) zpokel(zeitnet, 0, --i);
	}

	/* Reset again to rebuild memory pointers */
	zout(0, SWR);
	while (!(zin(GSR) & 0x01000000)) ;
	/* Enable 2, 4, and 8-word burst, word parity */
	// DR bit == use external cell FIFO
#ifdef ZEITNET_LOOPBACK
	zout(0x80004f24, GMR);     // Loopback
#else
	zout(0x80000f24, GMR);     // No loopback
#endif
	
	PRINTK("Zeitnet: Detected device, %dKB memory, %s interface.\n",
				 (zeitnet->memsz >> 10),
				 (zeitnet->copper)?("copper UTP5"):("MMF"));

	// XXX Initialize PHY
	InitPhy(zeitnet);

  /* Install interrupt service routine */
  vector = HalGetInterruptVector(PCIBus, 0, irqLevel, irqVector,
				 &irql, &affinity);
  stat = IoConnectInterrupt(&zeitnet->interruptObject, ZeitnetISR,
														zeitnet, NULL, vector, irql, irql, irqMode,
														FALSE, affinity, FALSE);
  if (stat != STATUS_SUCCESS) {
    DbgPrint("Zeitnet: Can't allocate IRQ.\n");
    IoDeleteDevice(zeitnetDev[devIndex]->deviceObject);
    zeitnetDev[devIndex] = NULL;
    return stat;
  }
  
  // Register driver with the U-Net interface
  {
    PUNET_ADAPTER adapter;
    adapter = UnetCreateAdapterObject(UNET_DEVID_ZEITNET, devIndex);
    if (adapter == NULL) {
      DbgPrint("Zeitnet: Can't get adapter object\n");
			IoDeleteDevice(zeitnetDev[devIndex]->deviceObject);
      zeitnetDev[devIndex] = NULL;
      return STATUS_NO_MEMORY;
    }
    ZeitnetFillAdapter(adapter);
		zeitnet->adapter_created = TRUE;
  }

	/* Initialize control memory and pointers */
	{
	 int pools, vccs, ld, rx;
	 ULONG curr;
	 pools = ZEITNET_NUM_POOLS;
	 if (ZN_SHAPER_SIZE * ZEITNET_NUM_SHAPERS > pools * ZN_POOL_DESCR_SIZE) {
		 pools = ZN_SHAPER_SIZE*ZEITNET_NUM_SHAPERS/ZN_POOL_DESCR_SIZE;
	 }
	 vccs = (zeitnet->memsz -
					 (ZN_SHAPER_SIZE * ZEITNET_NUM_SHAPERS) -
					 (ZN_POOL_DESCR_SIZE * pools)) /
		 (2 * ZN_VC_DESCR_SIZE + ZN_RX_LOOKUP_SIZE);
	 // XXX Hopefully vccs is the max number of channels provided by the
	 // uPD98401
	 zeitnet->max_conns = vccs; 
	 
	 ld = -1;
	 for (rx = 1; rx < vccs; rx <<= 1) ld++;

	 curr = rx * ZN_RX_LOOKUP_SIZE / 4;  // size of VC lookup table, in words
	 zpokel(zeitnet, curr, ZN_PMA);      // loc of pool descr pointers
	 zeitnet->pool_base = curr;

	 curr += pools * ZN_POOL_DESCR_SIZE / 4; // size of pool descr table
	 zpokel(zeitnet, curr, ZN_SMA);  // loc of shapers pointers

	 curr += ZEITNET_NUM_SHAPERS * ZN_SHAPER_SIZE / 4; // size of shapers
	 zpokel(zeitnet, curr, ZN_TOS);  // top of stack
	}

	/* Create mailboxes */
	if (InitMbx(zeitnet, ZN_MBX_RX_0, ZEITNET_MBX_RX_0_SIZE, ZN_RX_MBX_SIZE) ||
			InitMbx(zeitnet, ZN_MBX_RX_1, ZEITNET_MBX_RX_1_SIZE, ZN_RX_MBX_SIZE) ||
			InitMbx(zeitnet, ZN_MBX_TX_0, ZEITNET_MBX_TX_0_SIZE, ZN_TX_MBX_SIZE) ||
			InitMbx(zeitnet, ZN_MBX_TX_1, ZEITNET_MBX_TX_1_SIZE, ZN_TX_MBX_SIZE)) {
		DbgPrint("Zeitnet: Can't allocate mailboxes.\n");
		IoDeleteDevice(zeitnetDev[devIndex]->deviceObject);
    zeitnetDev[devIndex] = NULL;
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	/* Initialize shapers */
	for (i = 0; i < ZEITNET_NUM_SHAPERS; i++) {
		zpokel(zeitnet, 0, ZN_SCHED_pristat + i);
		zeitnet->shaper[i].refCnt = 0;
	}
	
	/* XXX mdw, 5-Dec-1996: Strangeness.
	 * It seems that if we set the ratio of I/M to 1, there's a serious
	 * problem with DMA starvation, at least when a loopback fiber is used.
	 * (Curiously, not with internal loopback!) The symptom is that an Rx packet
	 * will 'timeout' as the T1 timer rolls over, and Rx error 0x700 will be reported.
	 * This is very easy to reproduce: Do a loopback of 1000 packets of size 1450
	 * bytes (say). The timeout seems to occur mainly when a prefetch cycle is
	 * taking place; however, I was unable to determine the race condition there.
	 * It is possible that polling chip register (e.g., zwait) prevents DMA cycles;
	 * in this case change zwait(0) to zwait(1) to cause 1ms delays between polls.
	 * However, I was able to reproduce by polling for Rx immediately after Tx
	 * (even with pauses between polls). My guess it that Tx or Rx DMA is being cut
	 * short by the high rate of Tx shaping.
	 * The following setting seems to work for packets up to 1500 bytes, not
	 * tested further. Also note that there is the newer rev of the chip (uPD98401A)
	 * which has 'better DMA performance' and a deeper Rx FIFO.
	 */
	
	/* Set up UBR shaper at shaper #0 */
	zeitnet->shaper[0].I = ZEITNET_DEFAULT_I;
	zeitnet->shaper[0].M = ZEITNET_DEFAULT_M;
	zeitnet->shaper[0].P = ZEITNET_DEFAULT_P;
	zeitnet->shaper[0].C = ZEITNET_DEFAULT_C;
	zeitnet->shaper[0].refCnt = 2; // Always keep this one 
	InitShaper(zeitnet, 0,
						 zeitnet->shaper[0].I,
						 zeitnet->shaper[0].M,
						 zeitnet->shaper[0].P,
						 zeitnet->shaper[0].C);
	
	zeitnet->tx_conn_map = ExAllocatePool(NonPagedPoolCacheAligned,
																				zeitnet->max_conns *
																				sizeof(PZEITNET_CHAN));
	if (!zeitnet->tx_conn_map) {
		DbgPrint("Zeitnet: Can't allocate tx_conn_map for %d connections.\n",
						 zeitnet->max_conns);
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	memset(zeitnet->tx_conn_map, 0, zeitnet->max_conns * sizeof(PZEITNET_CHAN));
																				
	/* Initialize Rx, buffer pools, VCC map, VCI mapreg, etc. */
	zeitnet->rx_conn_map = ExAllocatePool(NonPagedPoolCacheAligned,
																				zeitnet->max_conns *
																				sizeof(PZEITNET_CHAN));
	if (!zeitnet->rx_conn_map) {
		DbgPrint("Zeitnet: Can't allocate rx_conn_map for %d connections.\n",
						 zeitnet->max_conns);
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	memset(zeitnet->rx_conn_map, 0, zeitnet->max_conns * sizeof(PZEITNET_CHAN));

	memset(zeitnet->pool_inuse, 0, ZEITNET_NUM_POOLS * sizeof(BOOLEAN));


	
	zout(0x0000ffff, VRR); // No VPI bits, all 16 VCI bits

	/* Enable interrupts and Tx/Rx */
	zout(0xffffff33, IMR); 
	zout(zin(GMR) | 0x03, GMR);

	return STATUS_SUCCESS;
}

void
ReleaseZeitnet(
	     IN int devIndex
	     )
{
  NTSTATUS stat;

  PRINTK("Zeitnet: releasing device %d\n", devIndex);

  if (zeitnetDev[devIndex] != NULL) {
		PZEITNET_DEV zeitnet = zeitnetDev[devIndex];

		if (zeitnet->adapter_created)
			UnetDestroyAdapterObject(UNET_DEVID_ZEITNET, devIndex);
    if (zeitnet->interruptObject)
      IoDisconnectInterrupt(zeitnet->interruptObject);
    stat = IoAssignResources(zeitnet->registryPath, NULL,
			     zeitnetDriverObject, NULL, NULL, NULL);
    if (stat != STATUS_SUCCESS)
      PRINTK("Zeitnet: IoAssignResource failed\n");
    if (zeitnet->deviceObject != NULL) {
      IoDeleteDevice(zeitnet->deviceObject);
    }
    zeitnetDev[devIndex] = NULL;
  }
}

static void
InitShaper(PZEITNET_DEV zeitnet, int shaper, UCHAR I, ULONG M, UCHAR P, UCHAR C) {

	zpokel(zeitnet, (I) << 24 | (M), ZN_SCHED_IM+shaper);
	zpokel(zeitnet, (P) << 24 | (C) << 16, ZN_SCHED_PCpc+shaper);
	zpokel(zeitnet, 0, ZN_SCHED_X+shaper);
	zpokel(zeitnet, 0, ZN_SCHED_Y+shaper);
	
	// XXX Right now all shapers have priority 0 (highest)
	zpokel(zeitnet, 0x01, ZN_SCHED_pristat+shaper); // enable
}

/* UDI entry points *******************************************************/

static NTSTATUS
ZeitnetAdapterInfo(
									 OUT PUNET_INFO_RESPONSE resp
									 )
{
  resp->version = UNET_DEV_VERSION;
	// Small messages always use buffers, not Rxd's
  resp->rxd_payload_size = 0; 
  resp->max_pdu = ZEITNET_MAX_PDU;
  resp->max_endpts = ZEITNET_MAX_ENDPT;
  resp->max_chans = ZEITNET_MAX_CHANS;

  resp->tx_buf_align = 2;
  resp->rx_buf_align = 2;

  resp->min_tx_fifo = ZEITNET_MIN_TX;
  resp->max_tx_fifo = ZEITNET_MAX_TX;
  resp->min_free_fifo = ZEITNET_MIN_FREE;
  resp->max_free_fifo = ZEITNET_MAX_FREE;
  resp->min_rx_fifo = ZEITNET_MIN_RX;
  resp->max_rx_fifo = ZEITNET_MAX_RX;

  return STATUS_SUCCESS;
}

#define CHK(val, min, max) do { \
  if(req->val < (min) || req->val > (max)) {\
     PRINTK("ZeitnetCreateEndpoint: illegal %s (%d)\n", #val, req->val);\
     return STATUS_INVALID_PARAMETER;\
  }\
  endpt->val = req->val;\
} while(0)
     
static NTSTATUS
ZeitnetCreateEndpoint(
											IN UNET_ENDPOINT_ID endptId,
											IN PUNET_CREATE_ENDPOINT_REQUEST req,
											OUT PUNET_CREATE_ENDPOINT_RESPONSE resp
											)
{
  PZEITNET_ENDPT endpt;
  PZEITNET_DEV zeitnet;
  PMDL mdl;
  int i;
	int size, used_size, lost_bytes, num_full_cells, pool_size, pool_size_log2;

  PRINTK("Zeitnet: CreateEndpoint %d 0x%x 0x%x\n", endptId, req, resp);

  if(endptId < 0 || endptId > ZEITNET_MAX_ENDPT || zeitnetEndpt[endptId] != 0)
    return STATUS_INVALID_PARAMETER;
  if(req->devIndex < 0 || req->devIndex > ZEITNET_MAX_DEVS ||
     zeitnetDev[req->devIndex] == 0)
    return STATUS_INVALID_PARAMETER;

	zeitnet = zeitnetDev[req->devIndex];

  // allocate endpoint structure
  endpt = ExAllocatePool(NonPagedPoolCacheAligned, sizeof(ZEITNET_ENDPT));
  if(endpt == 0) {
    PRINTK("Zeitnet: can't alloc endpt struct.\n");
    return STATUS_INSUFFICIENT_RESOURCES;
  }
  endpt->dev = zeitnetDev[req->devIndex];
	endpt->process = IoGetCurrentProcess();
	endpt->num_freed = 0;
	endpt->index = endptId;
  zeitnetEndpt[endptId] = endpt;

  // check sizes
  CHK(numTxFifoEntries, ZEITNET_MIN_TX, ZEITNET_MAX_TX);
  CHK(numFreeFifoEntries, ZEITNET_MIN_FREE, ZEITNET_MAX_FREE);
  CHK(numRxFifoEntries, ZEITNET_MIN_RX, ZEITNET_MAX_RX);
  CHK(bufferAreaSize, 0, 5*1024*1024); // max due to bug in MmInitialize
	// XXX Right now, restrict Rx buffers to fit within one page.
	// Larger packets can be received, across multiple batches.
	CHK(rxBufferSize, 0, PAGE_SIZE);

  // allocate buffer area
	if (req->bufferAreaSize) {
		endpt->bufferArea =
			MmAllocateContiguousMemory(endpt->bufferAreaSize, topPhysAddr);
		if(endpt->bufferArea == NULL) {
			PRINTK("Zeitnet: Can't allocate fixed buffer area.\n");
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		endpt->physBufferArea =
			(void *)(MmGetPhysicalAddress(endpt->bufferArea).LowPart);
		RtlZeroMemory(endpt->bufferArea, endpt->bufferAreaSize);
		mdl = MmCreateMdl(0, endpt->bufferArea, endpt->bufferAreaSize);
		MmBuildMdlForNonPagedPool(mdl);
		endpt->userBufferArea = (long)MmMapLockedPages(mdl, UserMode);
		resp->bufferArea = (void *)endpt->userBufferArea;
		ExFreePool(mdl);
		PRINTK("Zeitnet: locked buffer area @0x%08x\n", resp->bufferArea);
	}
	
  // allocate queues
  endpt->fifoAreaSize = 3*sizeof(void *) +
    endpt->numTxFifoEntries*sizeof(UNET_TX_BUFFER_DESCRIPTOR) +
    endpt->numFreeFifoEntries*sizeof(UNET_RECV_BUFFER_DESCRIPTOR) +
    endpt->numRxFifoEntries*sizeof(UNET_RECEIVE_DESCRIPTOR);
  endpt->txWall = ExAllocatePool(NonPagedPool, endpt->fifoAreaSize);
  if(endpt->txWall == NULL) {
    PRINTK("Zeitnet: Can't allocate fifos.\n");
    return STATUS_INSUFFICIENT_RESOURCES;
  }
  RtlZeroMemory(endpt->txWall, endpt->fifoAreaSize);
  mdl = MmCreateMdl(0, endpt->txWall, endpt->fifoAreaSize);
  MmBuildMdlForNonPagedPool(mdl);
  endpt->userFifoArea = (long)MmMapLockedPages(mdl, UserMode) +
    ((ULONG)endpt->txWall & (PAGE_SIZE-1));
  resp->txWall = (void *)endpt->userFifoArea;
  ExFreePool(mdl);
  PRINTK("Zeitnet: locked fifo area @0x%08x\n", resp->txWall);

  // clear channels
  endpt->numChans = 0;
  for(i=0; i<ZEITNET_MAX_CHANS; i++) {
    endpt->chans[i].inUse = 0;
  }

  // init other endpt pointers
  endpt->freeWall = (long*)(endpt->txWall+1);
  endpt->rxWall = (long*)(endpt->freeWall+1);
  endpt->txFirst = (UNET_TX_BUFFER_DESCRIPTOR*)(endpt->rxWall+1);
  endpt->txNext = endpt->txFirst;
  endpt->txLast = endpt->txDirty = endpt->txFirst + endpt->numTxFifoEntries-1;
  endpt->freeFirst = (UNET_RECV_BUFFER_DESCRIPTOR*)(endpt->txLast+1);
  endpt->freeNext = endpt->freeFirst;
  endpt->freeLast = endpt->freeFirst + endpt->numFreeFifoEntries-1;
  endpt->freePrefetchNext = endpt->freeFirst;
	endpt->freePrefetch = 0;
  endpt->rxFirst = (UNET_RECEIVE_DESCRIPTOR*)(endpt->freeLast+1);
  endpt->rxNext = endpt->rxFirst;
  endpt->rxLast = endpt->rxFirst + endpt->numRxFifoEntries-1;

	// Compute bytes of buffer actually used
	// XXX mdw: This might not be entirely correct. If Rx buffers seem to be
	// hosed on reception, check out the code below and be sure things are being
	// computed correctly.
	size = req->rxBufferSize;
	if (size % 64) {
		PRINTK("Zeitnet: RxBufferSize %d not a multiple of 64.\n", size);
		return STATUS_INVALID_PARAMETER; // XXX Cleanup here
	}
	pool_size = (size / 64);
	pool_size_log2 = 0;
	while (!(pool_size & 0x01)) {
		pool_size_log2++; pool_size >>= 1;
	}

	// It seems as though the entire Rx buffer is used, not just portions
	// of it on cell boundaries
	used_size = endpt->rxBufferUsedSize = size; 

  // initialize the Fifos
  for(i=0; i<endpt->numTxFifoEntries; i++) {
    endpt->txFirst[i].chan = UNET_CHANNEL_FREE;
  }
  for(i=0; i<endpt->numRxFifoEntries; i++) {
    endpt->rxFirst[i].chan = UNET_CHANNEL_FREE;
  }
  for(i=0; i<endpt->numFreeFifoEntries; i++) {
    endpt->freeFirst[i].buf = (ULONG)UNET_RXB_FREE;
  }
	
  // initialize the walls
  *endpt->txWall = (ULONG)endpt->txLast - (ULONG)endpt->txFirst;
  *endpt->freeWall = (ULONG)endpt->freeLast - (ULONG)endpt->freeFirst;
  *endpt->rxWall = (ULONG)endpt->rxLast - (ULONG)endpt->rxFirst;

	// Create free buffer pool
	for (endpt->pool_num = 0; endpt->pool_num < ZEITNET_NUM_POOLS; endpt->pool_num++) {
		if (!zeitnet->pool_inuse[endpt->pool_num]) break;
	}
	if (endpt->pool_num == ZEITNET_NUM_POOLS) {
		PRINTK("Zeitnet: No free buffer pools available.\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	zeitnet->pool_inuse[endpt->pool_num] = TRUE;
	
	for (i = 0; i < POOL_LIST_SIZE; i++) {
		endpt->pool[i].paddr = 0;
		endpt->pool[i].p_next =
			(i != POOL_LIST_SIZE-1) ?
			(MmGetPhysicalAddress(&(endpt->pool[i+1])).LowPart) :
			(MmGetPhysicalAddress(&(endpt->pool[0])).LowPart);
	}
	endpt->pool_next = endpt->pool_dirty = &(endpt->pool[0]);
	endpt->pool_last = &(endpt->pool[POOL_LIST_SIZE-1]);

	// One buffer per batch
	zpokel(zeitnet, ((1) << 16) | ((pool_size_log2) << 24),
				 zeitnet->pool_base + (2*endpt->pool_num));
	zpokel(zeitnet, MmGetPhysicalAddress(endpt->pool).LowPart,
				 zeitnet->pool_base + (2*endpt->pool_num) + 1);

	// Default shaper is 0 (UBR)
	endpt->tx_shaper = 0;

  // construct the response
  resp->freeWall = (ULONG*)(resp->txWall+1);
  resp->rxWall = (ULONG*)(resp->freeWall+1);
  resp->txFifo = (UNET_TX_BUFFER_DESCRIPTOR*)(resp->rxWall+1);
  resp->freeFifo = (UNET_RECV_BUFFER_DESCRIPTOR*)(resp->txFifo +
						  endpt->numTxFifoEntries);
  resp->rxFifo = (UNET_RECEIVE_DESCRIPTOR*)(resp->freeFifo +
					    endpt->numFreeFifoEntries);
	
	resp->rxBufferAllocSize = size;
	resp->rxBufferUsedSize = used_size;
	
  return STATUS_SUCCESS;
}

static NTSTATUS
ZeitnetDestroyEndpoint(
		     IN UNET_ENDPOINT_ID endptId
		     )
{
  static NTSTATUS ZeitnetDeactivateChannel(PUNET_DEACTIVATE_CHANNEL_REQUEST req);
  PZEITNET_ENDPT endpt = zeitnetEndpt[endptId];
  int chan;

  PRINTK("Zeitnet: DestroyEndpoint %d\n", endptId);

  if(endptId < 0 || endptId > ZEITNET_MAX_ENDPT)
    return STATUS_INVALID_PARAMETER;
  if(endpt == 0) return STATUS_SUCCESS;

  // destroy all channels
  for(chan=0; chan < ZEITNET_MAX_CHANS; chan++) {
    if(endpt->chans[chan].inUse) {
      UNET_DEACTIVATE_CHANNEL_REQUEST dcreq = {endptId, chan};
      (void)ZeitnetDeactivateChannel(&dcreq);
    }
  }

	// Clear out TLB
	StlbClear(endptId);

	// Disable pool
	endpt->dev->pool_inuse[endpt->pool_num] = FALSE;
	
	// Free Fifos and buffer area
  if(endpt->txWall != 0) {
    PMDL mdl = MmCreateMdl(0, endpt->txWall, endpt->fifoAreaSize);
    MmBuildMdlForNonPagedPool(mdl);
    MmUnmapLockedPages((void *)endpt->userFifoArea, mdl);
    ExFreePool(mdl);
    ExFreePool((void *)endpt->txWall);
  }
  if(endpt->bufferAreaSize != 0) {
    PMDL mdl = MmCreateMdl(0, endpt->bufferArea, endpt->bufferAreaSize);
    MmBuildMdlForNonPagedPool(mdl);
    MmUnmapLockedPages((void *)endpt->userBufferArea, mdl);
    ExFreePool(mdl);
    MmFreeContiguousMemory((void *)endpt->bufferArea);
  }
	
  ExFreePool(endpt);
  zeitnetEndpt[endptId] = 0;
  return STATUS_SUCCESS;
}

NTSTATUS
ZeitnetActivateChannel(
		     IN PUNET_ACTIVATE_CHANNEL_REQUEST req,
		     IN PUNET_ACTIVATE_CHANNEL_RESPONSE resp
		     )
{
  int endptId = req->endpointId;
  PZEITNET_ENDPT endpt = zeitnetEndpt[endptId];
  PZEITNET_DEV zeitnet;
	PZEITNET_CHAN chan;
  int channum;
	KIRQL oldIrqL;

  PRINTK("Zeitnet: ActivateChannel %d\n", endptId);

  if(endptId < 0 || endptId > ZEITNET_MAX_ENDPT || zeitnetEndpt[endptId] == 0)
    return STATUS_INVALID_PARAMETER;
	
  zeitnet = endpt->dev;

  // get a free channel
  channum = req->chan;
  if(channum < 0 || channum >= ZEITNET_MAX_CHANS || endpt->chans[channum].inUse) {
    for(channum=0; channum < ZEITNET_MAX_CHANS; channum++) {
      if(endpt->chans[channum].inUse == 0) break;
    }
    if(channum == ZEITNET_MAX_CHANS) {
      DbgPrint("Zeitnet: ActivateChannel ran out of channels.\n");
      return STATUS_INSUFFICIENT_RESOURCES;
    }
  }

	chan = &(endpt->chans[channum]);
	chan->endpt = endpt;
	chan->index = channum;

  // Record addresses
  memcpy(&(chan->txAddr), &(req->txAddr.atm), sizeof(UNET_ATM_ADDRESS));
  memcpy(&(chan->rxAddr), &(req->rxAddr.atm), sizeof(UNET_ATM_ADDRESS));

	// Open Tx connection
	KeRaiseIrql(POWER_LEVEL, &oldIrqL);
	zwait(1);
	zout(0x20000000, CMR);
	zwait(1);
	chan->tx_conn = (zin(CMR) & 0x003fff8) >> 3;
	KeLowerIrql(oldIrqL);
	
	if (!chan->tx_conn) {
		DbgPrint("Zeitnet: Out of Tx connections on chip.\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	zeitnet->tx_conn_map[chan->tx_conn] = chan;
	PRINTK("Zeitnet: Got Tx connection %d.\n", chan->tx_conn);
	
	chan->tx_shaper = endpt->tx_shaper;
	zeitnet->shaper[chan->tx_shaper].refCnt++;

	chan->tx_ring = MmAllocateContiguousMemory((ZEITNET_TXRING_SIZE+1)*ZN_TX_DESCR_SIZE, topPhysAddr);
	if (!chan->tx_ring) {
		DbgPrint("Zeitnet: Can't allocate Tx Ring.\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	PRINTK("Zeitnet: tx_ring allocated at 0x%lx\n", chan->tx_ring);
	chan->tx_ring_phys = MmGetPhysicalAddress(chan->tx_ring).LowPart;
	
	memset(chan->tx_ring, 0, (ZEITNET_TXRING_SIZE+1)*ZN_TX_DESCR_SIZE);
	{
		ZN_TX_DESCR *linkptr;
		linkptr = chan->tx_ring + ZEITNET_TXRING_SIZE;
		linkptr->status = 0x80000000;
		linkptr->pad = linkptr->size = 0;
		linkptr->addr = (ULONG)chan->tx_ring_phys;
	}
	chan->tx_ring_next = chan->tx_ring;
	chan->tx_ring_last = chan->tx_ring + (ZEITNET_TXRING_SIZE-1);
	chan->tx_ring_dirty = chan->tx_ring;
	chan->tx_ring_full = FALSE;

	chan->tx_ring_shadow = ExAllocatePool(NonPagedPoolCacheAligned,
																				(ZEITNET_TXRING_SIZE)*
																				sizeof(ZEITNET_TX_SHADOW));
	if (!chan->tx_ring_shadow) {
		DbgPrint("Zeitnet: Can't allocate Tx ring shadow.\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	KeRaiseIrql(POWER_LEVEL, &oldIrqL);
	zpokel(zeitnet, 0, chan->tx_conn * ZN_VC_DESCR_SIZE/4);
	zpokel(zeitnet, (0x80000000 | (chan->tx_shaper << 23) | (chan->txAddr.vci)),
				 chan->tx_conn * ZN_VC_DESCR_SIZE/4 + 1);
	zpokel(zeitnet, 0, chan->tx_conn * ZN_VC_DESCR_SIZE/4 + 2);
	zpokel(zeitnet, (ULONG)chan->tx_ring_phys,
				 chan->tx_conn * ZN_VC_DESCR_SIZE/4 + 6);
	KeLowerIrql(oldIrqL);

	// Open Rx connection
	KeRaiseIrql(POWER_LEVEL, &oldIrqL);
	zwait(1);
	zout(0x20000000, CMR);
	zwait(1);
	chan->rx_conn = (zin(CMR) & 0x003fff8) >> 3;
	KeLowerIrql(oldIrqL);
	
	if (!chan->rx_conn) {
		DbgPrint("Zeitnet: Out of Rx connections on chip.\n");
		// XXX Need to clean up here
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	zeitnet->rx_conn_map[chan->rx_conn] = chan;
	PRINTK("Zeitnet: Got Rx connection %d.\n", chan->rx_conn);

	KeRaiseIrql(POWER_LEVEL, &oldIrqL);
	zpokel(zeitnet, (endpt->pool_num) << 16, chan->rx_conn * ZN_VC_DESCR_SIZE/4);
	zpokel(zeitnet, (1<<12) | (1<<11) | 1366, // Max AAL5 PDU
				 chan->rx_conn * ZN_VC_DESCR_SIZE/4 + 1);
	zpokel(zeitnet, 0UL, chan->rx_conn * ZN_VC_DESCR_SIZE/4 + 2);

	// Set Rx LUT entry
	{
		ULONG pos, shift;
		pos = (chan->rxAddr.vci) >> 1;
		shift = (chan->rxAddr.vci & 0x01) ? 0:16;
		zpokel(zeitnet, (zpeekl(zeitnet, pos) & ~(0xffff << shift)) |
					 ((chan->rx_conn | 0x8000) << shift), pos);
	}
	KeLowerIrql(oldIrqL);
	
	/* Start up a buffer-prefetch cycle for the endpoint */
	if (endpt->freePrefetch <= START_PREFETCH_THRESH) PfStart(zeitnet, endptId);
			
	endpt->numChans++;
	chan->inUse = 1;
  resp->chan = channum;
  return STATUS_SUCCESS;
}

static NTSTATUS
ZeitnetDeactivateChannel(
		       IN PUNET_DEACTIVATE_CHANNEL_REQUEST req
		       )
{
  int endptId = req->endpointId;
  PZEITNET_ENDPT endpt = zeitnetEndpt[endptId];
  PZEITNET_DEV zeitnet;
	PZEITNET_CHAN chan;
	int num_spins = 0;
	KIRQL oldIrqL;

  PRINTK("Zeitnet: DeactivateChannel endpt=%d chan=%d\n",
	 endptId, req->chan);

  if(endptId < 0 || endptId > ZEITNET_MAX_ENDPT || zeitnetEndpt[endptId] == 0)
    return STATUS_INVALID_PARAMETER;
  if(req->chan < 0 || req->chan >= ZEITNET_MAX_CHANS)
    return STATUS_INVALID_PARAMETER;
	
  zeitnet = endpt->dev;

	chan = &(endpt->chans[req->chan]);

	// Wait for Tx to drain on this channel
	num_spins = 0;
	while ((chan->tx_ring_full) || (chan->tx_ring_dirty != chan->tx_ring_next)) {
		ZeitnetUpdateTxwall(zeitnet, endpt, TRUE);
		if (num_spins++ > 1000) {
			DbgPrint("Zeitnet: Timed out waiting for Tx entries to drain on %d:%d.\n",
							 endptId, req->chan);
			break;
		}
	}

	KeRaiseIrql(POWER_LEVEL, &oldIrqL);
	zwait(1);
	zout(0x24000000 | (chan->tx_conn << 3), CMR);
	zwait(1);
	if (!(zin(CMR) & 0x0003fff8)) {
		PRINTK("Zeitnet: Can't close chip Tx connection?\n");
	}

	zwait(1);
	zout(0x24000000 | (chan->rx_conn << 3), CMR);
	zwait(1);
	if (!(zin(CMR) & 0x0003fff8)) {
		PRINTK("Zeitnet: Can't close chip Rx connection?\n");
	}
	KeLowerIrql(oldIrqL);

	zeitnet->shaper[chan->tx_shaper].refCnt--;

  chan->inUse = 0;

	// XXX Need to mark Rx VC as free 

  return STATUS_SUCCESS;
}

static NTSTATUS
ZeitnetAdapterIoctl(  
											IN PUNET_ADAPTER_IOCTL_REQUEST req,
											OUT PUNET_ADAPTER_IOCTL_RESPONSE resp
											)
{

	int endptId = req->endpointId;
  PZEITNET_ENDPT endpt = zeitnetEndpt[endptId];
  PZEITNET_DEV zeitnet = endpt->dev;
	
	PRINTK("Zeitnet: AdapterIoctl called, req endpt %d, code 0x%lx, len %d, buf 0x%lx\n",
				 req->endpointId, req->code, req->bufferSize, req->buffer);
	PRINTK("Zeitnet: AdapterIoctl response len %d, buf 0x%lx\n",
				 resp->bufferSize, resp->buffer);

	switch (req->code) {
	case UNET_ZEITNET_IOCTL_TRAFFICSHAPING: {
		// Set traffic parameters for future channels
		UNET_ZEITNET_IOCTL_TRAFFICSHAPING_REQUEST *tshape;
		UNET_ZEITNET_IOCTL_TRAFFICSHAPING_RESPONSE *rshape;
		int shaper_num, i;
		
		if (req->bufferSize != sizeof(UNET_ZEITNET_IOCTL_TRAFFICSHAPING_REQUEST))
			return STATUS_BUFFER_TOO_SMALL;
		if (req->bufferSize != sizeof(UNET_ZEITNET_IOCTL_TRAFFICSHAPING_REQUEST))
			return STATUS_BUFFER_TOO_SMALL;
		tshape = req->buffer; rshape = resp->buffer;

		// Find shaper with these parameters
		for (i = 0; i < ZEITNET_NUM_SHAPERS; i++) {
			if ((zeitnet->shaper[i].refCnt) &&
					(zeitnet->shaper[i].I == tshape->I) &&
					(zeitnet->shaper[i].M == tshape->M) &&
					(zeitnet->shaper[i].P == tshape->P) &&
					(zeitnet->shaper[i].C == tshape->C)) break;
		}
		if (i == ZEITNET_NUM_SHAPERS) {
			// Ok, try to allocate a new shaper
			for (i = 0; i < ZEITNET_NUM_SHAPERS; i++) {
				if (zeitnet->shaper[i].refCnt == 0) break;
			}
			if (i == ZEITNET_NUM_SHAPERS) {
				DbgPrint("Zeitnet: All shapers allocated, can't complete request.\n");
				return STATUS_INSUFFICIENT_RESOURCES;
			}
			zeitnet->shaper[i].I = tshape->I;
			zeitnet->shaper[i].M = tshape->M;
			zeitnet->shaper[i].P = tshape->P;
			zeitnet->shaper[i].C = tshape->C;
			InitShaper(zeitnet, i, tshape->I, tshape->M, tshape->P, tshape->C);
		}
			
		zeitnet->shaper[i].refCnt++;
		
		rshape->I = zeitnet->shaper[endpt->tx_shaper].I;
		rshape->M = zeitnet->shaper[endpt->tx_shaper].M;
		rshape->P = zeitnet->shaper[endpt->tx_shaper].P;
		rshape->C = zeitnet->shaper[endpt->tx_shaper].C;
		zeitnet->shaper[endpt->tx_shaper].refCnt--;
		
		PRINTK("Zeitnet: De-allocated shaper %d (%d %d %d %d), allocated %d (%d %d %d %d)\n",
					 endpt->tx_shaper,
					 rshape->I, rshape->M, rshape->P, rshape->C,
					 i,
					 tshape->I, tshape->M, tshape->P, tshape->C);
		
		endpt->tx_shaper = i;
		break;
	}
	default:
		DbgPrint("Zeitnet: Ioctl code 0x%lx not recognized.\n", req->code);
		return STATUS_INVALID_PARAMETER;
		break;
	}
	
	return STATUS_SUCCESS;
	
}

/* Zeitnet Stlb routines *****************************************************/

// Do a prefetch cycle on this endpoint
static void PfStart(PZEITNET_DEV zeitnet, UNET_ENDPOINT_ID eid) {
	int i;
	PZEITNET_ENDPT endpt = zeitnetEndpt[eid];
	ULONG vaddr, paddr, voffset, saddr;
	int batches = 0;
	ULONG first = 0UL;
	int pf = endpt->freePrefetch;

	PRINTK("Zeitnet: PfStart called on endpoint 0x%x, freePf = %d\n",eid, pf);
	
	while (pf < NUM_PREFETCH) {
		if (endpt->pool_next->paddr) break; // Batches can be consumed out-of-order
		if (RXB_ISFREE(endpt->freePrefetchNext->buf)) break;
		vaddr = PAGE_ADDR(endpt->freePrefetchNext->buf);
		voffset = PAGE_OFFSET(endpt->freePrefetchNext->buf);

		// XXX In 'goto next_pf' cases, user Free FIFO should be bumped: Record
		// number of pages skipped here for next Rx cycle.
		
		if (voffset + endpt->rxBufferSize > PAGE_SIZE) {
			// Buffer crosses a page, just reject it.
			DbgPrint("Zeitnet: PfStart address 0x%lx crosses a page, rejecting.\n",
							 vaddr | voffset);
			// XXX Should push buffer onto user Rx FIFO here
			goto next_pf;
		}
		
		if (!(paddr = StlbTranslate(eid, vaddr, UNET_STLB_PROT_RW, TRUE, &saddr))) {
			PRINTK("Zeitnet: PfStart can't write-translate 0x%x:0x%lx\n", eid, vaddr);
			goto next_pf;
		}
		PRINTK("Zeitnet: Prefetch 0x%x:0x%lx -> 0x%lx\n",
					 eid, vaddr, paddr);
		pf++;

		endpt->pool_next->paddr = paddr | voffset;
		endpt->pool_next->vaddr = vaddr | voffset;
		endpt->pool_next->saddr = saddr | voffset;
		if (!batches) first = MmGetPhysicalAddress(endpt->pool_next).LowPart;
		batches++;
		endpt->pool_next = (endpt->pool_next == endpt->pool_last) ?
			(endpt->pool):(endpt->pool_next+1);

	next_pf:
		endpt->freePrefetchNext = (endpt->freePrefetchNext == endpt->freeLast)?(endpt->freeFirst):(endpt->freePrefetchNext+1);
	}

	if (batches) {
		KIRQL oldIrqL;
		PRINTK("Zeitnet: Pushing %d batches to free pool.\n", batches);
		// Signal new batches added
		KeRaiseIrql(POWER_LEVEL, &oldIrqL);
		endpt->freePrefetch += batches;
		zwait(0);
		zout((ULONG)first, CER);
		zout(0x34000000 | (endpt->pool_num << 16) | batches, CMR);
		zwait(0);
		KeLowerIrql(oldIrqL);
	}
	
}

static BOOLEAN
ZeitnetISR (
	  IN PKINTERRUPT Interrupt, 
	  IN PVOID ServiceContext
	  )
{
  ZEITNET_DEV *zeitnet = (PZEITNET_DEV)ServiceContext;
	ULONG gsr = zin(GSR);
	
  PRINTK("Zeitnet: INTERRUPT, GSR = 0x%8.8lx\n", gsr);
	
	if (gsr & ~0x05) {
		// Some error
		PRINTK("Zeitnet: INTERRUPT error, GSR = 0x%8.8lx\n", gsr);
	}
	if (gsr & 0x01) {
		// Rx MBX modified
		ZeitnetRx(zeitnet);
	}
  return TRUE;
}

/* Zeitnet reception  *********************************************************/

static void
ZeitnetRx(ZEITNET_DEV *zeitnet) {

	int mn = ZN_MBX_RX_0;
	ULONG mwa = zin(MWA(mn));
	ULONG pdirty = (zeitnet->p_mbx[mn] & ~0xffff) | mwa;
	PULONG dirty = (PULONG)((ULONG)zeitnet->mbx[mn] + (pdirty - (ULONG)zeitnet->p_mbx[mn]));
	PULONG pos = zeitnet->mbx_next[mn];
	PVOID tmp;

	while (pos != dirty) {
		ZN_RX_MBX_ENTRY *mbxentry = (ZN_RX_MBX_ENTRY *)pos;
		ULONG conn = mbxentry->conn;
		// XXX Assumes valid channel
		PZEITNET_CHAN chan = zeitnet->rx_conn_map[conn];
		PZEITNET_ENDPT endpt = chan->endpt;
		int len, len_cells, num_bufs, buf_num, tail;
		int num_cells = mbxentry->pkt_size;
		ULONG p_batch = mbxentry->addr;
		ZEITNET_FREE_POOL *batch = endpt->pool_dirty;
		PUCHAR p;

		PRINTK("Zeitnet: Rx on connection %d, freePrefetch=%d.\n", conn,
						 endpt->freePrefetch);
		PRINTK("Zeitnet: 0x%8.8lx 0x%8.8lx 0x%8.8lx 0x%8.8lx\n",
					 *(PULONG)pos, *((PULONG)pos + 1), *((PULONG)pos + 2),
					 *((PULONG)pos + 3));
		
		if (p_batch != MmGetPhysicalAddress(batch).LowPart) {
			ZEITNET_FREE_POOL *node = batch;
			DbgPrint("Zeitnet: Received batch not next dirty!\n");
			
			do {
				if (p_batch == node->p_next) {
					batch = (node == endpt->pool_last)?(endpt->pool):(node+1);
					break;
				}
				node = (node == endpt->pool_last)?(endpt->pool):(node+1);
			} while (node != endpt->pool_dirty);
			if (node == endpt->pool_dirty) {
				DbgPrint("Zeitnet: Didn't find batch pointer 0x%lx in pool ring!\n",
								 p_batch);
				goto drop;
			}
			endpt->pool_dirty = batch; // Reset pool_dirty to this entry
		}
		
	

		if (*((PULONG)pos+3) & (0x1 << 15)) {
			DbgPrint("Zeitnet: Rx error 0x%x\n", (*((PULONG)pos + 3) & 0x0f00));
			goto drop;
		}


		if ((long)endpt->rxNext - (long)endpt->rxFirst == (long)*endpt->rxWall) {
			// Whoops, Rx FIFO is full. Just drop it ...
			PRINTK("Zeitnet: User Rx FIFO full, dropping packet\n");
			goto drop;
		}
	
		len_cells = num_cells * 48; // Complete cells, incl. AAL5 trailer
		num_bufs = (len_cells / endpt->rxBufferUsedSize) + 1;
		tail = len_cells - ((num_bufs-1)*endpt->rxBufferUsedSize);
		
		for (buf_num = 0; buf_num < num_bufs; buf_num++) {
			
			PRINTK("Zeitnet: Batch [%d] V:0x%lx P:0x%lx S:0x%lx\n",
						 buf_num, batch->vaddr, batch->paddr, batch->saddr);			
#ifdef ZEITNET_DEBUG_PRINTPACKETS
			{
				int i;
				PRINTK("Zeitnet: Rx [%d] ", buf_num);
				for (i = 0; i < 50; i++) {
					PRINTK("%2.2x ", *((unsigned char *)batch->saddr + i));
				}
				PRINTK("\n");
			}
#endif
			
			if (buf_num == num_bufs-1) {
				p = (PUCHAR)((unsigned char *)batch->saddr) + tail - 6;
				// XXX Assumes little-endian CPU
				len = ((*p) << 8) | (*(p+1));
				PRINTK("Zeitnet: AAL5 len is %d.\n",len);
			
				endpt->rxNext->chan = chan->index;
				endpt->rxNext->length = len;
			}
			
			endpt->rxNext->pay_or_buf[buf_num] = batch->vaddr;

			// Get next batch ptr
			batch->paddr = 0UL;
			endpt->pool_dirty = (endpt->pool_dirty == endpt->pool_last) ?
				(endpt->pool):(endpt->pool_dirty+1);
			p_batch = batch->p_next;
			batch = endpt->pool_dirty;

			if (p_batch != MmGetPhysicalAddress(batch).LowPart) {
				ZEITNET_FREE_POOL *node = batch;
				DbgPrint("Zeitnet: Received batch not next dirty!\n");
				
				do {
					if (p_batch == node->p_next) {
						batch = (node == endpt->pool_last)?(endpt->pool):(node+1);
						break;
					}
					node = (node == endpt->pool_last)?(endpt->pool):(node+1);
				} while (node != endpt->pool_dirty);
				if (node == endpt->pool_dirty) {
					DbgPrint("Zeitnet: Didn't find batch pointer 0x%lx in pool ring!\n",
									 p_batch);
					goto drop;
				}
				endpt->pool_dirty = batch; // Reset pool_dirty to this entry
			}
			
			endpt->freeNext->buf = (ULONG)UNET_RXB_FREE;
			endpt->freeNext = (endpt->freeNext == endpt->freeLast)?
				endpt->freeFirst : endpt->freeNext+1;
			*endpt->freeWall = (long)endpt->freeNext - (long)endpt->freeFirst;
			endpt->freePrefetch--;
			
		}

		endpt->rxNext = (endpt->rxNext == endpt->rxLast)?
			endpt->rxFirst : endpt->rxNext + 1;

	drop:
		if (((((ULONG)pos) += 16) & 0xffff) == zeitnet->mbx_end[mn])
			pos = zeitnet->mbx[mn];
	}

	zeitnet->mbx_next[mn] = pos;
	zout(mwa, MTA(mn));
}


/* Zeitnet transmission ****************************************************/

// Update Tx wall for given endpoint
// Note that 'force' argument is ignored
static void
ZeitnetUpdateTxwall(ZEITNET_DEV *zeitnet,
										ZEITNET_ENDPT *endpt,
										BOOLEAN force) {

	int num_spins = 0;
	int mn = ZN_MBX_TX_0;
	ULONG mwa = zin(MWA(mn));
	ULONG pdirty = (zeitnet->p_mbx[mn] & ~0xffff) | mwa;
	PULONG dirty = (PULONG)((ULONG)zeitnet->mbx[mn] + (pdirty - (ULONG)zeitnet->p_mbx[mn]));
	PULONG pos = zeitnet->mbx_next[mn];
	PVOID tmp;
	
	PRINTK("--- mbx 0x%lx, end 0x%lx, pos 0x%lx, dirty 0x%lx, MWA 0x%lx\n",
				 zeitnet->mbx[mn],
				 zeitnet->mbx_end[mn],
				 pos, dirty, mwa);
	PRINTK("--- p_mbx 0x%lx, MSH 0x%lx, MSL 0x%lx, MBA 0x%lx, MTA 0x%lx\n",
				 zeitnet->p_mbx[mn],
				 zin(MSH(mn)),
				 zin(MSL(mn)),
				 zin(MBA(mn)),
				 zin(MTA(mn)));

	while (pos != dirty) {
		ZN_TX_MBX_ENTRY *mbxentry = (ZN_TX_MBX_ENTRY *)pos;
		ULONG conn = mbxentry->conn;
		PZEITNET_CHAN chan;
		ZEITNET_TX_SHADOW *shadow;

		PRINTK("Zeitnet: Tx MBX entry: 0x%lx\n", *(PULONG)pos);
		PRINTK("Zeitnet: Tx done on connection %d.\n", conn);
		chan = zeitnet->tx_conn_map[conn]; // XXX Assumes this is valid channel
		chan->endpt->num_freed++;

		shadow = &(chan->tx_ring_shadow[chan->tx_ring_dirty - chan->tx_ring]);

		if (shadow->page_id) {
			PRINTK("Zeitnet: Reaping 0x%x:0x%lx -> 0x%lx\n",
						 PAGE_OFFSET(shadow->page_id), PAGE_ADDR(shadow->page_id),
						 shadow->paddr);
			
			StlbDecRefcnt(PAGE_OFFSET(shadow->page_id),
										PAGE_ADDR(shadow->page_id),
										shadow->paddr);
		} else {
			// Multi-page buffer; packet directory used
			ZEITNET_PKTDIR_SHADOW *pdir = shadow->paddr;
			do {
				PRINTK("Zeitnet: Reaping 0x%x:0x%lx -> 0x%lx\n",
							 PAGE_OFFSET(pdir->shadow.page_id), PAGE_ADDR(pdir->shadow.page_id),
							 pdir->shadow.paddr);
				StlbDecRefcnt(PAGE_OFFSET(pdir->shadow.page_id),
											PAGE_ADDR(pdir->shadow.page_id),
											pdir->shadow.paddr);
				if (pdir->last) {
					MmFreeContiguousMemory(pdir->pktdir);
					break;
				}
				else pdir++;
			} while (1);
			ExFreePool((ZEITNET_PKTDIR_SHADOW *)shadow->paddr);
		}

		chan->tx_ring_full = FALSE;
		chan->tx_ring_dirty->status = 0; // Mark as invalid
		chan->tx_ring_dirty =
			(chan->tx_ring_dirty == chan->tx_ring_last)?
			(chan->tx_ring):(chan->tx_ring_dirty + 1);

	next_mbx_entry:
		if ((((ULONG)pos += 4) & 0xffff) == (ULONG)zeitnet->mbx_end[mn]) {
			pos = (ULONG)zeitnet->mbx[mn];
		}
	}

	zeitnet->mbx_next[mn] = pos;
	zout(mwa, MTA(mn));

	// Now bump Tx ring entries for this endpoint
	if (endpt->num_freed) {
		int num_walked = 0;

		UNET_TX_BUFFER_DESCRIPTOR *curWall = endpt->txDirty;

		while (endpt->num_freed) {
			UNET_TX_BUFFER_DESCRIPTOR *nextWall =
				(curWall == endpt->txLast)?(endpt->txFirst):(curWall+1);
			if (nextWall->chan != UNET_CHANNEL_ERROR) {
				nextWall->chan = UNET_CHANNEL_FREE;
				endpt->num_freed--;
			}
			curWall = nextWall;
			num_walked++;
		}
		endpt->txDirty = curWall;
		*(endpt->txWall) = (long)curWall - (long)endpt->txFirst;
	}
}

// Transmit buffers which cross multiple pages
static void
ZeitnetTxMultiple(ZEITNET_DEV *zeitnet,
									ZEITNET_ENDPT *endpt,
									PZEITNET_CHAN chan,
									ULONG vpage,
									ULONG voffset,
									int len) {
	ULONG paddr;
	KIRQL oldIrqL;
	int translen = 0;
	int pagenum = 0;
	int numpages = ((voffset + len) / PAGE_SIZE) + 1;
	int head = PAGE_SIZE - voffset;
	int tail = PAGE_SIZE - ((numpages * PAGE_SIZE) - (voffset + len));
	ZN_TX_PKTDIR *pktdir;
	ULONG p_pktdir;
	ZEITNET_PKTDIR_SHADOW *shadow;

	PRINTK("Zeitnet: Transmitting multi-page buffer: %d pages.\n", numpages);
	
	// Allocate and construct packet directory
	pktdir = MmAllocateContiguousMemory(numpages * sizeof(ZN_TX_PKTDIR),
																			topPhysAddr);
	if (!pktdir) {
		DbgPrint("Zeitnet: Can't allocate packet directory for %d-page buffer.\n",
						 numpages);
		endpt->txNext->chan = UNET_CHANNEL_ERROR;
		return;
	}
	p_pktdir = MmGetPhysicalAddress(pktdir).LowPart;
	PRINTK("Zeitnet: Packet directory 0x%lx at 0x%lx\n", pktdir, p_pktdir);

	shadow = ExAllocatePool(NonPagedPool, numpages * sizeof(ZEITNET_PKTDIR_SHADOW));
	if (!shadow) {
		DbgPrint("Zeitnet: Can't allocate packet directory shadow for %d-page buffer.\n",
						 numpages);
		endpt->txNext->chan = UNET_CHANNEL_ERROR;
		MmFreeContiguousMemory(pktdir);
		return;
	}

	while (pagenum < numpages) {
		paddr = StlbTranslate(endpt->index, vpage+(pagenum*PAGE_SIZE), UNET_STLB_PROT_RO, TRUE, NULL);
		if (!paddr) {
			PRINTK("Zeitnet: Can't translate address 0x%lx.\n", vpage+(pagenum*PAGE_SIZE));
			endpt->txNext->chan = UNET_CHANNEL_ERROR;
			return;
		}
		PRINTK("Zeitnet: Tx got 0x%x:0x%lx -> 0x%lx\n", endpt->index, vpage+(pagenum*PAGE_SIZE), paddr);

		shadow[pagenum].shadow.page_id = (endpt->index) | (vpage + (pagenum*PAGE_SIZE));
		shadow[pagenum].shadow.paddr = paddr;
		
		if (pagenum == 0) {
			pktdir[pagenum].size = head;
			pktdir[pagenum].addr = paddr | voffset;
			shadow[pagenum].last = FALSE;
		} else if (pagenum == (numpages - 1)) {
			pktdir[pagenum].size = tail | 0x80000000;
			pktdir[pagenum].addr = paddr;
			shadow[pagenum].last = TRUE;
			shadow[pagenum].pktdir = pktdir;
		} else {
			pktdir[pagenum].size = PAGE_SIZE;
			pktdir[pagenum].addr = paddr;
			shadow[pagenum].last = FALSE;
		}

		PRINTK("Zeitnet: Pktdir %d 0x%lx\n", pktdir[pagenum].size,
					 pktdir[pagenum].addr);

		pagenum++;
	}
	
	// Save shadow information
	{
		ZEITNET_TX_SHADOW *shadow2 =
			&(chan->tx_ring_shadow[chan->tx_ring_next - chan->tx_ring]);
		shadow2->page_id = 0;
		shadow2->paddr = shadow;
	}
		
#ifdef ZEITNET_DEBUG_PRINTPACKETS
	{
		int i;
		PRINTK("Zeitnet: Tx ");
		for (i = 0; i < MIN(50,len); i++) {
			PRINTK("%2.2x ", *((unsigned char *)(vpage | voffset) + i));
		}
		PRINTK("\n");
	}
#endif
	
	chan->tx_ring_next->pad = 0;
	chan->tx_ring_next->size = numpages * sizeof(ZN_TX_PKTDIR);
	chan->tx_ring_next->addr = p_pktdir;
	chan->tx_ring_next->status = 0xc0000000 | (0x01 << 17);
	
	KeRaiseIrql(POWER_LEVEL, &oldIrqL);
	zwait(0);
	zout(0x30000000 | (chan->tx_conn << 3), CMR);
	KeLowerIrql(oldIrqL);
	
	chan->tx_ring_next = (chan->tx_ring_next == chan->tx_ring_last) ?
		(chan->tx_ring):(chan->tx_ring_next + 1);
	
	// Must leave empty entry ('wall') after all valid Tx buffs
	{
		ZN_TX_DESCR *next_next;
		next_next =
			((chan->tx_ring_next) == (chan->tx_ring_last)) ?
			(chan->tx_ring) : (chan->tx_ring_next + 1);
		if (next_next == chan->tx_ring_dirty)
			chan->tx_ring_full = TRUE;
	}

	return;
}
	

// Transmit new entries on the given endpoint
static void
ZeitnetTransmit(
								ZEITNET_DEV *zeitnet,
								ZEITNET_ENDPT *endpt,
								int endptNum
								)
{
  int prefix = 0;
  int nextPrefix;
  int entry, len;
  int firstTx, i;
  unsigned short totalLen;
  volatile UNET_TX_BUFFER_DESCRIPTOR *txBegin;
	ULONG vaddr, vpage, voffset, paddr, paddr2;
	KIRQL oldIrqL;

	while(endpt->txNext->chan >= 0) {
		int channum = endpt->txNext->chan;
		PZEITNET_CHAN chan;
		
		PRINTK("Zeitnet: Transmit: Channel %d\n", channum);
		if (channum >= ZEITNET_MAX_CHANS || !(endpt->chans[channum].inUse)) {
			PRINTK("Zeitnet: Bad channel.\n");
			endpt->txNext->chan = UNET_CHANNEL_ERROR;
			prefix = 0;
			goto next_tx;
		}
		chan = &(endpt->chans[channum]);

		if (chan->tx_ring_full) {
			PRINTK("Zeitnet: Channel %d tx_ring is full.\n", channum);
			ZeitnetUpdateTxwall(zeitnet, endpt, FALSE);
			if (chan->tx_ring_full) return; // still full: abandon
		}

		len = endpt->txNext->length; // Save as user might change it during sleep
		if (len > ZEITNET_MAX_PDU) {
			DbgPrint("Zeitnet: Length too large.\n");
			endpt->txNext->chan = UNET_CHANNEL_ERROR;
			prefix = 0;
			goto next_tx;
		}

		PRINTK("Zeitnet: Tx 0x%lx, len %d\n", endpt->txNext->buf, len);

		nextPrefix = TX_CONTINUE(endpt->txNext->buf);
		if (!prefix && !nextPrefix) {
			// Single PDU entry
			vaddr = endpt->txNext->buf & ~0x01;
			vpage = PAGE_ADDR(vaddr); voffset = PAGE_OFFSET(vaddr);

			if (voffset + len > PAGE_SIZE) {
				// Buffer spans two pages
				ZeitnetTxMultiple(zeitnet, endpt, chan, vpage, voffset, len);
				goto next_tx;
			}

			// Buffer fits in one page
			paddr = StlbTranslate(endptNum, vpage, UNET_STLB_PROT_RO, TRUE, NULL);
			if (!paddr) {
				PRINTK("Zeitnet: Can't translate address.\n");
				endpt->txNext->chan = UNET_CHANNEL_ERROR;
				prefix = 0;
				goto next_tx;
			}
			PRINTK("Zeitnet: Tx got 0x%lx -> 0x%lx\n", vaddr, paddr|voffset);

			// Save shadow information
			{
				ZEITNET_TX_SHADOW *shadow =
					&(chan->tx_ring_shadow[chan->tx_ring_next - chan->tx_ring]);
				shadow->page_id = (vpage | endpt->index);
				shadow->paddr = paddr;
			}

#ifdef ZEITNET_DEBUG_PRINTPACKETS
			{
				int i;
				PRINTK("Zeitnet: Tx ");
				for (i = 0; i < len; i++) {
					PRINTK("%2.2x ", *((unsigned char *)vaddr + i));
				}
				PRINTK("\n");
			}
#endif

			chan->tx_ring_next->pad = 0;
			chan->tx_ring_next->size = len;
			chan->tx_ring_next->addr = paddr | voffset; 
			chan->tx_ring_next->status = 0xe0000000 | (0x01 << 17);

			PRINTK("--- Tx: Pushed into 0x%lx, first 0x%lx, last 0x%lx\n",
						 chan->tx_ring_next, chan->tx_ring, chan->tx_ring_last);

			KeRaiseIrql(POWER_LEVEL, &oldIrqL);
			zwait(0);
		 	zout(0x30000000 | (chan->tx_conn << 3), CMR);
			KeLowerIrql(oldIrqL);

			chan->tx_ring_next = (chan->tx_ring_next == chan->tx_ring_last) ?
				(chan->tx_ring):(chan->tx_ring_next + 1);

			PRINTK("--- Tx: Ring next now 0x%lx, dirty 0x%lx\n",
						 chan->tx_ring_next, chan->tx_ring_dirty);

			// Must leave empty entry ('wall') after all valid Tx buffs
			{
				ZN_TX_DESCR *next_next;
				next_next =
					((chan->tx_ring_next) == (chan->tx_ring_last)) ?
					(chan->tx_ring) : (chan->tx_ring_next + 1);
				if (next_next == chan->tx_ring_dirty)
					chan->tx_ring_full = TRUE;
			}

		} else {
			PRINTK("Zeitnet: Chained Tx PDUs not yet supported.\n");
			prefix = 0;
			goto next_tx;
		}

	next_tx:
		endpt->txNext = (endpt->txNext == endpt->txLast)?(endpt->txFirst):(endpt->txNext + 1);
		
	}
	return;
}

int
ZeitnetService(
							 IN UNET_ENDPOINT_ID index
							 )
{
  ZEITNET_DEV *zeitnet;
  ZEITNET_ENDPT *endpt;

  if(index < 0 || index > ZEITNET_MAX_ENDPT || zeitnetEndpt[index] == 0)
    return FALSE;

  endpt = zeitnetEndpt[index];
  zeitnet = endpt->dev;

  // Push any new Tx fifo entries onto the device Tx ring
  ZeitnetTransmit(zeitnet, endpt, index);

	// Update Tx wall for endpoint
	ZeitnetUpdateTxwall(zeitnet, endpt, FALSE);
			
	// Start prefetch for Rx buffers
	if (endpt->freePrefetch <= START_PREFETCH_THRESH) PfStart(zeitnet, index);

  return TRUE;
}

/* Collect all i/f routines in a struct **************************************/

static void
ZeitnetFillAdapter(
									 IN PUNET_ADAPTER UDIAdapter
									 )
{
  UDIAdapter->createEndpoint = ZeitnetCreateEndpoint;
  UDIAdapter->destroyEndpoint = ZeitnetDestroyEndpoint;
  UDIAdapter->activateChannel = ZeitnetActivateChannel;
  UDIAdapter->deactivateChannel = ZeitnetDeactivateChannel;
  UDIAdapter->adapterInfo = ZeitnetAdapterInfo;
  UDIAdapter->transmit = ZeitnetService;
  UDIAdapter->stopAdapter = 0;
	UDIAdapter->adapterIoctl = ZeitnetAdapterIoctl;
}

